"""Experimental 'python -m chemfp' command driver

This code is undocumented and only lightly tested. The current goal is
to get feedback from friendly users.

"""
# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

import sys
import argparse

import os

from . import commandline
from .commandline import cmdsupport

_commands = {}
_command_list = []

def command(command_name):
    def f(command_func):
        _commands[command_name] = command_func
        _command_list.append(command_name)
        return command_func
    return f

@command("simsearch")
def simsearch_command():
    from .commandline import simsearch
    simsearch.run()

@command("rdkit2fps")
def rdkit2fps_command():
    commandline.run_rdkit2fps()

@command("oe2fps")
def oe2fps_command():
    commandline.run_oe2fps()

@command("ob2fps")
def ob2fps_command():
    commandline.run_ob2fps()

@command("cdk2fps")
def cdk2fps_command():
    commandline.run_cdk2fps()

@command("sdf2fps")
def sdf2fps_command():
    from .commandline import sdf2fps
    sdf2fps.run()

@command("fpcat")
def fpcat_command():
    from .commandline import fpcat
    fpcat.run()
#### 

# Some bonus commands that aren't (yet?) command-line tools

def _add_in_out_args(parser):
    parser.add_argument(
        "--in", metavar="FORMAT", dest="input_format",
        choices=("fps", "fps.gz", "fps.zst", "fpb", "fpb.gz", "fpb.zst", "flush"),
        help="input fingerprint format. One of fps or fpb (with optional gz or zst compression), or flush. (default guesses from filename or is fps)")
    
    parser.add_argument(
        "-o", "--output", metavar="FILENAME",
        help="save the fingerprints to FILENAME (default=stdout)")
    parser.add_argument(
        "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "fps.zst", "fpb", "flush"),
        help="output fingerprint format. One of fps, fps.gz, fps.zst, fpb, or flush. (default guesses from output filename, or is 'fps')")

    parser.add_argument("filename", nargs="?",
                            help = "fingerprint filename. If not given, read from stdin.")
    
@command("count")
def count_command():
    from . import open as open_fingerprints
    parser = argparse.ArgumentParser(
        prog = "count",
        description = "count the nunber of fingerprints in the file",
        )
    parser.add_argument(
        "--in", metavar="FORMAT", dest="input_format",
        choices=("fps", "fps.gz", "fps.zst", "fpb", "fpb.gz", "fpb.zst", "flush"),
        help="input fingerprint format. One of fps or fpb (with optional gz or zst compression), or flush. (default guesses from filename or is fps)")
    parser.add_argument("filename", nargs="?",
                            help = "fingerprint filename. If not given, read from stdin.")

    args = parser.parse_args()

    reader = open_fingerprints(args.filename, format=args.input_format)
    if hasattr(reader, "arena"):
        print(len(reader))
    else:
        n = 0
        for block in reader.iter_blocks():
            n += block.count(b"\n")
        print(n)


@command("first")
def first_command():
    import itertools
    from . import open as open_fingerprints
    from . import open_fingerprint_writer
    from . import types
    parser = argparse.ArgumentParser(
        prog = "first",
        description = "select up to the the first N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints")
    _add_in_out_args(parser)

    args = parser.parse_args()

    reader = open_fingerprints(args.filename, format=args.input_format)
    if hasattr(reader, "arena"):
        subset = reader[:args.n]
        subset.save(args.output, format=args.output_format)
        return

    with open_fingerprint_writer(args.output, format=args.output_format,
                                     metadata=reader.metadata) as writer:
        writer.write_fingerprints(itertools.islice(reader, 0, args.n))
    
    
@command("last")
def last_command():
    import collections
    from . import open as open_fingerprints
    from . import open_fingerprint_writer
    from . import types
    parser = argparse.ArgumentParser(
        prog = "last",
        description = "select up to the the last N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints")
    _add_in_out_args(parser)

    args = parser.parse_args()

    reader = open_fingerprints(args.filename, format=args.input_format)
    if hasattr(reader, "arena"):
        subset = reader[:args.n]
        subset.save(args.output, format=args.output_format)
        return

    with open_fingerprint_writer(args.output, format=args.output_format,
                                     metadata=reader.metadata) as writer:
        writer.write_fingerprints(collections.deque(reader, maxlen=args.n))


@command("sample")
def sample_command():
    from . import load_fingerprints
    from . import types
    parser = argparse.ArgumentParser(
        prog = "sample",
        description = "sample N fingerprints")
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of fingerprints to sample (default: 10)")
    parser.add_argument("--upper-bound", "-u", action="store_true",
                            help="use all the fingerprints if there are fewer than 'n' of them")
    parser.add_argument("--no-reorder", action="store_true",
                            help="do not reorder the sampled fingerprints by popcount")
    parser.add_argument("--seed", type=int, default=None,
                            help="RNG seed to use for sampling (default: use the system RNG)")
    _add_in_out_args(parser)
    args = parser.parse_args()

    arena = load_fingerprints(args.filename, format=args.input_format, reorder=False)

    n = args.n
    if args.upper_bound:
        n = min(n, len(arena))
    elif n > len(arena):
        sys.stderr.write("Sample size %d is larger than the number of fingerprints (%d).\n" % (
            n, len(arena)))
        sys.stderr.write("Use a smaller value for --n or specify --upper-bound to use all available fingerprints.\n")
        raise SystemExit(1)
    sample = arena.sample(n, reorder=not args.no_reorder, rng=args.seed)
    sample.save(args.output, format=args.output_format)
    

# @command("train_test_split")
#### 
@command("fpb_text")
def fpb_text_command():
    from . import io, fpb_io
    
    def parse_label(s):
        try:
            label = s.encode("ascii")
        except UnicodeEncodeError:
            raise argparse.ArgumentTypeError("label must be ASCII")

        if b"\n" in label:
            raise argparse.ArgumentTypeError("label may not contain a newline")
        return label
        
    parser = argparse.ArgumentParser(
        prog = "fpb_text",
        description = "show the TEXT sections of an FPB file")
    parser.add_argument("--list-labels", action="store_true",
                            help="only show the labels of the TEXT blocks")
    parser.add_argument("--label", action="append", type=parse_label,
                            help="only show the named label(s). Can be used more than once.")
    parser.add_argument("--in", metavar="FORMAT", dest="source_format",
                            choices=("fpb", "fpb.gz", "fpb.zst"),
                            help="input FPB format")
    parser.add_argument("filename")

    args = parser.parse_args()

    if args.label is None:
        labels = None
    else:
        labels = set(args.label)
        
    # In principle I could work directly on the FPB file contents.
    # It's easier to load the file into memory and extract the data from there.
    source = args.filename
    format_name, compression = io.normalize_input_format(source, args.source_format)
    try:
        arena = fpb_io.open_fpb(source, format_name, compression, allow_mmap=True)
    except (ValueError, IOError) as err:
        sys.stderr.write("Cannot open FPB file: %s\n" % (err,))
        raise SystemExit(9)
    
    # If mmap'ed, the arena starts with FPB_MAGIC.
    # If read(), it does not.
    check_magic = arena.arena[:8].startswith(fpb_io.FPB_MAGIC)

    # Don't need to check for errors as open_fpb() did that.
    for chunk in fpb_io.iter_chunks(arena.arena, check_magic=check_magic):
        if chunk.tag == b"TEXT":
            label, nl, content = chunk.content[chunk.start:chunk.end].partition(b"\n")
            if args.list_labels:
                print(label.decode("ascii"))
            elif (labels is None) or (label in labels):
                label = label.decode("ascii")
                print(label)
                print("="*len(label))
                print(content.decode("utf8"))
                print("\n\n")
    

    
#### 
@command("fpb_dump")
def fpb_dump_command():
    from . import fpb_io
    fpb_io.main()

#### 
@command("report")
def report_command():
    parser = argparse.ArgumentParser(
        description = "report similarity search implementation details")
    args = parser.parse_args()
    from . import bitops
    bitops.use_environment_variables()
    bitops.print_report()


@command("license")
def license_command():
    parser = argparse.ArgumentParser(
        description = "show the chemfp license status")
    args = parser.parse_args()
        
    try:
        cmdsupport.license_check()
    except SystemExit:
        pass

# Text toolkit-related functions

_sdf_compression = {
    None: "auto",
    "sdf": "none",
    "sdf.gz": "gz",
    "sdf.zst": "zst",
    "sdf.bz2": "bz2",
}

def get_sdf_compression(parser, source, format):
    from . import _text_toolkit
    if format is None:
        compression = "auto"
    elif format == "sdf":
        compression = "none"
    elif format.startswith("sdf."):
        compression = format[4:]
    else:
        raise AssertionError(compression)

    try:
        return _text_toolkit._get_sdf_compression(source, compression)
    except ValueError as err:
        msg = str(err)
        assert msg.startswith("The text toolkit does not support "), msg
        assert msg.endswith("compression"), msg
        failure = msg[msg.index("'"):msg.rindex("'")]
        parser.error("Unsupported compression type '%s'" % (failure,))

## sdf_count: count the number of records in an SD file
@command("sdf_count")
def sdf_count_command():
    from . import text_toolkit, ParseError, gzio
    # from . import io as chemfp_io
    # import io
    
    parser = argparse.ArgumentParser(
        description = "count the number of records in an SD file",
    )
    ## g = parser.add_mutually_exclusive_group()
    ## g.add_argument("--fast", action="store_true",
    ##                    help=r"use simple text matching for text matching '\n\n$$$$\n")
    ## g.add_argument("--grep", action="store_true",
    ##                    help=r"use grep to count matches for '$$$$\n")
    parser.add_argument("--in", dest="input_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="input file format (default auto-detects based on extension)")
    parser.add_argument("--size", action="store_true",
                            help="also include the total number of bytes in the output")
    parser.add_argument("filename", nargs="*",
                            help="input filename(s). If not specified, read from stdin.")

    args = parser.parse_args()
    if args.filename is None:
        filenames = [None] # stdin
    else:
        filenames = args.filename

    include_size = args.size

    has_error = False
    total_n = total_size = 0
    for filename in filenames:
        n = size = 0
        compression = get_sdf_compression(parser, filename, args.input_format)
        try:
            reader = text_toolkit.read_sdf_records(filename, compression=compression)
        except IOError as err:
            sys.stderr.write("Cannot open input file: %s -- skipping\n" % (err,))
            has_error = True
            continue
        ## except Exception as err:
        ##     raise SystemExit("Cannot open input file: %s -- skipping\n" % (err,))

        try:
            with reader:
                if include_size:
                    len_ = len
                    n = 0
                    for record in reader:
                        size += len_(record)
                        n += 1
                else:
                    n = sum(1 for record in reader)
                
        except ParseError as err:
            sys.stderr.write("Cannot process input file: %s -- skipping\n" % (err,))
            has_error = True
            continue
        except gzio.BadGzipFile as err:
            sys.stderr.write("Failure decompressing input file: %s -- skipping\n" % (err,))
            has_error = True
            continue
        except IOError as err:
            sys.stderr.write("Failure reading input file: %s -- skipping\n" % (err,))
            has_error = True
            continue

        if filename is None:
            if include_size:
                sys.stdout.write("%10d\t%10d\n" % (n, size))
            else:
                sys.stdout.write("%10d\n" % (n,))
        else:
            if include_size:
                sys.stdout.write("%10d\t%10d\t%s\n" % (n, size, filename))
            else:
                sys.stdout.write("%10d\t%s\n" % (n, filename))
        total_size += size
        total_n += n
        sys.stdout.flush()

    if len(filenames) > 1:
        if include_size:
            sys.stdout.write("%10d\t%10d\t%s\n" % (total_n, total_size, "total"))
        else:
            sys.stdout.write("%10d\t%s\n" % (total_n, "total"))
            
    if has_error:
        raise SystemExit(1)

## sdf_first: select the first 'n' records in an SD file
@command("sdf_first")
def sdf_first_command():
    from . import io, types, text_toolkit
    import itertools
    parser = argparse.ArgumentParser(
        description = "select up to the first N records in an SD file",
    )
    parser.add_argument("-n", "--n", type=types.nonnegative_int, default=10,
                            help="number of records (default: 10)")
    parser.add_argument("--output", "-o", dest="destination",
                            metavar="FILENAME",
                            help="output file. If not specified, write to stdout.")
    parser.add_argument("--in", dest="input_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="input file format (default auto-detects based on extension or is 'sdf')")
    parser.add_argument("--out", dest="output_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="output file format (default auto-detects based on extension or is 'sdf')")
    # TODO: add support for compression level
    parser.add_argument("filename", nargs="?",
                            help="input filename. If not specified, read from stdin.")

    args = parser.parse_args()
    input_compression = get_sdf_compression(parser, args.filename, args.input_format)
    output_compression = get_sdf_compression(parser, args.destination, args.output_format)

    try:
        reader = text_toolkit.read_sdf_records(args.filename, input_compression)
    except Exception as err:
        raise SystemExit("Cannot open input file: %s" % (err,))

    with reader:
        try:
            outfile, outfile_close = io.open_compressed_output(args.destination, output_compression, "sdf_first")
        except Exception as err:
            raise SystemExit("Cannot open output file: %s" % (err,))

        try:
            for rec in itertools.islice(reader, 0, args.n):
                outfile.write(rec)
        finally:
            outfile_close()
    
## sdf_tags: list the tags used in an SD file

@command("sdf_tags")
def sdf_tags_command():
    from . import text_toolkit, ParseError, gzio
    import collections
    # from . import io as chemfp_io
    # import io
    
    parser = argparse.ArgumentParser(
        description = "list the tags used in an SD file",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--list", action="store_true",
                           help="list the unique tags (default)")
    group.add_argument("--count", action="store_true",
                            help="count the number of occurrences of each tag")
    parser.add_argument("--value", action="store_true",
                            help="for --list, when a tag is seen for the first time, also show its value")
    parser.add_argument("--multiline", action="store_true",
                            help="for --count, indicate if the tag value has multiple lines")
    parser.add_argument("--in", dest="input_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="input file format (default auto-detects based on extension)")
    parser.add_argument("-v", "--verbose", action="store_true",
                            help="show progress information")
    parser.add_argument("filename", nargs="*",
                            help="input filename(s). If not specified, read from stdin.")

    args = parser.parse_args()
    if args.filename is None:
        filenames = [None] # stdin
    else:
        filenames = args.filename

    verbose = args.verbose
    stdout = getattr(sys.stdout, "buffer", sys.stdout)

    do_list = args.list
    do_count = args.count
    do_multiline = args.multiline
    if not do_count and not do_list:
        do_list = True

    show_value = args.value
    
    has_error = False
    if do_list:
        seen = set()
    else:
        seen = collections.Counter()
        multiline = {}
        def show_counts():
            for tag, count in seen.most_common():
                if multiline.get(tag, False):
                    msg = "%10d\t%s\t(multiline)\n" % (count, tag.decode("utf8", errors="replace"))
                else:
                    msg = "%10d\t%s\n" % (count, tag.decode("utf8", errors="replace"))
                sys.stdout.write(msg)
            
    recno = 0
    try:
        for filename in filenames:
            compression = get_sdf_compression(parser, filename, args.input_format)
            try:
                reader = text_toolkit.read_sdf_records(filename, compression=compression)
            except IOError as err:
                sys.stderr.write("Cannot open input file: %s -- skipping\n" % (err,))
                has_error = True
                continue

            if verbose:
                sys.stderr.write("Processing %s\n" % (filename,))
                sys.stderr.flush()

            try:
                with reader:
                    for recno, record in enumerate(reader, recno+1):
                        if verbose and recno % 100000 == 0:
                            sys.stderr.write("Processed %d records. #unique tags found: %d\n" % (recno, len(seen)))
                            sys.stderr.flush()
                        if do_list:
                            for tag, value in text_toolkit.get_sdf_tag_pairs(record):
                                if tag not in seen:
                                    seen.add(tag)
                                    if show_value:
                                        stdout.write(tag + b"\t" + value + b"\n")
                                    else:
                                        stdout.write(tag + b"\n")
                                    stdout.flush()
                        elif do_multiline:
                            for tag, value in text_toolkit.get_sdf_tag_pairs(record):
                                seen[tag] += 1
                                if b"\n" in value:
                                    multiline[tag] = True
                        else:
                            seen.update(tag for tag, value in text_toolkit.get_sdf_tag_pairs(record))

            except ParseError as err:
                sys.stderr.write("Cannot process input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
            except gzio.BadGzipFile as err:
                sys.stderr.write("Failure decompressing input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
            except IOError as err:
                sys.stderr.write("Failure reading input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
    
    except KeyboardInterrupt:
        if do_count:
            sys.stderr.write("KeyboardInterrupt received after reading %d records.\n" % (recno,))
            show_counts()
        raise
            
    if do_count:
        show_counts()        
    
    if has_error:
        raise SystemExit(1)

            
## sdf_split: split the records of an SD file into multiple files

def parse_byte_count(s):
    c = s[-1:]
    if c in "BKMG":
        if c == "B":
            factor = 1
        elif c == "K":
            factor = 1000
        elif c == "M":
            factor = 1000000
        elif c == "G":
            factor = 1000000000
        s = s[:-1]
    else:
        factor = 1

    if not s.isdigit():
        raise ValueError("must be a number followed by optional scaling factor 'B', 'K', 'M', or 'G'")
    return int(s) * factor
            
@command("sdf_split")
def sdf_split_command():
    from . import io, types, text_toolkit, ParseError

    parser = argparse.ArgumentParser(
        description = "split the contents of the input file into 1 or more output files",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-n", "--n", type=types.nonnegative_int,
                            help="number of records per file (default: 1000)")
    group.add_argument("-b", "--byte-count", type=parse_byte_count, default=None,
                            help="switch a new file once the current file size exceeds this size")
    ### TODO: what about this? Figure out the sizes
    ## parser.add_argument("-c", "--chunks", type=types.nonnegative_int, default=None,
    ##                         help="total number of output files")
    parser.add_argument("--template", "-t",
                            metavar="STR", default="{filename}_{n:04d}.sdf",
                            help="Template for the output file. Default is '{filename}_{n:04d}.sdf'")
    parser.add_argument("--from", type=types.nonnegative_int, default=0, dest="from_",
                            help="First output index value. Must be a non-negative integer. (default: 0)")
    parser.add_argument("--in", dest="input_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="input file format (default auto-detects based on extension or is 'sdf')")
    parser.add_argument("--out", dest="output_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="output file format (default auto-detects based on template extension or is 'sdf')")
    parser.add_argument("-v", "--verbose", action="store_true",
                            help="show progress information")
    # TODO: add support for compression level
    parser.add_argument("filename", nargs="?",
                            help="input filename. If not specified, read from stdin.")

    args = parser.parse_args()
    input_compression = get_sdf_compression(parser, args.filename, args.input_format)

    template = args.template
    try:
        s = template.format(dirname="/there", sep="/", basename="blah", filename="stdin", n=456)
        t = template.format(dirname="/there", sep="/", basename="blah", filename="stdin", n=987)
    except ValueError as err:
        raise SystemExit("Unable to use --suffix %r: %s" % (suffix, err))
    if s == t:
        raise SystemExit("Unable to use --suffix %r: template must use {n} to change based on the output file index")
    output_compression = get_sdf_compression(parser, s, args.output_format)


    if args.filename is None:
        template_kwargs = dict(
            dirname = ".",
            sep = os.path.sep,
            basename = "stdin",
            filename = "stdin",
            )
    else:
        dirname, basename = os.path.split(args.filename)
        if not dirname:
            dirname = "."
        template_kwargs = dict(
            dirname = dirname,
            sep = os.path.sep,
            basename = basename,
            fiename = args.filename,
            )

    # Figure out how I'm going to split the files
    n = args.n
    byte_count = args.byte_count
    if byte_count is None:
        if n is None:
            n = 1000
        option = 0 # number of records
    else:
        option = 1 # byte count

        
    try:
        reader = text_toolkit.read_sdf_records(args.filename, input_compression)
    except Exception as err:
        raise SystemExit("Cannot open input file: %s" % (err,))

    output_count = args.from_
    outfile = outfile_close = None
    num_records = size = 0
    need_new_file = False
    try:
        try:
            for record in reader:
                if option == 1:
                    record_size = len(record)
                    if size == 0:
                        need_new_file = False # Always write at least one record
                    elif size + record_size > byte_count:
                        # Finish this file before writing a new record
                        need_new_file = True

                if need_new_file:
                    outfile_close()
                    outfile = None
                    need_new_file = False

                if outfile is None:
                    # Need to create a new output file.
                    output_filename = template.format(
                        n=output_count, **template_kwargs)
                    output_count += 1
                    if args.verbose:
                        sys.stderr.write("Opening %s\n" % (output_filename,))
                        sys.stderr.flush()
                    try:
                        outfile, outfile_close = io.open_compressed_output(output_filename, output_compression, "sdf_split")
                    except Exception as err:
                        raise SystemExit("Cannot open output file: %s" % (err,))
                    num_records = size = 0

                outfile.write(record)
                    
                if option == 0:
                    num_records += 1
                    need_new_file = (num_records == n)
                    
                elif option == 1:
                    size += record_size

                else:
                    raise AssertionError(option)
                    
        except ParseError as err:
            raise SystemExit("Error reading the file: %s" % (err,))
    finally:
        if outfile_close is not None:
            outfile_close()
        reader.close()
                
    if output_count == 0:
        # Nothing written so far. Create an empty output file.
        # (TODO: add a way to suppress this? filter uses '-e/--elide-empty-files')
        output_filename = prefix + suffix.format(n=0)
        try:
            outfile, outfile_close = io.open_compressed_output(output_filename, output_compression, "sdf_split")
        except Exception as err:
            raise SystemExit("Cannot open output file: %s" % (err,))
        outfile_close()

# sdf2csv: convert the tags in an SD file into CSV format
@command("sdf2csv")
def sdf2csv_command():
    from . import text_toolkit, ParseError, gzio, io
    import csv
    import io as python_io
    
    parser = argparse.ArgumentParser(
        description = "convert tags from an SD file into CSV format",
    )
    # TODO: 'tab' or other options?
    parser.add_argument("-t", "--tag", action="append",
                            help="tag to export (use once for each tag)")
    parser.add_argument("-T", "--no-tag", action="append",
                            help="tag to not include, if using tags from the first record (use once for each tag)")
    parser.add_argument("--no-id", action="store_true",
                            help="do not include the id in the output")
    parser.add_argument("--id-header", default="title",
                            help="text to use for the id header column instead of 'title'")
    parser.add_argument("--dialect", choices=("excel", "excel-tab", "unix"),
                            help="csv dialect")
    parser.add_argument("--missing",
                            help="value to use for missing tag")
    parser.add_argument("--in", dest="input_format",
                            metavar="FORMAT", choices=("sdf", "sdf.gz", "sdf.zst"),
                            help="input file format (default auto-detects based on extension)")
    parser.add_argument("-o", "--output", dest="destination",
                            help="csv output filename (default: stdout)")
    parser.add_argument("filename", nargs="*",
                            help="input filename(s). If not specified, read from stdin.")

    args = parser.parse_args()
    if args.filename is None:
        filenames = [None] # stdin
    else:
        filenames = args.filename

    if args.missing is None:
        missing = b""
    else:
        missing = args.missing.encode("utf8")

    tags = []
    headers = []
    if not args.no_id:
        tags.append(None)
        headers.append(args.id_header)

    if args.tag:
        get_tags_from_first_record = False
        # User-specified
        for tag in args.tag:
            if "=" not in tag:
                tags.append(tag.encode("utf8"))
                headers.append(tag)
                continue
            
            if tag[0] == "=":
                # Use the rest of the term as the tag.
                tag = tag[1:]
                tags.append(tag.encode("utf8"))
                headers.append(tag)
                continue

            header, _, tag = tag.partition("=")
            tags.append(tag.encode("utf8"))
            headers.append(header)
    else:
        get_tags_from_first_record = True
        if args.no_tag:
            exclude = set(tag.encode("utf8") for tag in args.no_tag)
        else:
            exclude = set()

    try:
        outfile, outfile_close = io.open_compressed_output(args.destination, None, "sdf2csv")
    except IOError as err:
        raise SystemExit("Cannot write output file: %s" % (err,))

    
    csv_writer = csv.writer(python_io.TextIOWrapper(outfile), args.dialect)
    def write_header():
        csv_writer.writerow(headers)

    if not get_tags_from_first_record:
        write_header()
    
    try:
        has_error = False
        for filename in filenames:
            compression = get_sdf_compression(parser, filename, args.input_format)
            # TODO: if 1 or 2 terms then use the read_sdf_ids_and_values() iterator
            try:
                reader = text_toolkit.read_sdf_records(filename, compression=compression)
            except IOError as err:
                sys.stderr.write("Cannot open input file: %s -- skipping\n" % (err,))
                has_error = True
                continue

            try:
                with reader:
                    for record in reader:
                        if get_tags_from_first_record:
                            # Need to get the tags from the first record
                            get_tags_from_first_record = False
                            for (tag, value) in text_toolkit.get_sdf_tag_pairs(record):
                                if tag in exclude:
                                    continue
                                headers.append(tag.decode("utf8"))
                                tags.append(tag)
                            write_header()

                        d = {}
                        # In case of duplicates, use the first
                        for (tag, value) in text_toolkit.get_sdf_tag_pairs(record):
                            if tag not in d:
                                d[tag] = value
                        d[None] = text_toolkit.get_sdf_id(record)

                        values = [d.get(tag, missing).decode("utf8") for tag in tags]
                        csv_writer.writerow(values)

            except ParseError as err:
                sys.stderr.write("Cannot process input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
            except gzio.BadGzipFile as err:
                sys.stderr.write("Failure decompressing input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
            except IOError as err:
                sys.stderr.write("Failure reading input file: %s -- skipping\n" % (err,))
                has_error = True
                continue
    finally:
        outfile_close()
        
    if has_error:
        raise SystemExit(1)


# Chemistry toolkit-related functions
@command("toolkits")
def toolkits_command():
    parser = argparse.ArgumentParser(
        description = "show cheminformatics toolkit availability")
    args = parser.parse_args()
    
    def _software():
        return software.partition("/")[2]
    
    try:
        import openbabel
    except ImportError as err:
        print("openbabel - not installed: %s" % (err,))
    else:
        from .openbabel_toolkit import software
        print("openbabel - version %s installed" % (_software(),))

    try:
        import rdkit
    except ImportError as err:
        print("rdkit - not installed: %s" % (err,))
    else:
        from .rdkit_toolkit import software
        print("rdkit - version %s installed" % (_software(),))

    try:
        import openeye
    except ImportError as err:
        print("oechem - not installed: %s" % (err,))
    else:
        from .openeye_toolkit import software
        
        from openeye import oechem, oegraphsim
        if not oechem.OEChemIsLicensed():
            print("oechem - version %s installed but OEChem is not licensed" % (_software(),))
        elif not oegraphsim.OEGraphSimIsLicensed():
            print("oechem - version %s installed and OEChem licensed but OEGraphSim is not licensed" % (_software(),))
        else:
            print("oechem - version %s installed and both OEChem and OEGraphSim are licensed" % (_software(),))


def _get_toolkit(opt, name1, name2, default_toolkit=None):
    from . import get_toolkit
    for name in (name1, name2):
        if name is None:
            continue
        try:
            tk = get_toolkit(name)
        except ValueError as err:
            sys.stderr.write("Cannot convert: %s\n" % (err,))
            raise SystemExit(1)
        return tk
    if default_toolkit is not None:
        return default_toolkit
    for name in ("rdkit", "openeye", "openbabel", "cdk"):
        try:
            tk = get_toolkit(name)
        except ValueError as err:
            continue
        sys.stderr.write("Using %s as neither --toolkit nor %s specified.\n" % (
            name, opt))
        return tk
    sys.stderr.write("No chemistry toolkit available.\n")
    raise SystemExit(1)

def _get_source_format(toolkit, source, format_name, reader_settings):
    try:
        format = toolkit.get_input_format_from_source(source, format_name)
        reader_args = format.get_reader_args_from_text_settings(reader_settings)
    except ValueError as err:
        sys.stderr.write(str(err) + "\n")
        raise SystemExit(1)
    return format, reader_args

def _get_destination_format(toolkit, destination, format_name, writer_settings):
    try:
        format = toolkit.get_output_format_from_destination(destination, format_name)
        writer_args = format.get_writer_args_from_text_settings(writer_settings)
    except ValueError as err:
        sys.stderr.write(str(err) + "\n")
        raise SystemExit(1)
    return format, writer_args

def _get_input_format(toolkit, format_name, reader_settings):
    try:
        format = toolkit.get_input_format(format_name)
        reader_args = format.get_reader_args_from_text_settings(reader_settings)
    except ValueError as err:
        sys.stderr.write(str(err) + "\n")
        raise SystemExit(1)
    return format, reader_args

def _get_output_format(toolkit, format_name, writer_settings):
    try:
        format = toolkit.get_output_format(format_name)
        writer_args = format.get_writer_args_from_text_settings(writer_settings)
    except ValueError as err:
        sys.stderr.write(str(err) + "\n")
        raise SystemExit(1)
    return format, writer_args

        
@command("convert")
def convert_command():
    from . import io
    
    parser = argparse.ArgumentParser(
        description = "convert from one struture format to another using a given toolkit"
        )
    parser.add_argument("--in", dest="in_format",
                            help="input structure format (default uses the input extension, or 'smi'")
    parser.add_argument("--out", dest="out_format",
                            help="output structure format (default uses the --output extension, or 'smi')")
    parser.add_argument("-o", "--output",
                            help="output filename (defaults to stdout)")
    parser.add_argument(
        "-R", metavar="NAME=VALUE", dest="reader_settings", action="append",
        help="specify a reader argument")
    cmdsupport.add_reader_aliases(parser)
    parser.add_argument(
        "-W", metavar="NAME=VALUE", dest="writer_settings", action="append",
        help="specify a writer argument")
    parser.add_argument("-t", "--toolkit")
    parser.add_argument("--in-toolkit")
    parser.add_argument("--out-toolkit")
    parser.add_argument("--via", dest="via_format", default="sdf")
    parser.add_argument(
        "--errors", choices=["strict", "report", "ignore"], default="ignore",
        help="how should structure parse errors be handled? (default=ignore)")

    parser.add_argument("filename", nargs="?",
                            help="input structure filename (defaults to stdin)")

    args = parser.parse_args()

    source = args.filename
    destination = args.output
    reader_settings = cmdsupport.get_reader_settings(parser, args)
    writer_settings = cmdsupport.get_writer_settings(parser, args)
    
    in_toolkit = _get_toolkit("--in-toolkit", args.in_toolkit, args.toolkit)
    out_toolkit = _get_toolkit("--out-toolkit", args.out_toolkit, args.toolkit, in_toolkit)

    error_handler = io.get_parse_error_handler(args.errors)

    if in_toolkit is out_toolkit:
        toolkit = in_toolkit
        input_format, reader_args = _get_source_format(toolkit, source, args.in_format, reader_settings)
        output_format, writer_args = _get_destination_format(toolkit, destination, args.out_format, writer_settings)
        
        with toolkit.read_ids_and_molecules(
                source, input_format, reader_args=reader_args, errors=error_handler) as reader:
            with toolkit.open_molecule_writer(
                            destination, output_format, writer_args=writer_args, errors=error_handler) as writer:
                    writer.write_ids_and_molecules(reader)
    else:
        via_input_format, via_reader_args = _get_input_format(out_toolkit, args.via_format, reader_settings)
        via_output_format, via_writer_args = _get_output_format(in_toolkit, args.via_format, writer_settings)
        
        input_format, reader_args = _get_source_format(in_toolkit, source, args.in_format, reader_settings)
        output_format, writer_args = _get_destination_format(out_toolkit, destination, args.out_format, writer_settings)
        
        with in_toolkit.read_ids_and_molecules(
                source, input_format, reader_args=reader_args, errors=error_handler) as reader:
            with out_toolkit.open_molecule_writer(
                            destination, output_format, writer_args=writer_args, errors=error_handler) as writer:
                
                for id, in_mol in reader:
                    # Transcode
                    rec = in_toolkit.create_string(
                        in_mol, via_output_format, writer_args=via_writer_args, errors=error_handler)
                    if rec is None:
                        continue
                    
                    out_mol = out_toolkit.parse_molecule(
                        rec, via_input_format, reader_args=via_reader_args, errors=error_handler)
                    if out_mol is None:
                        continue
                    
                    writer.write_id_and_molecule(id, out_mol)
    
###

epilog = """
Description of the available commands:    
 o simsearch: Tanimoto or Tversky fingerprint similarity search
 o rdkit2fps: use RDKit to generate fingerprints from a structure file
 o oe2fps: use OpenEye's OEChem and OEGraphSim to generate fingerprints
     from a structure file
 o ob2fps: use Open Babel to generate fingerprints from a structure file
 o cdk2fps: use CDK to generate fingerprints from a structure file
 o sdf2fps: extract fingerprints from a tag of records in an SD file
 o toolkits: show cheminformatics toolkit availability
 o fpcat: convert between fingerprint file formats
 o count: count the number of fingerprints
 o first: select the first 'n' fingerprints
 o last: select the last 'n' fingerprints
 o sample: select 'n' fingeprints at random 
 o fpb_dump: display low-level dump of the contents of an FPB file
 o fpb_add_cfpl: add or update the CFPL license key
 o report: report similarity search implementation details
 o license: report on the chemfp license status
 o sdf_count: count the number of records in an SD file
 o sdf_first: select the first 'n' records in an SD file
 o sdf_tags: list the tags in an SD file
 o sdf_split: split the records of an SD file into multiple files
 o sdf2csv: convert the tags in an SD file into CSV format
 o convert: convert from one structure format to another using a given toolkit

This interface is EXPERIMENTAL and UNSUPPORTED. Let me know if it's useful
or if there are other tools you would like.
"""
def _check_epilog():
    for line in epilog.splitlines():
        pass
    
parser = argparse.ArgumentParser(
    prog = "chemfp",
    usage = "python -m chemfp [-h] [-S] command [args]",
    description = "chemfp tools for working with cheminformatics fingerprints",
    formatter_class = argparse.RawDescriptionHelpFormatter,
    epilog = epilog)
parser.add_argument("--allow-signals", "-S", action = "store_true",
                        help = "let Python handle SIGPIPE and SIGITR signals instead of exiting")

def _get_command_list_str():
    terms = []
    for name in _command_list:
        terms.append(repr(name))
        terms.append(", ")
    assert len(terms) > 4
    del terms[-1]
    terms[-2] = " or "
    return "".join(terms)

_command_list_str = _get_command_list_str()
parser.add_argument("command", nargs="?",
                        help = "one of: " + _command_list_str)

parser.add_argument("args", nargs=argparse.REMAINDER,
                        help = "arguments passed to the subcommand; use command --help for details")

def main(argv=None):
    args = parser.parse_args(argv)
    command = args.command
    if command is None:
        parser.print_usage()
        print(epilog)
        return

    command_func = _commands.get(command, None)
    if command_func is None:
        parser.error("unsupported command: %r (Must be one of %s" % (
            command, _command_list_str))
        
    sys.argv[:] = [command] + args.args
    
    if args.allow_signals:
        def run(f):
            return f()
        cmdsupport.run = run

    command_func()

if __name__ == "__main__":
    import signal
    signal.signal(signal.SIGPIPE, signal.SIG_DFL) # Allow the output pipe to be closed
    main()
