"Internal support for the CDK toolkit"

# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

from __future__ import absolute_import

import sys
import re

from ._compat import BytesIO, unicode_type, bytes_type, integer_types, filename_types, PY2
                          

from . import _compat

from . import io
from . import base_toolkit
from . import _text_toolkit



#####

import jpype  # Must install JPype so chemfp can interface to the CDK jar
import jpype.imports # configure the import hooks
import jpype.nio

if not jpype.isJVMStarted():
    jpype.startJVM(None, '-Djava.awt.headless=true')

try:
    from org.openscience import cdk
except ImportError as err:
    raise ImportError("Unable to access the CDK jar via JPype. Is the jar on your CLASSPATH?: %s" % (
        err,))

import java
import javax

# Disable the InChI logging to stdout by increasing the log level
from org.apache.log4j import LogManager, Level
logger = LogManager.getLogger("net.sf.jnati")
logger.setLevel(Level.ERROR)


SOFTWARE = "CDK/" + str(cdk.CDK.getVersion())
EXTERNAL_NAME = "CDK"

has_inchi = True

_stdin = java.lang.System.in_
_stdout = java.lang.System.out

InvalidSmilesException = cdk.exception.InvalidSmilesException
STRICT = cdk.io.IChemObjectReader.Mode.STRICT
SmiFlavor = cdk.smiles.SmiFlavor

# cinfony uses
# cdk.silent.SilentChemObjectBuilder.getInstance()
_default_builder = cdk.DefaultChemObjectBuilder.getInstance()

_aromaticity_perceptor = cdk.aromaticity.Aromaticity(
    cdk.aromaticity.ElectronDonation.daylight(),
    cdk.graph.Cycles.or_(cdk.graph.Cycles.all(),
                             cdk.graph.Cycles.all(6)))

TITLE = cdk.CDKConstants.TITLE

SmilesParser = cdk.smiles.SmilesParser
SmilesGenerator = cdk.smiles.SmilesGenerator

_inchi_generator_factory = cdk.inchi.InChIGeneratorFactory.getInstance()

_getAllAtomContainers = cdk.tools.manipulator.ChemFileManipulator.getAllAtomContainers

#### File I/O

def close_chain(close1, close2):
    def close():
        close1()
        close2()
    return close

def _do_nothing():
    pass

import errno
import os
_reverse_error_message_dict = dict((os.strerror(i), i) for i in errno.errorcode)

def _get_error_message(err):
    msg = err.getMessage()
    if msg is None:
        msg = str(err)
    elif isinstance(err, java.lang.ArrayIndexOutOfBoundsException):
        msg = str(err)

    return msg
    
def FileNotFoundError(err):
    s = str(err.getMessage())
    # use rpartition() in case the filename contains a ' ('
    filename, _, message = s.rpartition(" (")
    message = message.rstrip(")")
    errnum = _reverse_error_message_dict.get(message, errno.ENOENT)
    
    if PY2:
        return IOError(errnum, message, filename)
    # Python will dispatch to the right subclass based on the errnum
    return OSError(errnum, message, filename)
    

def open_compressed_input_java(source, compression):
    if not compression:
        if source is None:
            return _stdin, _do_nothing
        if isinstance(source, filename_types):
            try:
                fis = java.io.FileInputStream(java.io.File(source))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            return fis, fis.close
        if isinstance(source, (java.io.InputStream, java.io.StringReader)):
            return source, _do_nothing
        raise ValueError("cdk does not support input from a Python file object")

    if compression == "gz":
        if source is None:
            gzis = java.util.zip.GZIPInputStream(_stdin)
            return gzis, gzis.close
        if isinstance(source, filename_types):
            try:
                fis = java.io.FileInputStream(java.io.File(source))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            gzis = java.util.zip.GZIPInputStream(fis)
            return gzis, close_chain(fis.close, gzis.close)
        if isinstance(source, (java.io.InputStream, java.io.StringReader)):
            gzis = java.util.zip.GZIPInputStream(source)
            return gzis, _do_nothing
        raise ValueError("cdk does not support input from a Python file object")

    if compression == "zst":
        # Defer the lookup until we know the source type because
        # the API should complain about not reading from fileobj
        # before it complains about missing zstd support.
        def _get_zstd():
            try:
                from com.github.luben import zstd
            except ImportError as err:
                raise ValueError("Please place zstd-jni on your CLASSPATH for zst compression support: %s"
                                     % (err,))
            return zstd
        
        if source is None:
            reader = _get_zstd().ZstdInputStream(_stdin)
            return reader, reader.close
        if isinstance(source, filename_types):
            try:
                fis = java.io.FileInputStream(java.io.File(source))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            reader = _get_zstd().ZstdInputStream(fis)
            return reader, close_chain(fis.close, reader.close)
        if isinstance(source, (java.io.InputStream, java.io.StringReader)):
            reader = _get_zstd().ZstdInputStream(source)
            return reader, _do_nothing
        raise ValueError("cdk does not support input from a Python file object")

    raise ValueError("CDK does not support compression type %r" % (compression,))

def open_string_input_java(content, compression, id_tag=None):
    if isinstance(content, unicode_type):
        if compression:
            raise ValueError("Compression does not work when the content is a Unicode string. "
                             "Use a byte string instead.")
        if id_tag is not None and not isinstance(id_tag, unicode_type):
            # XXX is this true?
            raise ValueError("id_tag must be a string, not bytes, when the content is a string")
        return java.io.ByteArrayInputStream(content.encode("utf8"));

    return java.io.ByteArrayInputStream(content)

def open_string_input_python(content, compression, id_tag=None):
    if isinstance(content, unicode_type):
        if compression:
            raise ValueError("Compression does not work when the content is a Unicode string. "
                             "Use a byte string instead.")
        if id_tag is not None and not isinstance(id_tag, unicode_type):
            # XXX is this true?
            raise ValueError("id_tag must be a string, not bytes, when the content is a string")
        return BytesIO(content.encode("utf8"))

    return BytesIO(content)

def open_compressed_output_java(destination, compression, external_name,
                                    encoding, encoding_errors, level):
    if not (encoding is None or encoding == "utf8"):
        raise ValueError("Unsupported encoding: %r" % (encoding,))

    if not compression:
        if destination is None:
            return _stdout, _do_nothing
        if isinstance(destination, filename_types):
            from errno import ENOENT
            try:
                fos = java.io.FileOutputStream(java.io.File(destination))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            return fos, fos.close
        if isinstance(destination, java.io.OutputStream):
            return fos, _do_nothing
        raise ValueError("cdk cannot write to a Python file object")

    if compression == "gz":
        # the Java standard library API doesn't support changing the compression level
        if destination is None:
            gzos = java.util.zip.GZIPOutputStream(_stdout)
            return gzos, gzos.close
        if isinstance(destination, filename_types):
            try:
                fos = java.io.FileOutputStream(java.io.File(destination))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            gzos = java.util.zip.GZIPInputStream(fos)
            return gzos, close_chain(gzos.close, fos.close)
        if isinstance(destination, java.io.OutputStream):
            gzos = java.util.zip.GZIPInputStream(destination)
            return gzos, gzos.close
        
        raise ValueError("cdk cannot write to a Python file object")

    if compression == "zst":
        try:
            from com.github.luben import zstd
        except ImportError as err:
            raise ValueError("Please place zstd-jni on your CLASSPATH for zst compression support: %s"
                                 % (err,))
        level = io._get_zst_level(level)
        if destination is None:
            zstdos = ZstdOutputStream(_stdout, level)
            return zstdos, zstdos.close
        if isinstance(destination, filename_types):
            try:
                fos = java.io.FileOutputStream(java.io.File(destination))
            except java.io.FileNotFoundException as err:
                raise FileNotFoundError(err)
            zstdos = ZstdOutputStream(_stdout, level)
            return gzos, close_chain(zstdos.close, fos.close)
        if isinstance(destination, java.io.OutputStream):
            zstdos = ZstdOutputStream(destination, level)
            return zstdos, zstdos.close

        raise ValueError("cdk cannot write to a Python file object")

    raise ValueError("Unsupported compression: " % (compression,))

def open_output_string_java():
    sio = java.io.StringWriter()
    def getvalue():
        return str(sio)
    return sio, sio.close, getvalue

def open_compressed_output_bytes_java(compression, level):
    if compression not in (None, "", "gz", "zst"):
        raise ValueError("CDK does not support compression type %r" % (compression,))
    if compression in (None, ""):
        bio = java.io.ByteArrayOutputStream()
        def getvalue():
            return bytes(bio.toByteArray())
        return bio, bio.close, getvalue
    if compression == "gz":
        bio = java.io.ByteArrayOutputStream()
        gzio = java.util.zip.GZIPOutputStream(bio)
        def getvalue():
            return bytes(bio.toByteArray())
        return gzio, close_chain(gzio.close, bio.close), getvalue
    if compression == "zst":
        raise ValueError("CDK does not yet support compression type %r" % (compression,))
    raise AssertionError

## class OutputToBytesJava(object):
##     def __init__(self, output_stream):
##         self.output_stream = output_stream

## def open_compressed_output_wrapped_java(
##         destination, compression, external_name,
##         encoding, encoding_errors, level):
##     output_stream, close = open_compressed_output_java(
##         destination, compression, external_name,
##         encoding, encoding_errors, level)
##     return OutputToBytesJava(output_stream), close

####
def _or_string(location):
    filename = location.filename
    if filename is None:
        return "<string>"
    return filename

def _or_bytes(location):
    filename = location.filename
    if filename is None:
        return "<bytes>"
    return filename

def _get_smiles_exception_msg(err):
    msg = str(err.getMessage())
    i = msg.find(", ")
    if i != -1:
        return msg[i+2:]
    return msg

def notify_smiles_error(error_handler, smiles, err, location=None):
    errmsg = _get_smiles_exception_msg(err)
    error_handler.error("CDK cannot parse the SMILES %s: %s" % (_compat.myrepr(smiles), errmsg), location)
    
class CDKFormatConfig(object):
    supports_io = True
    alternate_extensions = ()
    is_available = True

    def create_bytes(self, mol, id, writer_args, error_handler):
        s = self.create_string(mol, id, writer_args, error_handler)
        if s is not None:
            return s.encode("utf8")
        return None

# XXX this is a duplicate of the RDKit reader
def iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler, location, format_name,
                                         encoding, encoding_errors):
    it = iter_read_delimited_fields_as_bytes(infile, close, has_header, delimiter, error_handler, location, format_name)
    
    def iter_read_delimited_fields_as_strings(it, error_handler, format_name, encoding, encoding_errors):
        for field1_bytes, field2_bytes in it:
            if field1_bytes is None:
                field1 = None
            else:
                try:
                    field1 = field1_bytes.decode(encoding, encoding_errors)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the first field from the %s file, %r: %s"
                                        % (format_name, field1_bytes, err), location)
                    continue
            if field2_bytes is None:
                field2 = None
            else:
                try:
                    field2 = field2_bytes.decode(encoding, encoding_errors)
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the second field from the %s file, %r: %s"
                                        % (format_name, field2_bytes, err), location)
                    continue
            yield field1, field2
                
    return iter_read_delimited_fields_as_strings(it, error_handler, format_name, encoding, encoding_errors)

# XXX this is a duplicate of the RDKit reader
def iter_read_delimited_fields_as_bytes(infile, close, has_header, delimiter, error_handler, location, format_name):
    it = _iter_read_delimited_fields(infile, close, has_header, delimiter, error_handler, location, format_name)
    next(it)
    return it

# XXX this is a duplicate of the RDKit reader
def _iter_read_delimited_fields(infile, close, has_header, delimiter, error_handler, location, format_name):
    lineno = start_offset = end_offset = 0
    record = None
    def get_lineno():
        return lineno
    def get_recno():
        return lineno
    def get_offsets():
        return start_offset, end_offset
    def get_record():
        return record
    
    location.register(get_lineno = get_lineno,
                      get_recno = get_recno,
                      get_offsets = get_offsets,
                      get_record = get_record,
        )

    start_offset = 0
    split_two_terms_bytes = delimiter.split_two_terms_bytes
    try:
        yield  "Ready for processing!"
        file_iter = enumerate(infile, 1)
        if has_header:
            # Skip the first line
            header_line = next(file_iter, None)
            if header_line is None:
                return
            
        for lineno, record in file_iter:
            start_offset = end_offset
            end_offset = start_offset + len(record)
            line = record.rstrip(b"\r\n")
            terms = split_two_terms_bytes(line)

            if not terms:
                error_handler.error("Empty line in %s file" % (format_name,), location)
                continue

            yield terms
    finally:
        if close is not None:
            close()
        location.save(lineno = get_lineno(),
                      recno = get_recno(),
                      offsets = None,
                      record = None,
            )

##### iter_read_smiles_ids_and_molecules()

def iter_read_smiles_ids_and_molecules_cdk(
        reader, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors):
    if builder is None:
        builder = _default_builder
    it = _iter_read_smiles_ids_and_molecules_cdk(
        reader, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors)
    next(it)
    return it

def _iter_read_smiles_ids_and_molecules_cdk(
        reader, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors):
    delimiter_name = delimiter.name

    if delimiter_name is None or delimiter_name in ("to-eol", "native"):
        DELIMITER_METHOD = 0
    elif delimiter_name == "whitespace":
        DELIMITER_METHOD = 1
    elif delimiter_name in (" ", "\t"):
        DELIMITER_METHOD = 2
    else:
        raise AssertionError(delimiter_name)
    
    recno = 0
    lineno = 0
    mol = None
    error_line = None
    def get_mol():
        return mol
    if has_header:
        lineno += 1
    def get_lineno():
        return lineno
    def get_recno():
        return recno
    def get_record():
        if error_line is None:
            return None
        else:
            return str(error_line).encode("utf8") # is the "utf8" needed? XXX
    location.register(
        get_mol=get_mol,
        get_recno=get_recno,
        get_lineno=get_lineno,
        get_record=get_record)
    
    BAD_SMILES_INPUT = "bad.smiles.input"
    try:
        yield  "Ready for processing!"
        while reader.hasNext():
            mol = reader.next()
            recno += 1
            lineno += 1
            if mol.getAtomCount() == 0:
                error_line = mol.getProperty(BAD_SMILES_INPUT)
                if error_line is not None:
                    error_handler.error("CDK cannot parse a SMILES record", location)
                    error_line = None
                    continue
                else:
                    # Empty line. Treat it as a skipped record.
                    recno -= 1
                    continue

            id = mol.getTitle()
            if id is not None:
                id = str(id)
                if DELIMITER_METHOD == 0:
                    pass
                elif DELIMITER_METHOD == 1:
                    terms = id.split(None, 1)
                    if terms:
                        id = terms[0]
                        mol.setTitle(id)
                    else:
                        id = None
                else:
                    # " " or "\t"
                    terms = id.split(delimiter_name, 1)
                    if terms:
                        id = terms[0]
                        mol.setTitle(id)
                    else:
                        id = None
                
            yield id, mol                
                
            mol = None  # reset in case the lower-level handler skips a record/raises an error
    
    finally:
        # Don't save the final molecule.
        location.save(mol=None, record=None, recno=get_recno(), lineno=get_lineno())
        mol = None
        close()


def iter_read_smiles_ids_and_molecules_python(
        infile, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors):
    if builder is None:
        builder = _default_builder
    it = _iter_read_smiles_ids_and_molecules_python(
        infile, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors)
    next(it)
    return it
    
def _iter_read_smiles_ids_and_molecules_python(
        infile, close, builder, has_header, delimiter, error_handler, location,
        encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    iter_reader = iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler, location,
                                                        "SMILES", encoding, encoding_errors)
    try:
        yield  "Ready for processing!"
        for smiles, id in iter_reader:
            smiles_parser = SmilesParser(builder)
            try:
                mol = smiles_parser.parseSmiles(smiles)
            except InvalidSmilesException as err:
                notify_smiles_error(error_handler, smiles, err, location)
                continue
            else:
                if id is not None:
                    mol.setProperty(TITLE, id)
            yield id, mol
            mol = None  # reset in case the lower-level handler skips a record/raises an error
    finally:
        # Don't save the final molecule.
        location.save(mol=None)
        mol = None
        iter_reader.close()

# Copied and modified from _openeye_toolkit.Flavor
class Flavor(object):
    def __init__(self, register_name, canonical_options, flavors):
        self.register_name = register_name
        self.canonical_options = canonical_options
        self.flavors = flavors

    def get_modified_flavor(self, register_name, default=None):
        flavors = self.flavors.copy()
        if default is not None:
            flavors["Default"] = default
        return Flavor(register_name, self.canonical_options, flavors)

    def parse_flavor(self, value_string):
        if not value_string:
            return self.flavors["Default"]
        
        if value_string.isdigit():
            return int(value_string)

        ## # What about:
        ## if value_string[:1] == "+":
        ##     flavor = self.flavors["Default"]
        ##     value_string = value_string[1:]
        ##     if not value_string:  # Allow '+' by itself to mean default
        ##         return flavor
        ## else:
        ##     flavor = 0
        flavor = 0
        value_string = value_string.replace(",", "|") # Allow both ',' and '|' as separators
        err = None
        try:
            for term in value_string.split("|"):
                if term[:1] == "-":
                    term_flavor = self.flavors[term[1:]]
                    flavor &= ~term_flavor
                else:
                    term_flavor = self.flavors[term]
                    flavor |= term_flavor
        except KeyError:
            available_flavors = ", ".join(sorted(self.flavors))
            err = ValueError(
                "CDK %s format does not support the %r flavor option. Available flavors are: %s"
                % (self.register_name, term, available_flavors))
        if err is not None:
            # Bypass exception chaining in a way compatible with Python 2.7
            raise err
            
        return flavor

    def get_flavor(self, args):
        flavor = args["flavor"]
        if isinstance(flavor, integer_types):
            return flavor
        return self.parse_flavor(flavor)


_smiles_output_flavor = Flavor(
    "base_smiles",
    canonical_options = (
        "Canonical",
        "InChILabelling",
        "AtomAtomMap",
        "AtomicMass",
        "UseAromaticSymbols",
        "StereoTetrahedral",
        "StereoCisTrans",
        "StereoExTetrahedral",
        "StereoExCisTrans",
        "AtomicMassStrict",
        "Cx2dCoordinates",
        "Cx3dCoordinates",
        "CxAtomLabel",
        "CxAtomValue",
        "CxRadical",
        "CxMulticenter",
        "CxPolymer",
        "CxFragmentGroup",
        ),

    flavors = {
        "Canonical": SmiFlavor.Canonical,
        "InChILabelling": SmiFlavor.InChILabelling,
        "AtomAtomMap": SmiFlavor.AtomAtomMap,
        "AtomicMass": SmiFlavor.AtomicMass,
        "UseAromaticSymbols": SmiFlavor.UseAromaticSymbols,
        #"SuppressHydrogens": SmiFlavor.SuppressHydrogens, # commented out in cdk
        "StereoTetrahedral": SmiFlavor.StereoTetrahedral,
        "StereoCisTrans": SmiFlavor.StereoCisTrans,
        "StereoExTetrahedral": SmiFlavor.StereoExTetrahedral,
        "StereoExCisTrans": SmiFlavor.StereoExCisTrans,
        "AtomicMassStrict": SmiFlavor.AtomicMassStrict,
        "Stereo": SmiFlavor.Stereo, # alias
        "Cx2dCoordinates": SmiFlavor.Cx2dCoordinates,
        "Cx3dCoordinates": SmiFlavor.Cx3dCoordinates,
        "CxCoordinates": SmiFlavor.CxCoordinates, # alias
        "CxAtomLabel": SmiFlavor.CxAtomLabel,
        "CxAtomValue": SmiFlavor.CxAtomValue,
        "CxRadical": SmiFlavor.CxRadical,
        "CxMulticenter": SmiFlavor.CxMulticenter,
        "CxPolymer": SmiFlavor.CxPolymer,
        "CxFragmentGroup": SmiFlavor.CxFragmentGroup,
        "AtomAtomMapRenumber": SmiFlavor.AtomAtomMapRenumber, # ??? 0x200000 makes it unique
        "CxSmiles": SmiFlavor.CxSmiles, # alias
        "CxSmilesWithCoords": SmiFlavor.CxSmilesWithCoords, # alias
        # Generic -- can't handle with the current system; use 0
        "Unique": SmiFlavor.Unique,   # alias
        "Isomeric": SmiFlavor.Isomeric, # alias
        "Absolute": SmiFlavor.Absolute, # alias
        "UniversalSmiles": SmiFlavor.UniversalSmiles,  # alias

        # Default: must be specified in the modified copies
        })


def parse_implementation(s):
    if s is None or s != "chemfp":
        return "cdk"
    return "chemfp"

def _get_errmol_info(mol):
    title = mol.getProperty(TITLE)
    if title is None:
        return ""
    return " (input title=%r)" % (str(title),)

class BaseSmiFormat(CDKFormatConfig):
    write_file_handler = "chemfp"
    read_string_handler = "chemfp_cdk"
    
    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        builder = _default_builder
        has_header = bool(reader_args["has_header"])
        implementation = parse_implementation(reader_args["implementation"])
        if implementation == "cdk":
            # use the native IteratingSMILESReader
            infile, close = open_compressed_input_java(source, compression)
            if has_header:
                # skip the header line
                one_byte_array = jpype.JArray(jpype.JByte)(1)
                while 1:
                    retval = infile.read(one_byte_array)
                    if not retval: # end of file
                        break
                    if one_byte_array[0] == 10: # newline
                        break
                    
            reader = cdk.io.iterator.IteratingSMILESReader(infile, builder)
            iter_reader = iter_read_smiles_ids_and_molecules_cdk(
                reader, close, builder, has_header, delimiter,
                error_handler, location, encoding, encoding_errors)
        else:
            infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)
            if has_header:
                # skip the header line
                infile.readline()
            iter_reader = iter_read_smiles_ids_and_molecules_python(
                infile, close, builder, has_header, delimiter,
                error_handler, location, encoding, encoding_errors)
        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format,
                                               {"delimiter": delimiter.name,
                                                #"builder": builder,
                                                "has_header": has_header,
                                                "implementation": implementation})

        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)


    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        split_two_terms_bytes = delimiter.split_two_terms_bytes
        split_two_terms_string = delimiter.split_two_terms_string
        #builder = get_builder(reader_args["builder"])
        smiles_parser = SmilesParser(_default_builder)

        def cdk_smiles_parser(content):
            if isinstance(content, unicode_type):
                first_line, _, _ = content.partition("\n") # get the first record from the input
                terms = split_two_terms_string(first_line.rstrip("\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None

                smiles, id = terms
                try:
                    mol = smiles_parser.parseSmiles(smiles)
                except InvalidSmilesException as err:
                    notify_smiles_error(error_handler, smiles, err)
                    return None, None
                else:
                    if id is not None:
                        mol.setProperty(TITLE, id)
                return id, mol
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = split_two_terms_bytes(first_line.rstrip(b"\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid SMILES input")
                    return None, None

                smiles_bytes, id_bytes = terms
                try:
                    mol = smiles_parser.parseSmiles(smiles_bytes)
                except InvalidSmilesException as err:
                    notify_smiles_error(error_handler, smiles_bytes, err)
                    return None, None
                else:
                    if id_bytes is not None:
                        try:
                            id = id_bytes.decode("utf8")
                        except UnicodeDecodeError as err:
                            error_handler.error("Cannot decode the identifier %r: %s"
                                                % (id_bytes, err))
                            return None, None
                        mol.setProperty(TITLE, id)
                    else:
                        id = None
                return id, mol

        cdk_smiles_parser.__name__ = "cdk_" + self.register_name + "_parse_id_and_molecule"
        return cdk_smiles_parser
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args, "whitespace")
        smiles_parser = SmilesParser(_default_builder)

        if isinstance(content, unicode_type):
            first_line, _, _ = content.partition("\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip("\r"))

            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None

            smiles, id = terms
            try:
                mol = smiles_parser.parseSmiles(smiles)
            except InvalidSmilesException as err:
                notify_smiles_error(error_handler, smiles, err)
                return None, None
            
            if id is not None:
                mol.setProperty(TITLE, id)
            return id, mol
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))

            if terms is None:
                error_handler.error("The empty string is not a valid SMILES input")
                return None, None

            smiles_bytes, id_bytes = terms
            try:
                mol = smiles_parser.parseSmiles(smiles_bytes)
            except InvalidSmilesException as err:
                notify_smiles_error(error_handler, smiles_bytes, err)
                return None, None
            
            if id_bytes is not None:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id_bytes, err))
                    return None, None
                mol.setProperty(TITLE, id)
            else:
                id = None
            return id, mol
            


    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        flavor = self.output.flavor.get_flavor(writer_args)
        
        new_writer_args = {
            "delimiter": delimiter.name,
            "flavor": flavor,
            }
        smiles_generator = SmilesGenerator(flavor)
        #         if (useAromaticityFlag.isSet()) sg = sg.aromatic(); ???
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, smiles_generator),
            location)
            

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        flavor = self.output.flavor.get_flavor(writer_args)
        new_writer_args = {
            "delimiter": delimiter.name,
            "flavor": flavor,
        }
        outfile, close, getvalue = io._open_output_string()

        smiles_generator = SmilesGenerator(flavor)
        #         if (useAromaticityFlag.isSet()) sg = sg.aromatic(); ???
        metadata = base_toolkit.FormatMetadata(_or_string(location),
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, smiles_generator),
            getvalue, location)
            

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        flavor = self.output.flavor.get_flavor(writer_args)
        new_writer_args = {
            "delimiter": delimiter.name,
            "flavor": flavor,
        }
        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)

        smiles_generator = SmilesGenerator(flavor)
        #         if (useAromaticityFlag.isSet()) sg = sg.aromatic(); ???
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_smi_structures(outfile, close, error_handler, location, delimiter, smiles_generator),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, writer_args, "whitespace")
        flavor = self.output.flavor.get_flavor(writer_args)
        smiles_generator = SmilesGenerator(flavor)
        try:
            smiles = str(smiles_generator.create(mol))
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error("CDK cannot create SMILES string%s: %s" % (
                _get_errmol_info(mol), _get_error_message(err)))
            return None

        if id is None:
            id = mol.getProperty(TITLE)
            if id is None:
                id = u""

        return delimiter.get_format_string() % (smiles, id)
    
    def create_bytes(self, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8")

# The three categories follow OEChem 2.0's lead:
#   "smi" = isomeric, canonical (alternate is "ism")
#   "can" = non-isomeric, canonical
#   "usm" = non-isomeric, non-canonical

_parse_delimiter = base_toolkit.DelimiterParser("whitespace")

class SmiFormat(BaseSmiFormat):
    register_name = "smi"
    alternate_extensions = ("ism", "isosmi")
    is_available = True
    record_format = "smi"

    class input(object):
        arg_info = (
            ("has_header", base_toolkit.parse_boolean),
            ("delimiter", _parse_delimiter),
            ("implementation", parse_implementation),
            ("kekulise", base_toolkit.parse_boolean),
            )
        defaults = {
            "has_header": False,
            "delimiter": "to-eol",
            "implementation": None,
            "kekulise": True,
            }
    
    class output(object):
        flavor = _smiles_output_flavor.get_modified_flavor(
            # CDK's default is not canonical.
            # chemfp's "smi" format is canonical
            # CDK's default includes CxSmiles, but chemfp requires that you enable CX SMILES output.
            # (I may add 'cxsmi' in the future.)
            "smi",
            #default = SmiFlavor.Canonical | SmiFlavor.Isomeric | SmiFlavor.Stereo | SmiFlavor.AtomicMass | SmiFlavor.CxSmiles)
            default = SmiFlavor.Canonical | SmiFlavor.Isomeric | SmiFlavor.Stereo | SmiFlavor.AtomicMass)
        
        arg_info = (
            ("delimiter", _parse_delimiter),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "delimiter": None,
            "flavor": None,
            }

# canonical, non-isomeric
class CanFormat(BaseSmiFormat):
    register_name = record_format = "can"
    is_available = True

    input = SmiFormat.input
    
    class output(object):
        flavor = _smiles_output_flavor.get_modified_flavor(
            "can",
            default = SmiFlavor.Canonical)
        
        arg_info = (
            ("delimiter", _parse_delimiter),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "delimiter": "to-eol",
            "flavor": None,
            }

# non-canonical, non-isomeric SMILES
class UsmFormat(BaseSmiFormat):
    register_name = record_format = "usm"
    is_available = True

    input = SmiFormat.input
    
    class output(object):
        flavor = _smiles_output_flavor.get_modified_flavor(
            "usm",
            default = 0)
        
        arg_info = (
            ("delimiter", _parse_delimiter),
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "delimiter": "to-eol",
            "flavor": None,
            }


class SmilesStringFormat(base_toolkit.NotIoFormatConfigMixin, CDKFormatConfig):
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        smiles_parser = SmilesParser(_default_builder)
        if not reader_args["kekulise"]:
            smiles_parser.kekulise(False)
        
        def cdk_parser(content):
            # Follow the Daylight spec, which says that the SMILES ends with any whitespace
            terms = content.split(None, 1)
            if not terms:
                # In chemfp, the empty SMILES string returns an empty molecule.
                # CDK does the same thing, so give it the empty string
                # TODO: create the AtomContainer directly
                return None, smiles_parser.parseSmiles("")

            try:
                mol = smiles_parser.parseSmiles(terms[0])
            except InvalidSmilesException as err:
                notify_smiles_error(error_handler, smiles, err, location)

            # This is only supposed to have a SMILES. Ignore any possible id.
            return None, mol

        cdk_parser.__name__ = "cdk_" + self.register_name + "_parse_id_and_molecule"
        return cdk_parser
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Follow the Daylight spec, which says that the SMILES ends with any whitespace
        terms = content.split(None, 1)
        
        smiles_parser = SmilesParser(_default_builder)        
        if not terms:
            # In chemfp, the empty SMILES string returns an empty molecule.
            # CDK does the same thing, so give it the emtpy string
            # TODO: create the AtomContainer directly
            return None, smiles_parser.parseSmiles("")

        if not reader_args["kekulise"]:
            smiles_parser.kekulise(False)
        
        try:
            mol = smiles_parser.parseSmiles(terms[0])
        except InvalidSmilesException as err:
            notify_smiles_error(error_handler, terms[0], err)
            return None, None
        # This is only supposed to have a SMILES. Ignore any possible id.
        return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        flavor = self.output.flavor.get_flavor(writer_args)
        smiles_generator = SmilesGenerator(flavor)
        try:
            return str(smiles_generator.create(mol))
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error("CDK cannot create SMILES string%s: %s" % (
                _get_errmol_info(mol), _get_error_message(err)))
            return None


    def create_bytes(self, mol, id, writer_args, error_handler):
        return self.create_string(mol, id, writer_args, error_handler).encode("utf8")


class SmiStringFormat(SmilesStringFormat):
    register_name = record_format = "smistring"

    class input(object):
        arg_info = (
            ("kekulise", base_toolkit.parse_boolean),
            )
        defaults = {
            "kekulise": True,
            }

    class output(object):
        flavor = SmiFormat.output.flavor.get_modified_flavor("smistring")
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }
        

class CanStringFormat(SmilesStringFormat):
    register_name = record_format = "canstring"
    
    input = SmiStringFormat.input

    class output(object):
        flavor = CanFormat.output.flavor.get_modified_flavor("canstring")
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }
            

class UsmStringFormat(SmilesStringFormat):
    register_name = record_format = "usmstring"
    
    input = SmiStringFormat.input

    class output(object):
        flavor = UsmFormat.output.flavor.get_modified_flavor("usmstring")
        arg_info = (
            ("flavor", flavor.parse_flavor),
            )
        defaults = {
            "flavor": None,
            }


def gen_write_smi_structures(outfile, close, error_handler, location, delimiter, smiles_generator):
    def get_smiles_java_str(mol):
        return smiles_generator.create(mol)
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          True, "SMILES", get_smiles_java_str)
    next(gen)
    return gen

def gen_write_inchi_structures(outfile, close, error_handler, location, delimiter, include_id, inchi_options_str):
    def get_inchi_java_str(mol):
        return _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchi()
        
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, "InChI", get_inchi_java_str)
    next(gen)
    return gen

def gen_write_inchikey_structures(outfile, close, error_handler, location, delimiter, include_id, inchi_options_str):
    def get_inchikey_java_str(mol):
        return _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchiKey()
        
    gen = _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                          include_id, "InChIKey", get_inchikey_java_str)
    next(gen)
    return gen

def _gen_write_delimited_structures(outfile, close, error_handler, location, delimiter,
                                    include_id, format_label, get_java_str):
    recno = output_recno = 0
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno

    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno)

    fmt = delimiter.get_format_string()
    
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        try:
                            java_str = get_java_str(mol)
                        except (cdk.exception.CDKException, java.lang.Exception) as err:
                            error_handler.error(
                                "CDK cannot create the %s string%s: %s" % (
                                    format_label, _get_errmol_info(mol), _get_error_message(err)),
                                location)
                            continue
                        if java_str is None:
                            error_handler.error(
                                "CDK cannot create the %s string%s" %
                                (format_label, _get_errmol_info(mol)),
                                location)
                            continue
                        
                        s = str(java_str)
                        if include_id:
                            id = mol.getProperty(TITLE) # getTitle() ?
                            if id is None:
                                id = ""
                                    
                            if not outfile.write_string(fmt % (s, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(s + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        try:
                            java_str = get_java_str(mol)
                        except (cdk.exception.CDKException, java.lang.Exception) as err:
                            error_handler.error(
                                "CDK cannot create the %s string%s: %s" % (
                                    format_label, _get_errmol_info(mol), _get_error_message(err)),
                                location)
                            continue
                        if java_str is None:
                            error_handler.error("CDK cannot create the %s string%s" % (
                                format_label, _get_errmol_info(mol)),
                                                location)
                            continue
                        s = str(java_str)
                        if include_id:
                            if id is None:
                                id = mol.getProperty(TITLE) # or getTitle()
                                if id is None:
                                    id = ""
                                else:
                                    id = str(id)
                            if not outfile.write_string(fmt % (s, id), error_handler, location):
                                continue
                        else:
                            if not outfile.write_string(s + "\n", error_handler, location):
                                continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    try:
                        java_str = get_java_str(mol)
                    except (cdk.exception.CDKException, java.lang.Exception) as err:
                        error_handler.error(
                            "CDK cannot create the %s string%s: %s" % (
                            format_label, _get_errmol_info(mol), _get_error_message(err)),
                            location)
                        continue

                    if java_str is None:
                        error_handler.error("CDK cannot create the %s string" % (format_label,),
                                            location)
                        continue
                    s = str(java_str)
                    if include_id:
                        if id is None:
                            id = mol.getProperty(TITLE)
                            if id is None:
                                id = ""
                            else:
                                id = str(id)

                        if not outfile.write_string(fmt % (s, id), error_handler, location):
                            continue
                    else:
                        if not outfile.write_string(s + "\n", error_handler, location):
                            continue
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                        
    finally:
        if close is not None:
            close()
        location.save(output_recno = get_output_recno(),
                      recno = get_recno())
            

def iter_read_inchi_structures(infile, close, has_header, delimiter, builder, error_handler, location,
                               encoding, encoding_errors):
    x = _iter_read_inchi_structures(infile, close, has_header, delimiter, builder, error_handler, location,
                                    encoding, encoding_errors)
    next(x)
    return x

def _iter_read_inchi_structures(infile, close, has_header, delimiter, builder, error_handler, location,
                                encoding, encoding_errors):
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)
    delimited_field_reader = iter_read_delimited_fields_as_strings(infile, close, has_header, delimiter, error_handler,
                                                                   location, "InChI", encoding, encoding_errors)
    try:
        yield "Ready for processing!"
        for inchi, id in delimited_field_reader:
            inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi, builder)
            if inchi_to_struct is None:
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,), location)
                continue
            mol = inchi_to_struct.getAtomContainer()
            if not mol.getAtomCount():
                # An InChI string must have at leat one atom.
                # Invalid InChIs can get here.
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,), location)
                continue
            if id is not None:
                mol.setProperty(TITLE, id)
            yield id, mol
            mol = None # reset in case the lower-level handler skips a record/raises an error
    finally:
        location.save(mol=None)
        mol = None
        delimited_field_reader.close()
    

# Format is:
#    <inchi> <delimiter> <id>

def parse_nonnegative_int(value):
    if not value.isdigit():
        raise ValueError("must be non-negative integer")
    return int(value)


class InChIGeneratorOptions(object):
    arg_info = (
        # Not enabling the "ENG_OPTIONS"
        ("NOLABELS", base_toolkit.parse_boolean),
        ("AuxNone", base_toolkit.parse_boolean),
        ("MISMATCHISERROR", base_toolkit.parse_boolean),
        ("SNON", base_toolkit.parse_boolean),
        ("NEWPSOFF", base_toolkit.parse_boolean),
        ("DONOTADDH", base_toolkit.parse_boolean),

        # "non-standard InChI perception options"
        ("SREL", base_toolkit.parse_boolean),
        ("SRAC", base_toolkit.parse_boolean),
        ("SUCF", base_toolkit.parse_boolean),
        ("ChiralFlagON", base_toolkit.parse_boolean),
        ("ChiralFlagOFF", base_toolkit.parse_boolean),
        ("SUU", base_toolkit.parse_boolean),
        ("SLUUD", base_toolkit.parse_boolean),
        ("FIXEDH", base_toolkit.parse_boolean),
        ("RECMET", base_toolkit.parse_boolean),
        ("KET", base_toolkit.parse_boolean),
        ("_15T", base_toolkit.parse_boolean),
        ("NoADP", base_toolkit.parse_boolean),
        ("DISCONSALT", parse_nonnegative_int), # takes an integer, either 0 or non-zero
        ("RECONMETAL", parse_nonnegative_int), # takes an integer, either 0 or non-zero
        ("DISCONMETALCHKVAL", parse_nonnegative_int), # takes an integer, either 0 or non-zero
        ("MOVEPOS", parse_nonnegative_int),
        ("MERGESALTTG", parse_nonnegative_int),
        ("UNCHARGEDACIDS", parse_nonnegative_int),
        ("ACIDTAUT", parse_nonnegative_int),  # code says it's 0, 1, or 2
        ("options", str),  # space separated strings
        )
    defaults = dict.fromkeys((x[0] for x in arg_info), None) # all empty
    _get_flags = {}
    for k, v in arg_info:
        if v is base_toolkit.parse_boolean:
            _get_flags[k] = 0
        elif v is parse_nonnegative_int:
            _get_flags[k] = 1
        elif v is str:
            _get_flags[k] = 9
        else:
            raise AssertionError((k, v))
    del k, v
    @staticmethod
    def get_options_and_kwargs(writer_args):
        options = []
        kwargs = {}
        extra_options = []
        _get_flags = InChIGeneratorOptions._get_flags
        for k, v in writer_args.items():
            if v is None: # unknown or use default 
                continue
            flg = _get_flags.get(k, None)
            if flg is None: # unknown
                continue
            if flg == 0:
                if v:
                    options.append(k)
                    kwargs[k] = True
                else:
                    kwargs[k] = False
            elif flg == 1:
                if isinstance(v, int) and (0 <= v):
                    options.append("%s:%d" % (k, v))
                    kwargs[k] = v
                else:
                    raise ValueError("%s must be a non-negative integer: %r" % (k, v))
            elif flg == 9:
                extra_options = v.split()
                kwargs[k] = v
            else:
                raise AssertionError(flg)
        return " ".join(options + extra_options), kwargs

_inchi_input_arg_info = ()
_inchi_input_defaults = {
    "delimiter": "to-eol",
    }

class InChIFormat(CDKFormatConfig):
    register_name = record_format = "inchi"
    is_available = has_inchi
    read_string_handler = "chemfp"
    write_file_handler = "chemfp"

    class input(object):
        arg_info = _inchi_input_arg_info + (
            ("delimiter", _parse_delimiter),
            )
        defaults = _inchi_input_defaults.copy()
        
    class output(object):
        arg_info = InChIGeneratorOptions.arg_info + (
            ("delimiter", _parse_delimiter),
            ("include_id", base_toolkit.parse_boolean),
            )
        defaults = InChIGeneratorOptions.defaults.copy()
        defaults["delimiter"] = None
        defaults["include_id"] = True

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        new_reader_args = {
            "delimiter": delimiter.name,
            }
        builder = _default_builder
            
        infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)

        iter_reader = iter_read_inchi_structures(infile, close, False, delimiter, builder,
                                                 error_handler, location, encoding, encoding_errors)

        metadata = base_toolkit.FormatMetadata(location.filename, self.record_format, new_reader_args)
        
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        split_two_terms_bytes = delimiter.split_two_terms_bytes
        split_two_terms_string = delimiter.split_two_terms_string
        builder = _default_builder

        def cdk_inchi_parse_id_and_molecule(content):
            if isinstance(content, unicode_type):
                first_line, _, _ = content.partition("\n") # get the first record from the input
                terms = split_two_terms_string(first_line.rstrip("\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input")
                    return None, None

                inchi, id = terms
                inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi, builder)
                if inchi_to_struct is None:
                    error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                    return None, None
                mol = inchi_to_struct.getAtomContainer()
                if not mol.getAtomCount():
                    # An InChI string must have at leat one atom.
                    # Invalid InChIs can get here.
                    error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                    return None, None
                if id is not None:
                    mol.setProperty(TITLE, id)
                return id, mol
            else:
                first_line, _, _ = content.partition(b"\n") # get the first record from the input
                terms = split_two_terms_bytes(first_line.rstrip(b"\r"))
            
                if terms is None:
                    error_handler.error("The empty string is not a valid InChI input")
                    return None, None

                inchi_bytes, id_bytes = terms
                inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi_bytes, builder)
                if inchi_to_struct is None:
                    error_handler.error("CDK cannot parse the InChI %r" % (inchi_bytes,))
                    return None, None
                mol = inchi_to_struct.getAtomContainer()
                if not mol.getAtomCount():
                    # An InChI string must have at leat one atom.
                    # Invalid InChIs can get here.
                    error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                    return None, None
                
                if id_bytes is None:
                    id = None
                else:
                    try:
                        id = id_bytes.decode("utf8")
                    except UnicodeDecodeError as err:
                        error_handler.error("Cannot decode the identifier %r: %s"
                                            % (id_bytes, err))
                        return None, None
                    mol.setProperty(TITLE, id)
                return id, mol
            
        return cdk_inchi_parse_id_and_molecule
        

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        delimiter = base_toolkit.get_delimiter(self, reader_args)
        if isinstance(content, unicode_type):
            first_line, _, _ = content.partition("\n") # get the first record from the input
            terms = delimiter.split_two_terms_string(first_line.rstrip("\r"))
            
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi, id = terms
            inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi, _default_builder)
            if inchi_to_struct is None:
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                return None, None
            mol = inchi_to_struct.getAtomContainer()
            if not mol.getAtomCount():
                # An InChI string must have at leat one atom.
                # Invalid InChIs can get here.
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                return None, None

            if id is not None:
                mol.setProperty(TITLE, id)
                new_id = mol.getProperty(TITLE)
            return id, mol
        else:
            first_line, _, _ = content.partition(b"\n") # get the first record from the input
            terms = delimiter.split_two_terms_bytes(first_line.rstrip(b"\r"))
            
            if terms is None:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            
            inchi_bytes, id_bytes = terms
            inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi_bytes, _default_builder)
            if inchi_to_struct is None:
                error_handler.error("CDK cannot parse the InChI %r" % (inchi_bytes,))
                return None, None

            mol = inchi_to_struct.getAtomContainer()
            if not mol.getAtomCount():
                # An InChI string must have at leat one atom.
                # Invalid InChIs can get here.
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                return None, None

            if id_bytes is not None:
                try:
                    id = id_bytes.decode("utf8")
                except UnicodeDecodeError as err:
                    error_handler.error("Cannot decode the identifier %r: %s"
                                        % (id_bytes, err))
                    return None, None
                mol.setProperty(TITLE, id)

            return id, mol
    
    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_options_str),
            location)
    

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)

        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_output_string()
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_options_str),
            getvalue,
            location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)

        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchi_structures(outfile, close, error_handler, location, delimiter,
                                       include_id, inchi_options_str),
            getvalue,
            location)
    

    def create_string(self, mol, id, writer_args, error_handler):
        inchi_options_str, inchi_kwargs = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        try:
            java_str = _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchi()
        except java.lang.IllegalArgumentException:
            error_handler.error("CDK cannot create the InChI string")
            return None
        
        if java_str is None:
            error_handler.error("CDK cannot create the InChI string")
            return None

        inchi = str(java_str)
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            if id is None:
                id = mol.getProperty(TITLE)
                if id is None:
                    id = u""
            return delimiter.get_format_string() % (inchi, id)
        else:
            return inchi + u"\n"


class InChIKeyFormat(CDKFormatConfig):
    register_name = record_format = "inchikey"
    is_available = has_inchi
    read_string_handler = "chemfp"
    write_file_handler = "chemfp"

    input = None

    output = InChIFormat.output

    def open_molecule_writer(self, destination, outfile, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_inchikey_structures(
                outfile, close, error_handler, location, delimiter,
                include_id, inchi_options_str),
            location)
    

    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)

        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_output_string()
        metadata = base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(
                outfile, close, error_handler, location, delimiter,
                include_id, inchi_options_str),
            getvalue,
            location)

    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        delimiter = base_toolkit.get_delimiter(self, writer_args)
        inchi_options_str, new_writer_args = InChIGeneratorOptions.get_options_and_kwargs(writer_args)

        new_writer_args["delimiter"] = delimiter.name
        new_writer_args["include_id"] = include_id = bool(writer_args["include_id"])

        outfile, close, getvalue = io._open_compressed_output_bytes(compression, EXTERNAL_NAME, level)
        
        metadata = base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args)
        return base_toolkit.MoleculeStringWriter(
            metadata,
            gen_write_inchikey_structures(
                outfile, close, error_handler, location, delimiter,
                include_id, inchi_options_str),
            getvalue,
            location)
    

    def create_string(self, mol, id, writer_args, error_handler):
        inchi_options_str, inchi_kwargs = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        try:
            java_str = _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchiKey()
        except java.lang.IllegalArgumentException:
            error_handler.error("CDK cannot create the InChIKey string")
            return None
        
        if java_str is None:
            error_handler.error("CDK cannot create the InChIKey string")
            return None

        inchikey = str(java_str)
        if writer_args["include_id"]:
            delimiter = base_toolkit.get_delimiter(self, writer_args)
            if id is None:
                id = mol.getProperty(TITLE)
                if id is None:
                    id = u""
            return delimiter.get_format_string() % (inchikey, id)
        else:
            return inchikey + u"\n"

def iter_read_structures_cdk(reader, close, id_tag, error_handler, location,
                                 encoding, encoding_errors):
    it = _iter_read_structures_cdk(reader, close, id_tag, error_handler, location,
                                       encoding, encoding_errors)
    next(it)
    return it

def _iter_read_structures_cdk(reader, close, id_tag, error_handler, location,
                                  encoding, encoding_errors):
    # Register a way to get the current molecule
    recno = 0
    mol = None
    def get_recno():
        return recno
    def get_mol():
        return mol
    location.register(
        get_mol=get_mol,
        get_recno=get_recno,
        )
    if id_tag is None:
        id_tag = TITLE
    
    try:
        yield "Ready for processing!"
        while reader.hasNext():
            mol = reader.next()
            id = mol.getProperty(id_tag)
            if id is not None:
                id = str(id)
            recno += 1
            yield id, mol
            mol = None  # reset in case the lower-level handler skips a record/raises an error
    finally:
        location.save(mol=None, recno=get_recno())
        mol = None

def iter_read_sdf_structures_chemfp(infile, close, reader_factory, id_tag,
                                    error_handler, location, encoding, encoding_errors):
    if id_tag is None:
        id_tag_bytes = None
    else:
        id_tag_bytes = id_tag.encode(encoding)
    it = _iter_read_sdf_structures_chemfp(infile, close, reader_factory, id_tag_bytes,
                                            error_handler, location, encoding, encoding_errors)
                                          
    next(it)
    return it

def _iter_read_sdf_structures_chemfp(infile, close, reader_factory, id_tag_bytes,
                                     error_handler, location, encoding, encoding_errors):
    # Register a way to get the current molecule
    mol = None
    def get_mol():
        return mol
    location.register(get_mol=get_mol)

    record_reader = _text_toolkit.iter_sdf_ids_and_records(infile, close, id_tag_bytes, error_handler, location,
                                                           encoding, encoding_errors)
    try:
        yield "Ready for processing!"
        for (id, record) in record_reader:
            try:
                mol = reader_factory.parse_record(record, _default_builder)
            except (cdk.exception.CDKException, java.lang.Exception) as err:
                mol = None
                error_handler.error("CDK cannot parse the sdf record: %s" % (
                    _get_error_message(err),), location)
                continue
            mol.setTitle(id)
            yield id, mol
            mol = None
            
    finally:
        location.save(mol=None)
        mol = None
        

class InChIStringFormat(base_toolkit.NotIoFormatConfigMixin, CDKFormatConfig):
    register_name = record_format = "inchistring"
    is_available = has_inchi

    class input(object):
        arg_info = _inchi_input_arg_info
        defaults = _inchi_input_defaults.copy()
        
    class output(object):
        arg_info = InChIGeneratorOptions.arg_info
        defaults = InChIGeneratorOptions.defaults.copy()

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        builder = _default_builder
        def cdk_inchistring_parse_id_and_molecule(content):
            # Follow the SMILES lead and split on the first whitespace. (Is that right?)
            terms = content.split()
            if not terms:
                error_handler.error("The empty string is not a valid InChI input")
                return None, None
            inchi = terms[0]
            inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi, builder)
            if inchi_to_struct is None:
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                mol = None
            else:
                mol = inchi_to_struct.getAtomContainer()

                if not mol.getAtomCount():
                    error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                    return None, None
            
            # This is only supposed to have an InChI string. Ignore any possible id.
            return None, mol

        return cdk_inchistring_parse_id_and_molecule
    
    
    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        # Follow the SMILES lead and split on the first whitespace. (Is that right?)
        terms = content.split()
        if not terms:
            error_handler.error("The empty string is not a valid InChI input")
            return None, None
        inchi = terms[0]
        inchi_to_struct = _inchi_generator_factory.getInChIToStructure(inchi, _default_builder)
        if inchi_to_struct is None:
            error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
            mol = None
        else:
            mol = inchi_to_struct.getAtomContainer()

            if not mol.getAtomCount():
                error_handler.error("CDK cannot parse the InChI %r" % (inchi,))
                return None, None
            
        # This is only supposed to have an InChI string. Ignore any possible id.
        return None, mol

    def create_string(self, mol, id, writer_args, error_handler):
        inchi_options_str, inchi_kwargs = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        try:
            inchi = _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchi()
        except java.lang.IllegalArgumentException:
            error_handler.error("CDK cannot create the InChI string")
            return None
        
        if inchi is None:
            error_handler.error("CDK cannot create the InChI string")
            return None
        
        return str(inchi)

class InChIKeyStringFormat(base_toolkit.NotIoFormatConfigMixin, InChIKeyFormat):
    register_name = format_name = "inchikeystring"
    is_available = has_inchi    

    input = None
        
    output = InChIStringFormat.output

    def create_string(self, mol, id, writer_args, error_handler):
        inchi_options_str, inchi_kwargs = InChIGeneratorOptions.get_options_and_kwargs(writer_args)
        try:
            inchikey = _inchi_generator_factory.getInChIGenerator(mol, inchi_options_str).getInchiKey()
        except java.lang.IllegalArgumentException:
            error_handler.error("CDK cannot create the InChIKey string")
            return None
        
        if inchikey is None:
            error_handler.error("CDK cannot create the InChIKey string")
            return None
        
        return str(inchikey)

        
def parse_mode(s):
    if s is None:
        return "RELAXED"
    if s in ("RELAXED", "relaxed"):
        return "RELAXED"
    if s in ("STRICT", "strict"):
        return "STRICT"
    raise ValueError("must be one of 'RELAXED' or 'STRICT' or None: %r" % (s,))

def parse_program_name(s):
    if isinstance(s, bytes_type) or isinstance(s, unicode_type):
        return s
    raise ValueError("Must be a string")

# Figure out the SDF input and output settings

def _set_settings_options(settings_hashmap, python_class):
    setting_iterator = settings_hashmap.iterator()
    arg_info = []
    defaults = {}

    while setting_iterator.hasNext():
        setting = setting_iterator.next()
        name = str(setting.getName())
        if isinstance(setting, cdk.io.setting.BooleanIOSetting):
            arg_info.append( (name, base_toolkit.parse_boolean) )
            default = str(setting.getDefaultSetting())
            if default == "false":
                defaults[name] = False
                
            elif default == "true":
                defaults[name] = True
            else:
                raise AssertionError((name, setting, default))
        elif isinstance(setting, cdk.io.setting.StringIOSetting):
            arg_info.append( (name, str) )
            defaults[name] = setting.getDefaultSetting()
        else:
            raise AssertionError( ("What is ", name, setting) )

    arg_info.sort()
    python_class.arg_info = tuple(arg_info)
    python_class.defaults = defaults


def _set_io_settings(io_obj, kwargs, defaults):
    new_kwargs = {}
    for k, v in kwargs.items():
        if v is None:
            continue # use the default
        default = defaults[k]
        if default is True:
            if v:
                new_kwargs[k] = True
                continue
            io_obj.getSetting(k).setSetting("false")
            new_kwargs[k] = False
        elif default is False:
            if not v:
                new_kwargs[k] = False
                continue
            io_obj.getSetting(k).setSetting("true")
            new_kwargs[k] = True
        else:
            if v != default:
                io_obj.getSetting(k).setSetting(v)
            new_kwargs[k] = v
    return new_kwargs
    
    
def _add_mode(python_class):
    python_class.arg_info += (("mode", str),)
    python_class.defaults["mode"] = "RELAXED"
    
class MDLReaderOptions(object):
    arg_info = defaults = None

_set_settings_options(cdk.io.MDLReader().getSettings(), MDLReaderOptions)
_add_mode(MDLReaderOptions)

class MDLV2000ReaderOptions(object):
    arg_info = defaults = None

_set_settings_options(cdk.io.MDLV2000Reader().getSettings(), MDLV2000ReaderOptions)
_add_mode(MDLV2000ReaderOptions)

class MDLV3000ReaderOptions(object):
    arg_info = defaults = None

_set_settings_options(cdk.io.MDLV3000Reader().getSettings(), MDLV3000ReaderOptions)
_add_mode(MDLV3000ReaderOptions)

_mdl_version_pat_unicode = re.compile(u"""
   [^\\n]*\\n  # skip three lines
   [^\\n]*\\n
   [^\\n]*\\n
   [^\\n]{34}[Vv](2000|3000)  # look for the V2000 or V3000
""", re.MULTILINE | re.X)

_mdl_version_pat_byte = re.compile(b"""
   [^\\n]*\\n  # skip three lines
   [^\\n]*\\n
   [^\\n]*\\n
   [^\\n]{34}[Vv](2000|3000)  # look for the V2000 or V3000
""", re.MULTILINE | re.X)

def get_mdl_version(record):        
    if isinstance(record, unicode_type):
        m = _mdl_version_pat_unicode.match(record)
        if m is None:
            return 0
        s = m.group(1)
        if s == "2000":
            return 2000
        if s == "3000":
            return 3000
    else:            
        m = _mdl_version_pat_byte.match(record)
        if m is None:
            return 0
        s = m.group(1)
        if s == b"2000":
            return 2000
        if s == b"3000":
            return 3000
    raise AssertionError(("Unsupported", m.group(1), record))


class SDFReaderOptions(object):
    # Merge the three MDL input options
    arg_info = tuple(sorted(dict(
        MDLReaderOptions.arg_info +
        MDLV2000ReaderOptions.arg_info +
        MDLV3000ReaderOptions.arg_info).items()))
    defaults = MDLReaderOptions.defaults.copy()
    defaults.update(MDLV2000ReaderOptions.defaults)
    defaults.update(MDLV3000ReaderOptions.defaults)

def _configure_mdl_reader(mdl_reader, defaults, set_args, mode):
    props = None
    for k, v in set_args.items():
        if k not in defaults:
            continue
        if v is True:
            if props is None:
                props = java.util.Properties()
            props.setProperty(k, "true")
        elif v is False:
            if props is None:
                props = java.util.Properties()
            props.setProperty(k, "false")
        else:
            if props is None:
                props = java.util.Properties()
            props.setProperty(k, v)
            
    if props is not None:
        mdl_reader.addChemObjectIOListener(cdk.io.listener.PropertiesListener(props))
        mdl_reader.customizeJob()
    if mode == "STRICT":
        mdl_reader.setReaderMode(STRICT)
    return mdl_reader
                              
class SDFReaderFactory(object):
    def __init__(self, reader_args):
        self.reader_args = reader_args
        new_reader_args = {}
        set_args = {}
        
        defaults = SDFReaderOptions.defaults
        mode = "RELAXED"
        for k, v in reader_args.items():
            if v is None:
                continue # use the default
            if k == "implementation":
                continue
            default = defaults[k]
            if default is True:
                if v:
                    new_reader_args[k] = True
                    continue
                set_args[k] = False
                new_reader_args[k] = False
            elif default is False:
                if not v:
                    new_reader_args[k] = False
                    continue
                set_args[k] = True
                new_reader_args[k] = True
            elif k == "mode":
                mode = parse_mode(v)
            else:
                if v != default:
                    set_args[k] = v
                new_reader_args[k] = v

        new_reader_args["mode"] = mode
        self.new_reader_args = new_reader_args
        
        self._mode = mode
        self._set_args = set_args

        self.mdlv0_reader = _configure_mdl_reader(
            cdk.io.MDLReader(), MDLReaderOptions.defaults, set_args, mode)
        self.mdlv2000_reader = _configure_mdl_reader(
            cdk.io.MDLV2000Reader(), MDLV2000ReaderOptions.defaults, set_args, mode)
        self.mdlv3000_reader = _configure_mdl_reader(
            cdk.io.MDLV3000Reader(), MDLV3000ReaderOptions.defaults, set_args, mode)

        self._mdl_reader_table = {
               0: self.mdlv0_reader,  # XXX this doesn't seem usable. Remove and use v2000?
            2000: self.mdlv0_reader,
            2000: self.mdlv2000_reader,
            3000: self.mdlv3000_reader,
            }
    
    def get_reader(self, record):
        "Get the appropriate reader for the given record version"
        return self._mdl_reader_table[get_mdl_version(record)]

    def parse_record(self, record, builder):
        reader = self._mdl_reader_table[get_mdl_version(record)]
        reader.setReader(java.io.StringReader(record))
        reader = self.get_reader(record)
        return reader.read(builder.newAtomContainer())

    def get_iterating_reader(self, infile, builder):
        sdf_reader = cdk.io.iterator.IteratingSDFReader(infile, builder)
        sdf_reader.addSettings(self.mdlv0_reader.getSettings())
        sdf_reader.addSettings(self.mdlv2000_reader.getSettings())
        sdf_reader.addSettings(self.mdlv3000_reader.getSettings())
        if self._mode == "STRICT":
            sdf_reader.setReaderMode(STRICT)
        return sdf_reader

class SDFOutputOptions(object):
    arg_info = None
    defaults = None
    
    ## @staticmethod
    ## def set_writer_opts(sdf_writer, writer_args):
    ##     defaults = SDFOutputOptions.defaults
    ##     new_writer_args = {}
    ##     for k, v in writer_args.items():
    ##         if v is None:
    ##             continue # use the default
    ##         default = defaults[k]
    ##         if default is True:
    ##             if v:
    ##                 new_writer_args[k] = True
    ##                 continue
    ##             sdf_writer.getSetting(k).setSetting("false")
    ##             new_writer_args[k] = False
    ##         elif default is False:
    ##             if not v:
    ##                 new_writer_args[k] = False
    ##                 continue
    ##             sdf_writer.getSetting(k).setSetting("true")
    ##             new_writer_args[k] = True
    ##         else:
    ##             if v != default:
    ##                 sdf_writer.getSetting(k).setSetting(v)
    ##             new_writer_args[k] = v
    ##     return new_writer_args
        

_set_settings_options(cdk.io.SDFWriter().getSettings(), SDFOutputOptions)

# Output coroutine for SDF writer

def gen_write_structures_cdk(record_format, cdk_writer, close, error_handler, location):
    it = _gen_write_structures_cdk(record_format, cdk_writer, close, error_handler, location)
    next(it)
    return it

def _gen_write_structures_cdk(record_format, cdk_writer, close, error_handler, location):
    recno = output_recno = 0
    mol = None
    def get_recno():
        return recno
    def get_output_recno():
        return output_recno
    def get_mol():
        return mol
    location.register(get_recno=get_recno,
                      get_output_recno=get_output_recno,
                      get_mol=get_mol)
    try:
        yield_value = None
        while 1:
            command, args = yield yield_value
            yield_value = None
            try:
                if command == "write_molecules":
                    for mol in args:
                        recno += 1
                        try:
                            cdk_writer.write(mol)
                        except (cdk.exception.CDKException, java.lang.Exception) as err:
                            error_handler.error(
                                "CDK cannot create %s record%s: %s" % (
                                    record_format, _get_errmol_info(mol), _get_error_message(err)),
                                location)
                            continue
                        output_recno += 1

                elif command == "write_ids_and_molecules":
                    for id, mol in args:
                        recno += 1
                        try:
                            if id is None:
                                cdk_writer.write(mol)
                            else:
                                old_title = mol.getProperty(TITLE)
                                if old_title == id:
                                    cdk_writer.write(mol)
                                else:
                                    mol.setProperty(TITLE, id)
                                    try:
                                        cdk_writer.write(mol)
                                    finally:
                                        mol.setProperty(TITLE, old_title)
                        except (cdk.exception.CDKException, java.lang.Exception) as err:
                            error_handler.error(
                                "CDK cannot create %s record%s: %s" % (
                                    record_format, _get_errmol_info(mol), _get_error_message(err)),
                                location)
                            continue
                        output_recno += 1

                elif command == "write_molecule":
                    id, mol = args
                    recno += 1
                    try:
                        if id is None:
                            cdk_writer.write(mol)
                        else:
                            old_title = mol.getProperty(TITLE)
                            if old_title == id:
                                cdk_writer.write(mol)
                            else:
                                mol.setProperty(TITLE, id)
                                try:
                                    cdk_writer.write(mol)
                                finally:
                                    mol.setProperty(TITLE, old_title)
                    except (cdk.exception.CDKException, java.lang.Exception) as err:
                        error_handler.error(
                            "CDK cannot create %s record%s: %s" % (
                                record_format, _get_errmol_info(mol), _get_error_message(err)),
                            location)
                        continue
                    
                    output_recno += 1

                else:
                    raise AssertionError("Unknown command: %r" % (command,))
            except Exception as err:
                yield_value = err, sys.exc_info()[2]
                
    finally:
        if close is not None:
            close()
        location.save(recno = get_recno(),
                      output_recno = get_output_recno(),
                      mol = None)
        mol = None



class SDFFormat(CDKFormatConfig):
    register_name = record_format = "sdf"
    alternate_extensions = ("sd", "mdl")
    is_available = True
    read_string_handler = "chemfp_cdk"
    write_file_handler = "cdk"

    class input(object):
        arg_info = SDFReaderOptions.arg_info + (
            ("mode", parse_mode),
            ("implementation", parse_implementation),
            )
        io_defaults = SDFReaderOptions.defaults
        defaults = io_defaults.copy()
        defaults["mode"] = "RELAXED"
        defaults["implementation"] = "cdk"

    output = SDFOutputOptions

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        reader_factory = SDFReaderFactory(reader_args)
        new_reader_args = reader_factory.new_reader_args            
        
        implementation = parse_implementation(reader_args["implementation"])
        if implementation == "cdk":
            infile, close = open_compressed_input_java(source, compression)
            reader = reader_factory.get_iterating_reader(infile, _default_builder)

            iter_reader = iter_read_structures_cdk(reader, close, id_tag, error_handler, location,
                                                   encoding, encoding_errors)
        elif implementation == "chemfp":
            # It looks like this is 50% slower than using native CDK
            infile, close = io.open_compressed_input(source, compression, EXTERNAL_NAME)
            iter_reader = iter_read_sdf_structures_chemfp(
                infile, close, reader_factory, id_tag,
                error_handler, location, encoding, encoding_errors)
        else:
            raise AssertionError(implementation)
        
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with MolfileFormat
        reader_factory = SDFReaderFactory(reader_args)
        if id_tag is None:
            id_tag = TITLE
            
        def cdk_sdf_parse_id_and_molecule(content):
            try:
                mol = reader_factory.parse_record(content, _default_builder)
            except (cdk.exception.CDKException, java.lang.Exception) as err:
                error_handler.error("CDK cannot parse the sdf record: %s" % (_get_error_message(err),))
                return None, None
            if mol is None:
                # This was not a molecule?
                error_handler.error("CDK cannot parse the sdf record")
                return None, None
            id = mol.getProperty(id_tag)
            if id is not None:
                id = str(id)
            return id, mol

        return cdk_sdf_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        reader_factory = SDFReaderFactory(reader_args)
        try:
            mol = reader_factory.parse_record(content, _default_builder)
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error("CDK cannot parse the sdf record: %s" % (_get_error_message(err),))
            return None, None
        if mol is None:
            # This was not a molecule?
            error_handler.error("CDK cannot parse the sdf record")
            return None, None

        if id_tag is None:
            id = mol.getProperty(TITLE)
        else:
            id = mol.getProperty(id_tag)
        if id is not None:
            id = str(id)
        return id, mol

    def open_molecule_writer(self, destination, output_stream, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        sdf_writer = cdk.io.SDFWriter(output_stream)
        close = close_chain(sdf_writer.close, close)
        new_writer_args = _set_io_settings(sdf_writer, writer_args, self.output.defaults)
        new_writer_args
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_structures_cdk(self.record_format, sdf_writer, close, error_handler, location),
            location)
    
    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        output_stream, close, getvalue = open_output_string_java()
        sdf_writer = cdk.io.SDFWriter(output_stream)
        close = close_chain(sdf_writer.close, close)
        new_writer_args = _set_io_settings(sdf_writer, writer_args, self.output.defaults)
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)

        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args),
            gen_write_structures_cdk(self.record_format, sdf_writer, close, error_handler, location),
            getvalue,
            location)
            
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        output_stream, close, getvalue = open_compressed_output_bytes_java(compression, level)
        sdf_writer = cdk.io.SDFWriter(output_stream)
        close = close_chain(sdf_writer.close, close)
        new_writer_args = _set_io_settings(sdf_writer, writer_args, self.output.defaults)
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)

        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args),
            gen_write_structures_cdk(self.record_format, sdf_writer, close, error_handler, location),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        output_stream, close, getvalue = open_output_string_java()
        sdf_writer = cdk.io.SDFWriter(output_stream)
        _set_io_settings(sdf_writer, writer_args, self.output.defaults)
        try:
            if id is None:
                sdf_writer.write(mol)
            else:
                old_id = mol.getProperty(TITLE)
                if old_id == id:
                    sdf_writer.write(mol)
                else:
                    mol.setProperty(TITLE, id)
                    try:
                        sdf_writer.write(mol)
                    finally:
                        mol.setProperty(TITLE, old_id)
                            
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error(
                "CDK cannot create SDF record%s: %s" % (
                    _get_errmol_info(mol), _get_error_message(err)))
            return None
        
        sdf_writer.close()
        return getvalue()



class MolfileFormat(base_toolkit.NotIoFormatConfigMixin, CDKFormatConfig):
    register_name = record_format = "molfile"
    alternate_extensions = ()
    is_available = True

    input = SDFFormat.input
    class output:
        arg_info = [x for x in SDFOutputOptions.arg_info if x[0] != "writeProperties"]
        defaults = SDFOutputOptions.defaults.copy()
        del defaults["writeProperties"]
    
    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        # Note: keep this synchronized with SDFFormat
        reader_factory = SDFReaderFactory(reader_args)
        if id_tag is None:
            id_tag = TITLE
            
        def cdk_molfile_parse_id_and_molecule(content):
            try:
                mol = reader_factory.parse_record(content, _default_builder)
            except (cdk.exception.CDKException, java.lang.Exception) as err:
                error_handler.error("CDK cannot parse the molfile record: %s" % (
                    _get_error_message(err),))
                return None, None
            if mol is None:
                # This was not a molecule?
                error_handler.error("CDK cannot parse the molfile record")
                return None, None
            id = mol.getProperty(id_tag)
            if id is not None:
                id = str(id)
            return id, mol

        return cdk_molfile_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        reader_factory = SDFReaderFactory(reader_args)
        try:
            mol = reader_factory.parse_record(content, _default_builder)
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error("CDK cannot parse the molfile record: %s" % (
                _get_error_message(err),))
            return None, None
        if mol is None:
            # This was not a molecule?
            error_handler.error("CDK cannot parse the molfile record")
            return None, None

        if id_tag is None:
            id = mol.getProperty(TITLE)
        else:
            id = mol.getProperty(id_tag)
        if id is not None:
            id = str(id)
            
        return id, mol
        
    def create_string(self, mol, id, writer_args, error_handler):
        output_stream, close, getvalue = open_output_string_java()
        sdf_writer = cdk.io.SDFWriter(output_stream)
        _set_io_settings(sdf_writer, writer_args, self.output.defaults)
        sdf_writer.getSetting("writeProperties").setSetting("false")
        try:
            if id is None:
                sdf_writer.write(mol)
            else:
                old_id = mol.getProperty(TITLE)
                if old_id == id:
                    sdf_writer.write(mol)
                else:
                    mol.setProperty(TITLE, id)
                    try:
                        sdf_writer.write(mol)
                    finally:
                        mol.setProperty(TITLE, old_id)
            
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error(
                "CDK cannot create molfile record%s: %s" % (
                    _get_errmol_info(mol), _get_error_message(err)))
            return None
        sdf_writer.close()
        s = getvalue()
        if not s.endswith("\n$$$$\n"):
            raise AssertionError(("Expected an SDF record", s))            
        return s[:-5]

#### Single record formats

def iter_read_single_structure_cdk(
            record_format,
            reader, close, id_tag, error_handler, location,
            encoding, encoding_errors):
    it = _iter_read_single_structure_cdk(
        record_format,
        reader, close, id_tag, error_handler, location,
        encoding, encoding_errors)
    next(it)
    return it

def _iter_read_single_structure_cdk(
            record_format,
            reader, close, id_tag, error_handler, location,
            encoding, encoding_errors):
    mol = None
    recno = 0
    def get_mol():
        return mol
    def get_recno():
        return recno
    location.register(
        get_mol=get_mol,
        get_recno=get_recno,
        )
    try:
        yield "ready!"
        try:
            chemfile = reader.read(cdk.ChemFile())
        except java.lang.Exception as err:
            error_handler.error("CDK cannot parse the %s record: %s" % (
                record_format, _get_error_message(err)), location)
            return
        
        containers = _getAllAtomContainers(chemfile)
        if containers.size() == 0:
            # This was not a molecule?
            error_handler.error("CDK cannot parse the %s record" % (record_format,),
                                    location)
            return
        recno += 1
        mol = containers[0]
        id = mol.getProperty(TITLE)
        if id is not None:
            id = str(id)
        yield id, mol
        
    finally:
        mol = None
        location.save(
            recno=get_recno(),
            mol=None)
            

class BaseSingleRecordFormat(CDKFormatConfig):
    ## register_name = record_format = "sdf"
    ## alternate_extensions = ("sd", "mdl")
    ## is_available = True
    read_string_handler = "cdk"
    write_file_handler = "cdk"

    reader_class = None
    writer_class = None
    ## class input(object):
    ##     arg_info = (
    ##         ("mode", parse_mode),
    ##         )
            
    ##     defaults = {
    ##         "mode": "RELAXED",
    ##     }

    ## output = SDFOutputOptions

    def read_ids_and_molecules(self, source, compression, id_tag, reader_args, error_handler, location,
                               encoding, encoding_errors):
        infile, close = open_compressed_input_java(source, compression)
        reader, new_reader_args = self.open_reader_impl(infile, reader_args)

        iter_reader = iter_read_single_structure_cdk(
            self.record_format,
            reader, close, id_tag, error_handler, location,
            encoding, encoding_errors)

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_reader_args)
        return base_toolkit.IdAndMoleculeReader(metadata, iter_reader, location)

    def make_id_and_molecule_parser(self, id_tag, reader_args, error_handler):
        reader, new_reader_args = self.open_reader_impl(None, reader_args)
        if id_tag is None:
            id_tag = TITLE
            
        def cdk_parse_id_and_molecule(content):
            reader.setReader(java.io.StringReader(content))
            try:
                chemfile = reader.read(cdk.ChemFile())
            except java.lang.Exception as err:
                error_handler.error("CDK cannot parse the %s record: %s" % (
                    self.record_format, _get_error_message(err)))
                return None, None
            containers = _getAllAtomContainers(chemfile)
            if containers.size() == 0:
                # This was not a molecule?
                error_handler.error("CDK cannot parse the %s record" % (self.record_format,),
                                        location)
                return None, None
            mol = containers[0]
            id = mol.getProperty(TITLE)
            if id is not None:
                id = str(id)
            return id, mol

        setattr(cdk_parse_id_and_molecule, "__name__",
                    "cdk_" + self.record_format + "_parse_id_and_molecule")
        return cdk_parse_id_and_molecule

    def parse_id_and_molecule(self, content, id_tag, reader_args, error_handler):
        if isinstance(content, str):
            f = java.io.ByteArrayInputStream(content.encode("utf8"))
        else:
            f = java.io.ByteArrayInputStream(content)
        reader, new_reader_args = self.open_reader_impl(f, reader_args)
        try:
            chemfile = reader.read(cdk.ChemFile())
        except java.lang.Exception as err:
            error_handler.error("CDK cannot parse the %s record: %s" % (
                self.record_format, _get_error_message(err)))
            return None, None
            
        containers = _getAllAtomContainers(chemfile)
        if containers.size() == 0:
            mol = reader.next()
            error_handler.error("CDK cannot parse the %s record" % (self.record_format,))
            return None, None
        mol = containers[0]
        if id_tag is None:
            id = mol.getProperty(TITLE)
        else:
            id = mol.getProperty(id_tag)
        if id is not None:
            id = str(id)
        return id, mol

    def open_molecule_writer(self, destination, output_stream, close, writer_args, error_handler, location,
                             encoding, encoding_errors):
        writer, new_writer_args = self.open_writer_impl(output_stream, writer_args)
        close = close_chain(writer.close, close)
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)
        return base_toolkit.MoleculeWriter(
            metadata,
            gen_write_structures_cdk(self.record_format, writer, close, error_handler, location),
            location)
    
    def open_molecule_writer_to_string(self, writer_args, error_handler, location):
        output_stream, close, getvalue = open_output_string_java()
        writer, new_writer_args = self.open_writer_impl(output_stream, writer_args)
        close = close_chain(writer.close, close)

        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)

        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_string(location), self.record_format, new_writer_args),
            gen_write_structures_cdk(self.record_format, writer, close, error_handler, location),
            getvalue,
            location)
            
    def open_molecule_writer_to_bytes(self, compression, writer_args, error_handler, location, level):
        output_stream, close, getvalue = open_compressed_output_bytes_java(compression, level)
        writer, new_writer_args = self.open_writer_impl(output_stream, writer_args)
        close = close_chain(writer.close, close)
        
        new_writer_args = self.output.set_writer_opts(sdf_writer, writer_args)
        metadata = base_toolkit.FormatMetadata(location.filename,
                                               self.record_format, new_writer_args)

        return base_toolkit.MoleculeStringWriter(
            base_toolkit.FormatMetadata(_or_bytes(location), self.record_format, new_writer_args),
            gen_write_structures_cdk(self.record_format, writer, close, error_handler, location),
            getvalue,
            location)
            

    def create_string(self, mol, id, writer_args, error_handler):
        output_stream, close, getvalue = open_output_string_java()
        cdk_writer, new_writer_args = self.open_writer_impl(output_stream, writer_args)
        try:
            if id is None:
                cdk_writer.write(mol)
            else:
                old_title = mol.getProperty(TITLE)
                if old_title == id:
                    cdk_writer.write(mol)
                else:
                    mol.setProperty(TITLE, id)
                    try:
                        cdk_writer.write(mol)
                    finally:
                        mol.setProperty(TITLE, old_title)
        except (cdk.exception.CDKException, java.lang.Exception) as err:
            error_handler.error(
                "CDK cannot create %s record%s: %s" % (
                    self.record_format, _get_errmol_info(mol), _get_error_message(err)))
            return None
        cdk_writer.close()
        return getvalue()

    ### Base implementation
    def open_reader_impl(self, input_stream, reader_args):
        if input_stream is None:
            reader = self.reader_class()
        else:
            reader = self.reader_class(input_stream)
        new_reader_args = _set_io_settings(reader, reader_args, self.input.defaults)
        return reader, new_reader_args

    def open_writer_impl(self, output_stream, writer_args):
        writer = self.writer_class(output_stream)
        new_writer_args = _set_io_settings(writer, writer_args, self.output.defaults)
        return writer, new_writer_args
    

def _get_input_settings(reader_class):
    class input:
        pass
    _set_settings_options(reader_class().getSettings(), input)
    return reader_class, input

def _get_output_settings(writer_class):
    class output:
        pass
    _set_settings_options(writer_class().getSettings(), output)
    return writer_class, output

class PDBFormat(BaseSingleRecordFormat):
    register_name = record_format = "pdb"
    alternate_extensions = ("ent",)

    reader_class, input = _get_input_settings(cdk.io.PDBReader)
    writer_class, output = _get_output_settings(cdk.io.PDBWriter)

#### The CIF format doesn't seem to generate atom containers
## class CIFFormat(BaseSingleRecordFormat):
##     register_name = record_format = "cif"
##     reader_class, input = _get_input_settings(cdk.io.CIFReader)
##     output = None

## class CDKOWLFormat(BaseSingleRecordFormat):
##     register_name = record_format = "owl"
##     reader_class, input = _get_input_settings(cdk.io.rdf.CDKOWLReader)
##     writer_class, output = _get_output_settings(cdk.io.rdf.CDKOWLWriter)

class CDKSourceCodeFormat(BaseSingleRecordFormat):
    register_name = record_format = "cdkjava"
    alternate_extensions = ("java",)
    input = None
    writer_class, output = _get_output_settings(cdk.io.CDKSourceCodeWriter)

def _CMLReaderWrapper(stream):
    return cdk.io.CMLReader(stream.read())
    
def _CMLWriterWrapper(stream):
    x = stream.read()
    print(x)
    return cdk.io.CMLReader(x)
    
class _CMLWriterWrapper(object):
    pass
    
class CMLFormat(BaseSingleRecordFormat):
    register_name = record_format = "cml"
    reader_class, input = _get_input_settings(cdk.io.CMLReader)
    writer_class, output = _get_output_settings(cdk.io.CMLWriter)


class CTXFormat(BaseSingleRecordFormat):
    register_name = record_format = "ctx"
    reader_class, input = _get_input_settings(cdk.io.CTXReader)
    output = None

class CrystClustFormat(BaseSingleRecordFormat):
    register_name = record_format = "crystclust"
    reader_class, input = _get_input_settings(cdk.io.CrystClustReader)
    writer_class, output = _get_output_settings(cdk.io.CrystClustWriter)

class GamessFormat(BaseSingleRecordFormat):
    register_name = record_format = "gamess"
    # use "out" or "log"?
    reader_class, input = _get_input_settings(cdk.io.GamessReader)
    output = None
    
## Gaussian03Reader.java GaussianInputWriter.java
## Gaussian98Reader.java

# ./storage/io/src/main/java/org/openscience/cdk/io/GhemicalMMReader.java    

# HINReader.java HINWriter.java


#./storage/ctab/src/main/java/org/openscience/cdk/io/MDLRXNV2000Reader.java
#./storage/ctab/src/main/java/org/openscience/cdk/io/MDLRXNV3000Reader.java
#./storage/ctab/src/main/java/org/openscience/cdk/io/MDLRXNWriter.java

class Mol2Format(BaseSingleRecordFormat):
    register_name = record_format = "mol2"
    reader_class, input = _get_input_settings(cdk.io.Mol2Reader)
    writer_class, output = _get_output_settings(cdk.io.Mol2Writer)
    
# Mopac7Reader.java Mopac7Writer.java

#./storage/smiles/src/main/java/org/openscience/cdk/io/MoSSOutputReader.java

# PCCompoundASNReader.java
# PCCompoundXMLReader.java
# PCSubstanceXMLReader.java
# PMPReader.java
# RGroupQueryReader.java RGroupQueryWriter.java
# ShelXReader.java ShelXWriter.java
#    RssWriter.java
#VASPReader.java

class XYZFormat(BaseSingleRecordFormat):
    register_name = record_format = "xyz"
    reader_class, input = _get_input_settings(cdk.io.XYZReader)
    writer_class, output = _get_output_settings(cdk.io.XYZWriter)

# ZMatrixReader.java

####
format_registry = base_toolkit.FormatRegistry(EXTERNAL_NAME, "cdk")

format_registry.register_formats(
    SmiFormat(),
    CanFormat(),
    UsmFormat(),
    SDFFormat(),
    SmiStringFormat(),
    CanStringFormat(),
    UsmStringFormat(),
    MolfileFormat(),
    
    InChIFormat(),
    InChIKeyFormat(),
    InChIStringFormat(),
    InChIKeyStringFormat(),

    PDBFormat(),
    
    #CIFFormat(), # It doesn't seem to support reading structures
    #CDKOWLFormat(),
    CDKSourceCodeFormat(),
    CMLFormat(),
    CTXFormat(),
    CrystClustFormat(),
    GamessFormat(),
    Mol2Format(),
    XYZFormat(),
    )
        
################ The toolkit API



class CDKToolkit(base_toolkit.BaseToolkit):
    def _open_molecule_writer_impl(self, format_config, destination, compression,
                                   writer_args, error_handler, location,
                                   encoding, encoding_errors, level):
        # Open the file and pass to the format
        handler = format_config.write_file_handler
        if handler == "chemfp":
            exc = None
            try:
                outfile, close = io.open_compressed_output_wrapped(destination, compression, EXTERNAL_NAME,
                                                                   encoding, encoding_errors, level)
            except ValueError as err:
                exc = ValueError("CDK does not support compression type %r: %s" % (compression, err))
            if exc is not None:
                raise exc
        elif handler == "cdk":
            exc = None
            try:
                outfile, close = open_compressed_output_java(destination, compression, EXTERNAL_NAME,
                                                             encoding, encoding_errors, level)
            except ValueError as err:
                exc = ValueError("CDK does not support compression type %r: %s" % (compression, err))
            if exc is not None:
                raise exc
            
        else:
            raise AssertionError(handler)
            

        return format_config.open_molecule_writer(destination, outfile, close, writer_args,
                                                  error_handler, location, encoding, encoding_errors)
    
    def _read_ids_and_molecules_from_string_impl(self,
                                                 format_config, content, compression, id_tag,
                                                 reader_args, error_handler, location):
        handler = format_config.read_string_handler
        if handler == "chemfp_cdk":
            if parse_implementation(reader_args["implementation"]) == "cdk":
                source = open_string_input_java(content, compression, id_tag)
            else:
                source = open_string_input_python(content, compression, id_tag)
        elif handler == "chemfp":
            source = open_string_input_python(content, compression, id_tag)
        elif handler == "cdk":
            source = open_string_input_java(content, compression, id_tag)
        else:
            raise AssertionError(handler)
        return format_config.read_ids_and_molecules(source, compression, id_tag, reader_args,
                                                        error_handler, location,
                                                        "utf8", "strict")

    def _check_read_encoding(self, format, encoding, encoding_errors):
        base_toolkit.only_utf8(format, encoding, encoding_errors)

    def _check_write_encoding(self, format, encoding, encoding_errors):
        base_toolkit.only_utf8(format, encoding, encoding_errors)


toolkit = CDKToolkit(format_registry)
