# Copyright (c) 2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

"""cdk_toolkit - chemfp toolkit API based on CDK

The cdk_toolkit implements the chemfp toolkit API using CDK
as the underlying chemistry toolkit. 
"""

from __future__ import absolute_import

# The public "cdk_toolkit" toolkit interface shouldn't expose internal details

from . import _cdk_toolkit as _tk

__all__ = [
    "is_licensed",
    "get_formats",
    "get_input_formats",
    "get_output_formats",
    "get_format",
    "get_input_format",
    "get_output_format",
    "get_input_format_from_source",
    "get_output_format_from_destination",
    "read_molecules",
    "read_molecules_from_string",
    "read_ids_and_molecules",
    "read_ids_and_molecules_from_string",
    "make_id_and_molecule_parser",
    "parse_molecule",
    "parse_id_and_molecule",
    "create_string",
    "open_molecule_writer",
    "open_molecule_writer_to_string",
    "open_molecule_writer_to_bytes",
    "copy_molecule",
    "add_tag",
    "get_tag",
    "get_tag_pairs",
    "get_id",
    "set_id",
    ]

name = "cdk"
software = _tk.SOFTWARE

def is_licensed():
    """Return True - CDK is always licensed

    :returns: True
    """
    return True


_format_registry = _tk.format_registry
_toolkit = _tk.toolkit

# format API

def get_formats(include_unavailable=False):
    """Get the list of structure formats that CDK supports

    If *include_unavailable* is True then also include CDK formats
    which aren't available to this specific version of CDK.

    :param include_unavailable: include unavailable formats?
    :type include_unavailable: True or False
    :returns: a list of Format objects
    """
    return _format_registry.get_formats(include_unavailable)

def get_input_formats():
    """Get the list of supported CDK input formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_formats()

def get_output_formats():
    """Get the list of supported CDK output formats

    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_formats()

def get_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not available.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_format(format)

def get_input_format(format):
    """Get the named input format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not an input format.

    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_input_format(format)

def get_output_format(format):
    """Get the named format, or raise a ValueError

    This will raise a ValueError if CDK does not implement
    the format *format_name* or that format is not an output format.
    
    :param format_name: the format name
    :type format_name: a string
    :returns: a list of :class:`chemfp.base_toolkit.Format` objects
    """
    return _format_registry.get_output_format(format)

def get_input_format_from_source(source=None, format=None):
    """Get the most appropriate format given the available source and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *source* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.
    
    :param source: the structure data source.
    :type source: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_input_format_from_source(source, format)

def get_output_format_from_destination(destination=None, format=None):
    """Get the most appropriate format given the available destination and format information

    If *format* is a :class:`chemfp.base_toolkit.Format` then
    return it. If it's a Format-like object with "name" and "compression"
    attributes use it to make a real Format object with the same
    attributes. If it's a string then use it to create a Format object.

    If *format* is None, use the *destination* to auto-detect the format.
    If auto-detection is not possible, assume it's an uncompressed
    SMILES file.

    :param destination: The structure data source.
    :type destination: a filename (as a string), a file object, or None to read from stdin
    :param format: format information, if known.
    :type format: a Format(-like) object, string, or None
    :returns: a :class:`chemfp.base_toolkit.Format` object
    """
    return _format_registry.get_output_format_from_destination(destination, format)

# read/write API

def read_molecules(source=None, format=None, id_tag=None, reader_args=None,
                   errors="strict", location=None,
                   encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads CDK molecules from a structure file

    Iterate through the *format* structure records in *source*. If *format*
    is None then auto-detect the format based on the *source*. For SD files,
    use *id_tag* to get the record id from the given SD tag instead of the
    title line. (read_molecules() will ignore the *id_tag*. It exists to
    make it easier to switch between reader functions.)

    Note: the reader returns a new CDK molecule each time.

    The *reader_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * has_header - True or False
      * sanitize - True or default sanitizes; False for unsanitized processing

    * InChI
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing
      
    * SDF
    
      * sanitize - True or default sanitizes; False for unsanitized processing
      * removeHs - True or default removes explicit hydrogens; False leaves them in the structure
      * strictParsing - True or default for strict parsing; False for lenient parsing

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If None
    then a default Location will be created.

    See :func:`chemfp.cdk_toolkit.read_ids_and_molecules` if you want (id, molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader parameters passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    return _toolkit.read_molecules(source, format, id_tag, reader_args, errors, location,
                                   encoding, encoding_errors)

def read_molecules_from_string(content, format, id_tag=None, reader_args=None,
                               errors="strict", location=None):
    """Return an iterator that reads CDK molecules from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.cdk_toolkit.read_molecules` for details about the other
    parameters.  See :func:`chemfp.cdk_toolkit.read_ids_and_molecules_from_string`
    if you  want to read (id, CDK molecule) pairs instead of just molecules.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeReader` iterating CDK molecules
    """
    return _toolkit.read_molecules_from_string(content, format, id_tag, reader_args,
                                               errors, location)

def read_ids_and_molecules(source=None, format=None, id_tag=None, reader_args=None,
                           errors="strict", location=None, encoding="utf8", encoding_errors="strict"):
    """Return an iterator that reads (id, CDK molecule) pairs from a structure file

    See :func:`chemfp.cdk_toolkit.read_molecules` for full parameter details.
    The major difference is that this returns an iterator of (id, CDK molecule)
    pairs instead of just the molecules.

    :param source: the structure source
    :type source: a filename, file object, or None to read from stdin
    :param format: the input structure format
    :type format: a format name string, or Format object, or None to auto-detect
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, CDK molecule) pairs
    """
    return _toolkit.read_ids_and_molecules(source, format, id_tag, reader_args, errors, location,
                                           encoding, encoding_errors)

def read_ids_and_molecules_from_string(content, format, id_tag=None, reader_args=None,
                                       errors="strict", location=None):
    """Return an iterator that reads (id, CDK molecule) pairs from a string containing structure records

    *content* is a string containing 0 or more records in the format *format*. See
    :func:`chemfp.cdk_toolkit.read_molecules` for details about the other
    parameters. See :func:`chemfp.cdk_toolkit.read_molecules_from_string`
    if you just want to read the CDK molecules instead of (id, molecule) pairs.

    :param content: the string containing structure records
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track parser state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.IdAndMoleculeReader` iterating (id, CDK molecule) pairs
    """
    return _toolkit.read_ids_and_molecules_from_string(content, format, id_tag, reader_args,
                                                       errors, location)

def make_id_and_molecule_parser(format, id_tag=None, reader_args=None, errors="strict"):
    """Create a specialized function which takes a record and returns an (id, CDK molecule) pair

    The returned function is optimized for reading many records from individual
    strings because it only does parameter validation once. However, I haven't
    really noticed much of a performance difference between this and
    :func:`chemfp.cdk_toolkit.parse_id_and_molecule` so you can probably
    so I suggest you use that function directly instead of making a specialized function.
    (Let me know if making a specialized function is useful.)

    See :func:`chemfp.cdk_toolkit.read_molecules` for details about the
    other parameters.

    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a function of the form ``parser(record string) -> (id, CDK molecule)``
    """
    return _toolkit.make_id_and_molecule_parser(format, id_tag, reader_args, errors)

def parse_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from the *content* string and return a CDK molecule.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.cdk_toolkit.read_molecules`
    for details about the other parameters. See :func:`chemfp.cdk_toolkit.parse_id_and_molecule`
    if you want the (id, CDK molecule) pair instead of just the molecule.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a CDK molecule
    """
    return _toolkit.parse_molecule(content, format, id_tag, reader_args, errors)

def parse_id_and_molecule(content, format, id_tag=None, reader_args=None, errors="strict"):
    """Parse the first structure record from *content* and return the (id, CDK molecule) pair.

    *content* is a string containing a single structure record in format *format*.
    (Additional records are ignored). See :func:`chemfp.cdk_toolkit.read_molecules`
    for details about the other parameters.

    See :func:`chemfp.cdk_toolkit.read_molecules` for details about the
    other parameters. See :func:`chemfp.cdk_toolkit.parse_molecule`
    if just want the CDK molecule and not the the (id, CDK molecule) pair.

    :param content: the string containing a structure record
    :type content: a string
    :param format: the input structure format
    :type format: a format name string, or Format object
    :param id_tag: SD tag containing the record id
    :type id_tag: string, or None to use the record title
    :param reader_args: reader arguments passed to the underlying toolkit
    :type reader_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: an (id, CDK molecule) pair
    """
    return _toolkit.parse_id_and_molecule(content, format, id_tag, reader_args, errors)


def create_string(mol, format, id=None, writer_args=None, errors="strict"):
    """Convert a CDK molecule into a structure record in the given format as a Unicode string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: a CDK molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :returns: a Unicode string
    """
    return _toolkit.create_string(mol, format, id, writer_args, errors)

def create_bytes(mol, format, id=None, writer_args=None, errors="strict", level=None):
    """Convert a CDK molecule into a structure record in the given format as a byte string

    If *id* is not None then use it instead of the molecule's own
    title. Warning: this may briefly modify the molecule, so may not
    be thread-safe.

    :param mol: the molecule to use for the output
    :type mol: a CDK molecule
    :param format: the output structure format
    :type format: a format name string, or Format object
    :param id: an alternate record id
    :type id: a string, or None to use the molecule's own id
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a byte string
    """
    return _toolkit.create_bytes(mol, format, id, writer_args, errors, level)


def open_molecule_writer(destination=None, format=None, writer_args=None,
                         errors="strict", location=None,
                         encoding="utf8", encoding_errors="strict", level=None):
    """Return a MoleculeWriter which can write CDK molecules to a destination.

    A :class:`chemfp.base_toolkit.MoleculeWriter` has the methods ``write_molecule``,
    ``write_molecules``, and ``write_ids_and_molecules``, which are ways to write
    a CDK molecule, a CDK molecule iterator, or an (id, CDK molecule) pair
    iterator to a file.
    
    Molecules are written to *destination*. The output format can be a
    string like "sdf.gz" or "smi", a :class:`chemfp.base_toolkit.Format`,
    or Format-like object with "name" and "compression" attributes, or None
    to auto-detect based on the *destination*. If auto-detection is not
    possible, the output will be written as uncompressed SMILES.
    
    The *writer_args* dictionary parameters depend on the format. These include:

    * SMILES
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * isomericSmiles - True to generate isomeric SMILES
      * kekuleSmiles - True to generate SMILES in Kekule form
      * canonical - True to generate a canonical SMILES
      * allBondsExplicit - True to write explict '-' and ':' bonds, even if they can be inferred; default is False
      * allHsExplicit - True to write explicit hydrogen counts; default is False
      * cxsmiles - True to include CXSMILES annotations; default is False

    InChI and InChIKey
    
      * delimiter - one of "tab", "space", "to-eol", the space or tab characters, or None
      * include_id - True or default to include the id as the second column; False has no id column
      * options - an options string passed to the underlying InChI library
      * logLevel - an integer log level
      * treatWarningAsError - True raises an exception on error; False or default keeps processing

    SDF
    
      * includeStereo - True include stereo information; False or default does not
      * kekulize - True or default creates the connection table with bonds in Kekeule form
      * v3k - True to alway export in V3000 format

    The *errors* parameter specifies how to handle errors. "strict" raises
    an exception, "report" sends a message to stderr and goes to the next
    record, and "ignore" goes to the next record.

    The *location* parameter takes a :class:`chemfp.io.Location` instance. If
    None then a default Location will be created.
    
    :param destination: the structure destination
    :type destination: a filename, file object, or None to write to stdout
    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer parameters passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer(destination, format, writer_args, errors, location,
                                         encoding, encoding_errors, level)
                                         

def open_molecule_writer_to_string(format, writer_args=None, errors="strict", location=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a string.

    See :func:`chemfp.cdk_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a Unicode string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer_to_string(format, writer_args, errors, location)

def open_molecule_writer_to_bytes(format, writer_args=None, errors="strict", location=None, level=None):
    """Return a MoleculeStringWriter which can write molecule records in the given format to a text string.

    See :func:`chemfp.cdk_toolkit.open_molecule_writer` for full
    parameter details.

    Use the writer's :meth:`chemfp.base_toolkit.MoleculeStringWriter.getvalue`
    to get the output as a byte string.

    :param format: the output structure format
    :type format: a format name string, or Format(-like) object, or None to auto-detect
    :param writer_args: writer arguments passed to the underlying toolkit
    :type writer_args: a dictionary
    :param errors: specify how to handle errors
    :type errors: one of "strict", "report", or "ignore"
    :param location: object used to track writer state information
    :type location: a :class:`chemfp.io.Location` object, or None
    :param level: compression level to use for compressed formats
    :type level: None, a positive integer, or one of the strings 'min', 'default', or 'max'
    :returns: a :class:`chemfp.base_toolkit.MoleculeStringWriter` expecting CDK molecules
    """
    return _toolkit.open_molecule_writer_to_bytes(format, writer_args, errors, location, level)


# molecule manipulation API

def copy_molecule(mol):
    """Return a new CDK molecule which is a copy of the given molecule

    :param mol: the molecule to copy
    :type mol: a CDK molecule
    :returns: a new CDK Mol instance
    """
    return mol.clone()

def add_tag(mol, tag, value):
    """Add an SD tag value to the CDK molecule

    :param mol: the molecule
    :type mol: a CDK molecule
    :param tag: the SD tag name
    :type tag: string
    :param value: the text for the tag
    :type value: string
    :returns: None
    """
    mol.setProperty(tag, value)

def get_tag(mol, tag):
    """Get the named SD tag value, or None if it doesn't exist

    :param mol: the molecule
    :type mol: a CDK molecule
    :param tag: the SD tag name
    :type tag: string
    :returns: a string, or None
    """
    return mol.getProperty(tag)
    
def get_tag_pairs(mol):
    """Get a list of all SD tag (name, value) pairs for the molecule

    :param mol: the molecule
    :type mol: a CDK molecule
    :returns: a list of (string name, string value) pairs
    """
    it = ((str(x.getKey()), str(x.getValue())) for x in mol.getProperties().items())
    return [(name, value) for (name, value) in it if name[:4] != "cdk:"]

def get_id(mol):
    """Get the molecule's id from CDK's "cdk:Title" property

    :param mol: the molecule
    :type mol: a CDK molecule
    :returns: a string
    """
    id = mol.getProperty(_tk.TITLE)
    if id is not None:
        id = str(id)
    return id

def set_id(mol, id):
    """Set the molecule's id as CDK's "cdk:Title" property

    :param mol: the molecule
    :type mol: a CDK molecule
    :param id: the new id
    :type id: string
    :returns: None
    """
    mol.setProperty(_tk.TITLE, id)
