# Copyright (c) 2020-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

## /LICENSE ##

from __future__ import absolute_import, print_function

from . import cdk_toolkit as _tk
from ._cdk_toolkit import cdk, java, _default_builder

from . import types

from .bitops import byte_from_bitlist


software = _tk.software
from . import SOFTWARE as CHEMFP_SOFTWARE

_available_types = []
_unavailable_types = []

def _register_if(status,
                 fingerprint_type_class,
                 unavailable_message=None):
    if status:
        _available_types.append(fingerprint_type_class)
        return
    if unavailable_message is None:
        unavailable_message = ("This version of CDK does not support the %s fingerprint"
                               % (fingerprint_type_class.name,) )
    _unavailable_types.append((fingerprint_type_class.name, unavailable_message))

########

# Some constants shared by the fingerprinter and the command-line code.
DEFAULT_SIZE = 1024 # DaylightFingerprinter.DEFAULT_SIZE
DEFAULT_SEARCH_DEPTH = 7 # DaylightFingerprinter.DEFAULT_SEARCH_DEPTH
DEFAULT_PATH_LIMIT = 42000 # DaylightFingerprinter.PATH_LIMIT
DEFAULT_HASH_PSEUDO_ATOMS = False

DEFAULT_PERCEIVE_STEREOCHEMISTRY = False

all_arguments = types.FingerprintArguments()
_add = all_arguments.add_argument

_add("size", decoder=types.positive_int, metavar="INT", default=DEFAULT_SIZE,
         help = "fingerprint size")
_add("searchDepth", decoder=types.nonnegative_int, metavar="INT", default=DEFAULT_SEARCH_DEPTH,
         help = "search depth")
_add("pathLimit", decoder=types.nonnegative_int, metavar="INT", default=DEFAULT_PATH_LIMIT,
         help = "path limit")
_add("hashPseudoAtoms", decoder=types.zero_or_one, metavar="0|1", default=int(DEFAULT_HASH_PSEUDO_ATOMS),
         encoder = types.encode_bool_as_zero_or_one,
         help = "include pseudo-atoms in path enumeration")
_add("perceiveStereochemistry", decoder=types.zero_or_one, metavar="0|1", default=int(DEFAULT_PERCEIVE_STEREOCHEMISTRY),
         encoder = types.encode_bool_as_zero_or_one,
         help = "re-perceive stereochemistry from 2D/3D coordinates")

class BaseCDKFingerprintType(types.ThreadsafeFingerprinterMixin,
                             types.FingerprintType):
    toolkit = _tk
    software = _tk.software + " " + CHEMFP_SOFTWARE
    fingerprinter_can_fail = True

class VariableSizeFingerprint(BaseCDKFingerprintType):
    def __init__(self, fingerprint_kwargs):
        super(VariableSizeFingerprint, self).__init__(fingerprint_kwargs)
        self.num_bits = fingerprint_kwargs["size"]

class FixedSizeFingerprint(BaseCDKFingerprintType):
    pass

#    public final String getVersionDescription() {
#        StringBuilder sb = new StringBuilder();
#        sb.append("CDK-")
#          .append(getClass().getSimpleName())
#          .append("/")
#          .append(CDK.getVersion()); // could version fingerprints separetely
#        for (Map.Entry<String,String> param : getParameters()) {
#            sb.append(' ').append(param.getKey()).append('=').append(param.getValue());
#        }
#        return sb.toString();
#    }

#### CDK-Daylight -- The Daylight-like fingerprinter

# cdk.fingerprint.Fingerprinter(int size, int searchDepth)
# #type=CDK-Fingerprinter/2.0 searchDepth=7 pathLimit=2000 hashPseudoAtoms=true
#    new SimpleImmutableEntry<>("searchDepth", Integer.toString(searchDepth)),
#    new SimpleImmutableEntry<>("pathLimit", Integer.toString(pathLimit)),
#    new SimpleImmutableEntry<>("hashPseudoAtoms", Boolean.toString(hashPseudoAtoms))

# getSize()
# getSearchDepth()
# setHashPseudoAtoms()
# setPathLimit()


# CDK says "The FingerPrinter assumes that hydrogens are explicitly given!"
# 
DaylightFingerprinter = cdk.fingerprint.Fingerprinter

def make_daylight_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = DaylightFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_daylight_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_daylight_fingerprinter.__qualname__ = "cdk_daylight_fingerprinter"
    cdk_daylight_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_daylight_fingerprinter


class CDKDaylightFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Daylight-like fingerprint, version 2.0
    
    The CDK-Daylight/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Daylight/2.0"
    _make_fingerprinter = staticmethod(make_daylight_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")

_register_if(True, CDKDaylightFingerprintType_v20)

#### CDK-GraphOnly

# Modified version of DaylightFingerprinter but bond symbols are all ""
GraphOnlyFingerprinter = cdk.fingerprint.GraphOnlyFingerprinter

def make_graphonly_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = GraphOnlyFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_graphonly_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_graphonly_fingerprinter.__qualname__ = "cdk_graphonly_fingerprinter"
    cdk_graphonly_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_graphonly_fingerprinter


class CDKGraphOnlyFingerprintType_v20(VariableSizeFingerprint):
    """CDK's GraphOnly fingerprint, version 2.0
    
    The CDK-GraphOnly/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-GraphOnly/2.0"
    _make_fingerprinter = staticmethod(make_graphonly_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")

_register_if(True, CDKGraphOnlyFingerprintType_v20)

#### CDK-MACCS

# "This class assumes that aromaticity perception, atom typing and adding of
#  implicit hydrogens have been performed prior to generating the fingerprint."
 
MACCSFingerprinter = cdk.fingerprint.MACCSFingerprinter

def make_maccs_fingerprinter():
    fingerprinter = MACCSFingerprinter()
    n = fingerprinter.getSize()
    if n != 166:
        raise AssertionError(("Unexpected MACCSFingerprinter size", n, 166))
    def cdk_maccs_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 166)
    cdk_maccs_fingerprinter.__qualname__ = "cdk_maccs_fingerprinter"
    cdk_maccs_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_maccs_fingerprinter

class CDKMACCSFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the 166 MACCS keys, version 2.0

    The CDK-MACCS/2.0 fingerprints have no parameters.
    """
    num_bits = 166
    name = "CDK-MACCS/2.0"
    _make_fingerprinter = staticmethod(make_maccs_fingerprinter)

_register_if(True, CDKMACCSFingerprintType_v20)
    
#### CDK-EState

# "This fingerprinter generates 79 bit fingerprints using the E-State fragments."

EStateFingerprinter = cdk.fingerprint.EStateFingerprinter

def make_estate_fingerprinter():
    fingerprinter = EStateFingerprinter()
    n = fingerprinter.getSize()
    if n != 79:
        raise AssertionError(("Unexpected EStateFingerprinter size", n, 79))
    def cdk_estate_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 79)
    cdk_estate_fingerprinter.__qualname__ = "cdk_estate_fingerprinter"
    cdk_estate_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_estate_fingerprinter

class CDKEStateFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the EState fingerprint, version 2.0

    The CDK-EState/2.0 fingerprints have no parameters.
    """
    num_bits = 79
    name = "CDK-EState/2.0"
    _make_fingerprinter = staticmethod(make_estate_fingerprinter)

_register_if(True, CDKEStateFingerprintType_v20)
    
#### CDK-Extended

ExtendedFingerprinter = cdk.fingerprint.ExtendedFingerprinter

def make_extended_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size >= 26):
        raise ValueError("size must be 26 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)
    fingerprinter = ExtendedFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_extended_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_extended_fingerprinter.__qualname__ = "cdk_extended_fingerprinter"
    cdk_extended_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_extended_fingerprinter


class CDKExtendedFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Extended fingerprint, version 2.0
    
    The CDK-Extended/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Extended/2.0"
    _make_fingerprinter = staticmethod(make_extended_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")

_register_if(True, CDKExtendedFingerprintType_v20)


#### CDK-Hybridization

# "The FingerPrinter assumes that hydrogens are explicitly given!"
# "Furthermore, if pseudo atoms or atoms with malformed symbols are present, ..."
# "Unlike the {@link Fingerprinter}, this fingerprinter does not take into
#     account aromaticity. Instead, it takes into account SP2 hybridization."

HybridizationFingerprinter = cdk.fingerprint.HybridizationFingerprinter

def make_hybridization_fingerprinter(
        size = DEFAULT_SIZE,
        searchDepth = DEFAULT_SEARCH_DEPTH,
        pathLimit = DEFAULT_PATH_LIMIT,
        hashPseudoAtoms = DEFAULT_HASH_PSEUDO_ATOMS):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    if not (searchDepth >= 0):
        raise ValueError("searchDepth must be 0 or greater")
    if not (pathLimit > 0):
        raise ValueError("pathLimit must be 1 or greater")
    hashPseudoAtoms = bool(hashPseudoAtoms)

    fingerprinter = HybridizationFingerprinter(size, searchDepth)
    fingerprinter.setPathLimit(pathLimit)
    fingerprinter.setHashPseudoAtoms(hashPseudoAtoms)
    def cdk_hybridization_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_hybridization_fingerprinter.__qualname__ = "cdk_hybridization_fingerprinter"
    cdk_hybridization_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_hybridization_fingerprinter

class CDKHybridizationFingerprintType_v20(VariableSizeFingerprint):
    """CDK's Hybridization fingerprint, version 2.0
    
    The CDK-Hybridization/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * searchDepth - maximum path length (default: 7)
    * pathLimit - maximum number of paths (default: 42000)
    * hashPseudoAtoms - include pseudo-atoms in the hash calculation (default: 0)
    """
    name = "CDK-Hybridization/2.0"
    _make_fingerprinter = staticmethod(make_hybridization_fingerprinter)
    _arguments = all_arguments.select_names("size searchDepth pathLimit hashPseudoAtoms")
    _format_string = ("size={size} searchDepth={searchDepth} "
                          "pathLimit={pathLimit} hashPseudoAtoms={hashPseudoAtoms}")

_register_if(True, CDKHybridizationFingerprintType_v20)

#### CDK-KlekotaRoth

# "SMARTS based substructure fingerprint based on <b>Chemical substructures
#  that enrich for biological activity</b> {@cdk.cite Klekota01112008}.
#
#   Presence of 4860 substructures"

# For some reason I am unable to get the KlekotaRothFingerprinter through jpype
# on macOS.
KlekotaRothFingerprinter = getattr(cdk.fingerprint, "KlekotaRothFingerprinter", None)

_klekota_roth_fingerprinter = None
def make_klekota_roth_fingerprinter():
    global _klekota_roth_fingerprinter
    if _klekota_roth_fingerprinter is not None:
        return _klekota_roth_fingerprinter

    fingerprinter = KlekotaRothFingerprinter()
    n = fingerprinter.getSize()
    if n != 4860:
        raise AssertionError(("Unexpected KlekotaRothFingerprinter size", n, 4860))
    def cdk_klekota_roth_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 4860)
    cdk_klekota_roth_fingerprinter.__qualname__ = "cdk_klekota_roth_fingerprinter"
    cdk_klekota_roth_fingerprinter.cdk_fingerprinter = fingerprinter
    _klekota_roth_fingerprinter = cdk_klekota_roth_fingerprinter # cache for future reuse    
    return cdk_klekota_roth_fingerprinter

class CDKKlekotaRothFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the Klekota-Roth fingerprints

    The CDK-KlekotaRoth/2.0 fingerprints have no parameters.
    """
    num_bits = 4860
    name = "CDK-KlekotaRoth/2.0"
    _make_fingerprinter = staticmethod(make_klekota_roth_fingerprinter)

_register_if(KlekotaRothFingerprinter is not None, CDKKlekotaRothFingerprintType_v20,
                 "jpype is unable to load CDK's KlekotaRothFingerprinter class")
    
#### CDK-Pubchem

# "this class is not thread-safe and uses stores intermediate steps internally."

PubchemFingerprinter = cdk.fingerprint.PubchemFingerprinter

def make_pubchem_fingerprinter():
    fingerprinter = PubchemFingerprinter(_default_builder)
    n = fingerprinter.getSize()
    if n != 881:
        raise AssertionError(("Unexpected PubchemFingerprinter size", n, 881))
    def cdk_pubchem_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 881)
    cdk_pubchem_fingerprinter.__qualname__ = "cdk_pubchem_fingerprinter"
    cdk_pubchem_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_pubchem_fingerprinter


class CDKPubchemFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the Klekota-Roth fingerprints

    The CDK-Pubchem/2.0 fingerprints have no parameters.
    """
    num_bits = 881
    name = "CDK-Pubchem/2.0"
    _make_fingerprinter = staticmethod(make_pubchem_fingerprinter)

_register_if(True, CDKPubchemFingerprintType_v20)

#### CDK-Substructure

SubstructureFingerprinter = cdk.fingerprint.SubstructureFingerprinter

def make_substructure_fingerprinter():
    fingerprinter = SubstructureFingerprinter()
    n = fingerprinter.getSize()
    if n != 307:
        raise AssertionError(("Unexpected SubstructureFingerprinter size", n, 307))
    def cdk_substructure_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 307)
    cdk_substructure_fingerprinter.__qualname__ = "cdk_substructure_fingerprinter"
    cdk_substructure_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_substructure_fingerprinter

class CDKSubstructureFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's Substructure fingerprints

    The CDK-Substructure/2.0 fingerprints have no parameters.
    """
    num_bits = 307
    name = "CDK-Substructure/2.0"
    _make_fingerprinter = staticmethod(make_substructure_fingerprinter)

_register_if(True, CDKSubstructureFingerprintType_v20)

#### CDK-ShortestPath

# "It is recommended to use atomtyped container before generating the fingerprints."

# "The FingerPrinter calculates fingerprint based on the Shortest
# Paths between two atoms. It also takes into account ring system,
# charges etc while generating a fingerprint."

# "The FingerPrinter assumes that hydrogens are explicitly given!
# Furthermore, if pseudo atoms or atoms with malformed symbols are
# present, their atomic number is taken as one more than the last
# element currently supported in {@link PeriodicTable}.

ShortestPathFingerprinter = cdk.fingerprint.ShortestPathFingerprinter

def make_shortest_path_fingerprinter(size = DEFAULT_SIZE):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = ShortestPathFingerprinter(size)
    def cdk_shortest_path_fingerprinter(mol):
        try:
            fp = fingerprinter.getBitFingerprint(mol)
        except java.lang.Exception as err:
            if "atom has unset atom type" in str(err.getMessage()):
                # Set the atom types in a copy
                mol2 = mol.clone()
                cdk.tools.manipulator.AtomContainerManipulator.percieveAtomTypesAndConfigureAtoms(mol2)
                fp = fingerprinter.getBitFingerprint(mol2)
            else:
                raise
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_shortest_path_fingerprinter.__qualname__ = "cdk_shortest_path_fingerprinter"
    cdk_shortest_path_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_shortest_path_fingerprinter

class CDKShortestPathFingerprintType_v20(VariableSizeFingerprint):
    """CDK's ShortestPath fingerprint, version 2.0
    
    The CDK-ShortestPath/2.0 :class:`.FingerprintType` parameter is:
    
    * size - the number of bits in the fingerprint (default: 1024)
    """
    name = "CDK-ShortestPath/2.0"
    fingerprinter_can_fail = True
    _make_fingerprinter = staticmethod(make_shortest_path_fingerprinter)
    _arguments = all_arguments.select_names("size")
    _format_string = "size={size}"

_register_if(True, CDKShortestPathFingerprintType_v20)
    
#### CDK-Circular

#  "Circular fingerprints: for generating fingerprints that are
#  functionally equivalent to ECFP-2/4/6 and FCFP-2/4/6 fingerprints,
#  which are partially described by Rogers et al. {@cdk.cite
#  Rogers2010}."

CircularFingerprinter = cdk.fingerprint.CircularFingerprinter

CLASS_ECFP0 = CircularFingerprinter.CLASS_ECFP0
CLASS_ECFP2 = CircularFingerprinter.CLASS_ECFP2
CLASS_ECFP4 = CircularFingerprinter.CLASS_ECFP4
CLASS_ECFP6 = CircularFingerprinter.CLASS_ECFP6

CLASS_FCFP0 = CircularFingerprinter.CLASS_FCFP0
CLASS_FCFP2 = CircularFingerprinter.CLASS_FCFP2
CLASS_FCFP4 = CircularFingerprinter.CLASS_FCFP4
CLASS_FCFP6 = CircularFingerprinter.CLASS_FCFP6

# ECFP

def make_ecfp0_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP0, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp0_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp0_fingerprinter.__qualname__ = "cdk_ecfp0_fingerprinter"
    cdk_ecfp0_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp0_fingerprinter

class CDKECFP0FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP0 fingerprint, version 2.0
    
    The CDK-ECFP0/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP0/2.0"
    _make_fingerprinter = staticmethod(make_ecfp0_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKECFP0FingerprintType_v20)

def make_ecfp2_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP2, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp2_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp2_fingerprinter.__qualname__ = "cdk_ecfp2_fingerprinter"
    cdk_ecfp2_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp2_fingerprinter

class CDKECFP2FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP2 fingerprint, version 2.0
    
    The CDK-ECFP2/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP2/2.0"
    _make_fingerprinter = staticmethod(make_ecfp2_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKECFP2FingerprintType_v20)

def make_ecfp4_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP4, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp4_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp4_fingerprinter.__qualname__ = "cdk_ecfp4_fingerprinter"
    cdk_ecfp4_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp4_fingerprinter

class CDKECFP4FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP4 fingerprint, version 2.0
    
    The CDK-ECFP4/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP4/2.0"
    _make_fingerprinter = staticmethod(make_ecfp4_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKECFP4FingerprintType_v20)

def make_ecfp6_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_ECFP6, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_ecfp6_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_ecfp6_fingerprinter.__qualname__ = "cdk_ecfp6_fingerprinter"
    cdk_ecfp6_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_ecfp6_fingerprinter

class CDKECFP6FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the ECFP6 fingerprint, version 2.0
    
    The CDK-ECFP6/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-ECFP6/2.0"
    _make_fingerprinter = staticmethod(make_ecfp6_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKECFP6FingerprintType_v20)
    
# FCFP

def make_fcfp0_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP0, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp0_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp0_fingerprinter.__qualname__ = "cdk_fcfp0_fingerprinter"
    cdk_fcfp0_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp0_fingerprinter

class CDKFCFP0FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP0 fingerprint, version 2.0
    
    The CDK-FCFP0/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP0/2.0"
    _make_fingerprinter = staticmethod(make_fcfp0_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKFCFP0FingerprintType_v20)
    
def make_fcfp2_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP2, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp2_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp2_fingerprinter.__qualname__ = "cdk_fcfp2_fingerprinter"
    cdk_fcfp2_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp2_fingerprinter

class CDKFCFP2FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP2 fingerprint, version 2.0
    
    The CDK-FCFP2/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP2/2.0"
    _make_fingerprinter = staticmethod(make_fcfp2_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKFCFP2FingerprintType_v20)

def make_fcfp4_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP4, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp4_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp4_fingerprinter.__qualname__ = "cdk_fcfp4_fingerprinter"
    cdk_fcfp4_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp4_fingerprinter

class CDKFCFP4FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP4 fingerprint, version 2.0
    
    The CDK-FCFP4/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP4/2.0"
    _make_fingerprinter = staticmethod(make_fcfp4_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKFCFP4FingerprintType_v20)
    
def make_fcfp6_fingerprinter(size = DEFAULT_SIZE, perceiveStereochemistry = DEFAULT_PERCEIVE_STEREOCHEMISTRY):
    if not (size > 0):
        raise ValueError("size must be 1 or greater")
    fingerprinter = CircularFingerprinter(CLASS_FCFP6, size)
    fingerprinter.setPerceiveStereo(perceiveStereochemistry)
    def cdk_fcfp6_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), size)
    cdk_fcfp6_fingerprinter.__qualname__ = "cdk_fcfp6_fingerprinter"
    cdk_fcfp6_fingerprinter.cdk_fingerprinter = fingerprinter
    return cdk_fcfp6_fingerprinter

class CDKFCFP6FingerprintType_v20(VariableSizeFingerprint):
    """CDK's implementation of the FCFP6 fingerprint, version 2.0
    
    The CDK-FCFP6/2.0 :class:`.FingerprintType` parameters are:
    
    * size - the number of bits in the fingerprint (default: 1024)
    * perceiveStereochemistry - if True, re-perceive stereochemistry from 2D/3D coordinates
    """
    name = "CDK-FCFP6/2.0"
    _make_fingerprinter = staticmethod(make_fcfp6_fingerprinter)
    _arguments = all_arguments.select_names("size perceiveStereochemistry")
    _format_string = "size={size} perceiveStereochemistry={perceiveStereochemistry}"

_register_if(True, CDKFCFP6FingerprintType_v20)

#### CDK-AtomPairs2D

# "Generates an atom pair 2D fingerprint as implemented in PaDEL given
# an {@link IAtomContainer}, that extends the {@link Fingerprinter}."

# fixed length

AtomPairs2DFingerprinter = cdk.fingerprint.AtomPairs2DFingerprinter
_atom_pairs_2d_fingerprinter = None

def make_atom_pairs_2d_fingerprinter():
    global _atom_pairs_2d_fingerprinter
    x = _atom_pairs_2d_fingerprinter
    if x is not None:
        return x
    
    fingerprinter = AtomPairs2DFingerprinter()
    n = fingerprinter.getSize()
    if n != 780:
        raise AssertionError(("Unexpected AtomPairs2DFingerprinter size", n, 780))
    def cdk_atom_pairs_2d_fingerprinter(mol):
        fp = fingerprinter.getBitFingerprint(mol)
        return byte_from_bitlist(fp.getSetbits(), 780)
    cdk_atom_pairs_2d_fingerprinter.__qualname__ = "cdk_atom_pairs_2d_fingerprinter"
    cdk_atom_pairs_2d_fingerprinter.cdk_fingerprinter = fingerprinter
    _atom_pairs_2d_fingerprinter = cdk_atom_pairs_2d_fingerprinter # cache for future reuse
    return cdk_atom_pairs_2d_fingerprinter

class CDKAtomPairs2DFingerprintType_v20(types.NoFingerprintParametersMixin, FixedSizeFingerprint):
    """CDK's implementation of the atom pairs fingerprint from Yap Chun Wei's PaDEL.

    The CDK-AtomPairs2D/2.0 :class:`.FingerprintType` takes no parameters.
    """
    num_bits = 780
    name = "CDK-AtomPairs2D/2.0"
    _make_fingerprinter = staticmethod(make_atom_pairs_2d_fingerprinter)


_register_if(True, CDKAtomPairs2DFingerprintType_v20)

#### CDK-Signature

# Can't support -- variable length
# cdk/fingerprint/SignatureFingerprinter.java


####

def resolve_chemfp_fingerprint_types(registry):
    for name, reason in _unavailable_types:
        registry.register_unavailable(name, reason)
    for fingerprint_type in _available_types:
        registry.register_fingerprint_type(fingerprint_type)

