"""I/O routines for the FPB format.

The FPB format contains 8 bytes ("magic") followed by a sequence of
chunks. Each chunk contains three fields: size, tag, and data. These
are laid out as:
   8  bytes - number of bytes in the data field, uint64 in little endian format
   4  bytes - the tag name, as letters [A-Za-z][A-Za-z0-9]{3}
  <n> bytes - the data

The known chunks are:
  META - the metadata chunk. The content is exactly the '#' lines
          from the FPS format, including the '#' and newlines.
  TEXT - human-readable text. The first line is a ASCII label, the
          rest is UTF-8 text.
  AREN - the arena chunk. The fingerprint data as a chunk of bytes. Required.
  POPC - the popcount chunk. Starting indices to popcount segments. Optional.
  FPID - fingerprint identifiers, as UTF-8 encoded strings and a lookup index.
  HASH - a hash table to look up a record position given its identifier.
  CFPL - a chemfp FPB license token.
  FEND - the last chunk; chunk processing ends at this point.

The META, POPC, HASH, and CFPL chunks are optional. The AREN, FPID,
and FEND chunks are required. The TEXT section is not used by a chemfp
arena.

"""

# Copyright (c) 2010-2021 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import with_statement, absolute_import, print_function

from . import _compat
from ._compat import BytesIO, xrange, myrepr
import array
import ctypes
import itertools
import shutil
import struct
import sys
import tempfile
import mmap
import re
import io as stdlib_io

from ._compat import next

try:
    # Only available on Python 2.6 and later
    _getsizeof = sys.getsizeof
except AttributeError:
    def _getsizeof(s):
        return 37  # At least that's what my machine returns

# Only available on Python 2.6 and later
SpooledTemporaryFile = tempfile.SpooledTemporaryFile


import _chemfp
from . import Metadata, ChemFPError, FingerprintWriter
from . import bitops
from . import arena as arena_module
from . import fps_io
from . import io

if type(u"") == type(""):
    # Python 3, only support Unicode identifiers (will utf-8 encode)
    _djb_hash = _chemfp.djb_hash
else:
    # Python 2. Support byte and Unicode identifiers.
    # The underlying C code does not auto utf8 encode.
    _real_djb_hash = _chemfp.djb_hash
    def _djb_hash(s):
        if isinstance(s, unicode):
            s = s.encode("utf8")
        return _real_djb_hash(s)

class FPBError(ChemFPError):
    pass


# Debug object. Wrap it around an output file to see what's going on.
class _Watch(object):
    "THIS IS PART OF THE PRIVATE API"
    def __init__(self, output):
        self.output = output
    def seek(self, where, whence=None):
        if whence is not None:
            print("seek(%r, %r)" % (where, whence))
        else:
            print("seek(%r)" % (where,))
        self.output.seek(where)
    def tell(self):
        where = self.output.tell()
        print("tell()->", where)
        return where

    def write(self, s, __file__ = __file__):
        import inspect, sys, linecache
        lineno = inspect.getframeinfo(sys._getframe(1)).lineno
        line = linecache.getline(__file__, lineno)
        print("Wrote", self.output.tell(), repr(s[:26]), len(s), str(lineno) + ": " + line.strip()[:30])
        self.output.write(s)

    def writelines(self, lines):
        lines = list(lines)
        #import inspect, sys, linecache
        #lineno = inspect.getframeinfo(sys._getframe(1)).lineno
        #line = linecache.getline(__file__, lineno)
        if len(lines) < 5:
            print(len(lines), "lines at ", self.output.tell(), repr(lines))
        else:
            print(len(lines), "lines at ", self.output.tell(), repr(lines[:5]) + "...")
        self.output.writelines(lines)
        

    def read(self, amount=None):
        import inspect, sys, linecache
        lineno = inspect.getframeinfo(sys._getframe(1)).lineno
        line = linecache.getline(__file__, lineno)
        where = self.output.tell()
        s = self.output.read(amount)
        print("Read", where, repr(s[:25]), len(s), str(lineno) + ": " + line.strip()[:30])
        return s

    def fileno(self):
        import inspect, sys, linecache
        lineno = inspect.getframeinfo(sys._getframe(1)).lineno
        line = linecache.getline(__file__, lineno)
        print("fileno()", str(lineno) + ": " + line.strip()[:30])
        return self.output.fileno()


###### Functions for reading/writing a chunk

_pack_chunk_size = struct.Struct("<Q").pack
_unpack_chunk_size = struct.Struct("<Q").unpack

def write_chunk(output, name, data):
    assert len(name) == 4
    #assert only letters in name
    output.write(_pack_chunk_size(len(data)) + name + data)


class Chunk(object):
    def __init__(self, tag, content, start, end):
        self.tag = tag
        self.content = content
        self.start = start
        self.end = end

    def __repr__(self):
        return "Chunk(%r, ..., %d, %d)" % (self.tag, self.start, self.end)

    def __len__(self):
        return self.end - self.start

    def __getitem__(self, i):
        if isinstance(i, slice):
            start, end, stride = i.indices(self.end-self.start)
            assert stride == 1, (stride, "Python 2.5 only allows stride=1")
            #return self.content[self.start+start:self.start+end:stride]
            return self.content[self.start+start:self.start+end]
        else:
            i = xrange(self.start, self.end)[i]
            return self.content[i]

    def find(self, text, start=None, end=None):
        start, end, stride = slice(start, end).indices(self.end-self.start)
        return self.content.find(text, self.start+start, self.start+end) - self.start

    def get_subchunk(self, start=0, end=None):
        start, end, stride = slice(start, end).indices(self.end-self.start)
        if end < start:
            raise ValueError("negative width subchunks are not supported")
        return Chunk(None, self.content, self.start+start, self.start+end)

_tag_pattern = re.compile(br"[a-zA-Z][a-zA-Z0-9]{3}$")
def _is_valid_tag(tag):
    return bool(_tag_pattern.match(tag))

def _check_magic(magic):
    if magic != FPB_MAGIC:
        raise FPBError("Content does not start with the FPB signature (found %s)" % (myrepr(magic),))

def iter_chunks(content, check_magic=True):
    if check_magic:
        _check_magic(content[:8])
        offset = 8
    else:
        offset = 0

    max_size = len(content)

    while offset < max_size:
        chunk_size_str = content[offset:offset+8]
        if len(chunk_size_str) != 8:
            raise FPBError("Unexpected end of content at offset %d" % (offset,))
        
        tag = content[offset+8:offset+12]
        if not _is_valid_tag(tag):
            raise FPBError("Unexpected chunk tag %r at offset %d" % (tag, offset+8))
        
        chunk_size, = _unpack_chunk_size(chunk_size_str)
        if offset + 12 + chunk_size > max_size:
            raise FPBError("Chunk size of %d for %s (at offset %d) extends beyond the file size of %d"
                           % (chunk_size, tag, offset, max_size - 12 - offset))

        start = offset+12
        end = start + chunk_size
        yield Chunk(tag, content, start, end)
        offset = end


# Verify that the chunks look right, and return a dict mapping tag to Chunk

def _validate_chunk_order(chunks):
    named_chunks = {
        b"META": None,
        b"AREN": None,
        b"POPC": None,
        b"FPID": None,
        b"HASH": None,
        b"CFPL": None,
        }
    tag = "" # initialize tag so it's valid even if there are no chunks
    for chunk in chunks:
        tag = chunk.tag
        if tag == b"FEND":
            break
        if tag not in named_chunks:
            continue  # ignore
        value = named_chunks[tag]
        if value is not None:
            raise FPBError("Chunk %r occurs more than once, which isn't allowed" % (tag,))
        named_chunks[tag] = chunk
        
    if tag != b"FEND":
        raise FPBError("Missing required FEND chunk")

    for required_tag in (b"AREN", b"FPID"):
        if named_chunks[required_tag] is None:
            raise FPBError("The required %s chunk is missing" % (required_tag,))

    return named_chunks


######## Function and making the magic
    
FPB_MAGIC = b"FPB1\r\n\0\0"

def write_magic(output):
    output.write(FPB_MAGIC)



####### Read and write the "META" data chunk data #######

# The META chunk is identical to the header section of the FPS file.
# That includes the leading "#" for each line, and an optional
# leading '#FPS1'.

def write_metadata_chunk(output, metadata):
    # XXX Make a better API. I don't like going through str()
    f = BytesIO()
    io.write_fps1_header(f, metadata)
    write_chunk(output, b"META", f.getvalue())

def extract_metadata(chunk, arena_num_bytes):
    if chunk is None:
        # The META chunk is missing, which is okay because it's
        # optional in the FPB format.
        # It's easier in chemfp code to always have a .metadata
        # attribute, which I can do because the arena_num_bytes is
        # always known. So, synthesize one.
        metadata = Metadata(num_bytes=arena_num_bytes)
    else:
        # XXX Need to figure out how to handle error reporting
        location = io.Location("<META>")
        metadata, chunk, chunk_reader = fps_io.read_header(BytesIO(chunk[:]), location)
        if chunk is not None:
            raise FPBError("Extra data after final META configuration line: %r" % (chunk,))
    return metadata


####### Read and write the "AREN"a data chunk data #######

# The primary purpose of the FPB format is to make it possible to
# memory map fingerprints such that the fingerprints are aligned
# correctly for the high-speed chemfp popcount methods. (Eg, so that
# fingerprints are 64-bit/8-byte aligned for the POPCNT instruction.)
#
# The fingerprints are in a contiguous chunk. Each fingerprint takes
# 'num_bytes' bytes. In order to preserve alignment, extra NUL bytes
# may be added to the end of the fingerprint. The total fingerprint
# entry (fingerprint + extra bytes) is 'storage_size' bytes long.
#
# There is a spacer so the first fingerprint can be aligned correctly.
# This can be between 0 and 255 NUL bytes long, and the spacer size
# is stored in a single byte.
#
# If the fingerprints are ordered by popcount then the POPC chunk, if
# present, specifies the start and send of each popcount region.

# The 'AREN'a starts with a header:
#   <num_bytes: 4 bytes>  -- the number of bytes in a fingerprint
#   <storage_size: 4 bytes>  -- number of bytes in fingerprint + extra bytes
#   <spacer_size: 1 byte>   -- the number of spacer bytes used so the fingerprint
#          chunk starts on an aligned file position.
#   <spacer : $spacer_size> NUL bytes> -- up to 255 NUL bytes, used for alignment.

# The fingerprints are N fingerprint fields, ordered sequentially.
#   <fp0: $storage_size bytes> -- the first fingerprint
#   <fp1: $storage_size bytes> -- the second fingerprint
#      ...
# The last fingerprint ends at the last byte of the arena chunk.
#
# Each fingerprint contains:
#   <fingerprint: $num_bytes bytes> -- the actual fingerprint data
#   <extra: $storage_size-$num_bytes bytes> -- the 'extra' NULL padding bytes
#       used so storage_size is a multiple of the alignment.
#
# To get the number of fingerprints in the arena:
#    (len(arena content) - 4 - 4 - 1 - $spacer_size) // $storage_size


def _get_arena_storage_size(num_bytes, alignment):
    """Return the smallest number >= `num_bytes` which is a multiple of `alignment`

    This determines the storage_size for each fingerprint in the arena.
    """
    remainder = num_bytes % alignment
    if remainder == 0:
        return num_bytes
    return num_bytes + (alignment - remainder)

_pack_arena_spacer_size = struct.Struct("<B").pack

def _get_arena_alignment_spacer(position, alignment):
    """Return the smallest spacer where current position + len(spacer) is `alignment` aligned

    A spacer is a single length byte followed by that many NUL bytes.
    The smallest spacer is "\\0" and the largest is 256 bytes long.
    """
    remainder = (position+1) % alignment
    if remainder == 0:
        count = 0
    else:
        count = alignment - remainder
    if count >= 256:
        # "256 bytes ought to be enough for enough for anybody."
        raise ValueError("Alignment size %r is too large. The maximum supported alignment is 256 bytes")
    return _pack_arena_spacer_size(count) + b"\0" * count

_pack_arena_sizes = struct.Struct("<II").pack
_unpack_arena_sizes = struct.Struct("<II").unpack
_unpack_arena_fingerprint_size = struct.Struct("<I").unpack
_unpack_arena_storage_size = struct.Struct("<I").unpack

def _write_arena_header(output, spacer, num_fingerprints, num_bytes, storage_size):
    # Write the chunk size, 'AREN', num_bytes/fp, storage_size/fp, and spacer
    chunk_size = 4 + 4 + len(spacer) + num_fingerprints*storage_size
    #print "chunk size is", chunk_size, "num_bytes", num_bytes, "storage_size", storage_size, "num_fingerprints", num_fingerprints, "spacer", repr(spacer)
    chunk_size_bytes = _pack_chunk_size(chunk_size)
    output.write(chunk_size_bytes +
                 b"AREN" +
                 _pack_arena_sizes(num_bytes, storage_size))

# The easy interface, if you know everything up-front
def _write_arena_chunk(output, num_fingerprints, num_bytes, storage_size, alignment, arena,
                       start_padding, end_padding):
    start_offset = start_padding
    end_offset = len(arena) - end_padding
    if not (0 <= start_offset <= len(arena)):
        raise ValueError("start_offset must be in the arena fingerprint data")
    if not (start_offset <= end_offset <= len(arena)):
        raise ValueError("end_padding must be between the start padding and the "
                         "end of the arena fingerprint data")
    
    # Figure out what the spacer will be.
    # Chemfp 3.3 switched to 128 byte initial alignment.
    # With AVX2 on Skylake, 64 byte alignment was 7% faster than 8 byte alignment.
    # 128 byte alignment showed no difference, but perhaps it will on the
    # upcoming AVX 512. In any case, it doesn't really hurt.
    if num_bytes % 64 == 0:
        # use 128 byte alignment for 64-byte fingerprints and multiples thereof
        spacer_alignment = max(alignment, 128)
    elif num_bytes % 32 == 0:
        spacer_alignment = max(alignment, 32)
    else:
        spacer_alignment = alignment
    spacer = _get_arena_alignment_spacer(output.tell()+8+4+4+4, spacer_alignment)

    # Write the preamble
    _write_arena_header(output, spacer, num_fingerprints, num_bytes, storage_size)
    output.write(spacer)
    
    # Write the arena. Sadly, I can't just point to the substring to
    # copy so I have to copy and write a section at a time.
    start = start_offset
    while 1:
        end = start + 1024*10240  # 10 MB at a time
        if end > end_offset:
            end = end_offset
        output.write(arena[start:end])
        if end == end_offset:
            break
        start = end


# This writer uses the FPID writer and may also use the POPC and HASH writers
def _make_unordered_fpb_writer(output,
                               metadata,
                               alignment=8,
                               include_popc=True,  # Include the POPC section in the output?
                               include_hash=True,  # Include the HASH section in the output?
                               id_spool=None,      # Scratch space for the fpid writer ids
                               offset_spool=None,  # Scratch space for the fpid writer offsets
                               num_entries=None,   # Slight optimization if the size is known and correct.
                               close=None,
                               ):
    """Send (id, fp) pairs to 'output' as an FPB file

    THIS IS A PRIVATE API

    Creates a coroutine which writes an FPB file. It's used like this:
       writer = _make_unordered_fpb_writer(...)   # create generator
       next(writer)                               # start processing
       while data:
           writer.send(id_fp_pairs)               # send new id/fp tuples
       for _ in writer:                           # write any remaining data and close
           raise AssertionError                   # should not get here

    The (id, fp) pairs are written in input order to the FPB file. That
    is, they are not reordered by popcount. However, if the input is
    already in popcount order, and if `include_popc` is specified, then
    

    `output` must be a file-like object which support write(s), tell()
    and seek(where).
    `metadata` is a Metadata, and it must have a num_bytes. (This may be loosened.)
    alignment sets the on-disk alignment. It can be no larger than 256.
    include_popc and include_hash specify if the POPC and HASH chunks should be written.
       (the AREN and FPID fields are always written)
    id_spool and offset_spool are file-like objects used for temporary storage
       when writing the FPID chunk.
    If you know the 'num_entries' then there's one less seek.
    """
    
    # If no spools are given, use memory
    if id_spool is None:
        id_spool = BytesIO()
    if offset_spool is None:
        offset_spool = BytesIO()

    # Write the magic and the META chunks
    write_magic(output)

    #output = _Watch(output)  # Uncomment to see the file I/O

    num_bytes = metadata.num_bytes
    write_metadata_chunk(output, metadata)

    # Need to know this point to get the correct alignment.
    # Make also need to remember this position to seek to it later.
    chunk_start_where = output.tell()

    # Round up the num_bytes to the nearest multiple of alignment
    storage_size = _get_arena_storage_size(num_bytes, alignment)

    # Spacer goes after the num_bytes/spacer_size so that the
    # fingerprints with the right alignment.
    spacer = _get_arena_alignment_spacer(chunk_start_where+8+4+4+4, alignment)

    # Extra bytes added to the end of each fingerprint so it fits the
    # storage size.
    extra_bytes = b"\0" * (storage_size - num_bytes)
        
    # Write the start of the 'AREN'a chunk, either with correct data
    # (if I know the expected total number) of entries, or with a
    # dummy, to be replaced later.
    if num_entries is not None:
        # Since I know the information, might as well stick it in now.
        _write_arena_header(output, spacer, num_entries, num_bytes, storage_size)
    else:
        # Write a fake entry
        output.write(b"\xff\xff\xff\xff\xff\xff\xff\xffAREN\0\0\0\0\0\0\0\0")

    # Move forward so the fingerprints are properly aligned.
    output.write(spacer)

    ## Add fingerprints to the ARENa chunk. I do this first since that
    ## takes the most space. Send ids to the id writer (which may
    ## spool) and hash writer.  At the end, copy those to the main
    ## output.

    # Keep track of the number of fingerprints with a given popcount.
    # This is used for the POPC chunk.
    popcount_counts = [0] * (num_bytes*8+1)
    prev_popcount = 0
    is_sorted = bool(include_popc)  # Only include a POPC section if requested,
                                    # and if the inputs are in popcount order.

    # Start the text and hash writers.
    fpid_writer = _make_data_chunk_writer(output, b"FPID",
                                          data_spool=id_spool, offset_spool=offset_spool)
    next(fpid_writer)  # prime the column writer

    if include_hash:
        hash_writer = _make_hash_writer(output)
        next(hash_writer)  # prime the hash writer

    # Process the fingerprints
    num_fingerprints = 0
    byte_popcount = bitops.byte_popcount
    try:
        while 1:
            # Get a list of (id, fp) pairs, or 'None' to indicate the end of input.
            id_fp_pairs = (yield "next")
            if id_fp_pairs is None:
                break

            # It's about 25% faster to separate ids and fps in order to
            # send the ids as a batch to the text_writer and hash_writer,
            # compared to one-by-one. Probably because of the Python
            # function call overhead.
            ids, fps = zip(*id_fp_pairs)

            ids = [id.encode("utf8") for id in ids]

            # Ensure that the inputs have consistent length.
            for fp in fps:
                if len(fp) != num_bytes:
                    # Inconsistent length. Go back, figure out the index, and complain.
                    # (I optimized for the likely case where there are no errors.)
                    i = fps.index(fp) 
                    raise ValueError("Fingerprint for id %r is %d bytes long, expected %d bytes: %r"
                                     % (ids[i].decode("utf8"), len(fp), num_bytes, fp))

            # Save the fingerprints as soon as I get them.
            for fp in fps:
                output.write(fp+extra_bytes)
                if is_sorted:
                    # I can build a POPC so long as the input is sorted by popcount.
                    popcount = byte_popcount(fp)
                    if popcount >= prev_popcount:
                        popcount_counts[popcount] += 1
                        prev_popcount = popcount
                    else:
                        # Once it isn't sorted, I can't build a POPC
                        is_sorted = False
                        popcount_counts = None
                num_fingerprints += 1

            # Save the ids in the FPID chunk and (if requested) the HASH chunk.
            fpid_writer.send(ids)
            if include_hash:
                hash_writer.send(ids)

        # Finished getting id/fp pairs. Save the rest of the data.

        # Save the rest of the data.
        if num_entries != num_fingerprints:
            # Save the current location then jump back and correct the header sizes.
            chunk_end_where = output.tell()
            try:
                output.seek(chunk_start_where)
            except IOError as err:
                if num_entries is None:
                    msg = "Unable to seek in %r to set the number of fingerprints: %s"
                else:
                    msg = "Unable to seek in %r to update the number of fingerprints: %s"
                raise IOError(err.errno, msg % (output, err.strerror))

            _write_arena_header(output, spacer, num_fingerprints, num_bytes, storage_size)
            # ... and jump back to the end.
            output.seek(chunk_end_where)


        if is_sorted:  # is only true if include_popc is true
            # Save the 'POPC'ount section
            _write_popcount_counts(output, popcount_counts)

        # Close the text and hash writers. They will write chunks to output.
        for _ in fpid_writer:
            raise AssertionError

        if include_hash:
            for _ in hash_writer:
                raise AssertionError

        # Finish writing the file.
        write_fend(output)
        output.flush()
    finally:
        if close is not None:
            try:
                close()
            finally:
                close = None
        id_spool.close()
        offset_spool.close()


# Write to an ArenaBuilder until it's too large.
# Dump it to a file and switch to a new ArenaBuilder.
# When done, if there are no files, return the arena from the ArenaBuilder.
# Do a parallel merge of the files

def _make_ordered_writer(output, metadata, alignment=8, include_popc=True, include_hash=True,
                         max_spool_size=None, tmpdir=None, close=None):
    arena_writer = arena_module._make_arena_writer(metadata, reorder=True, alignment=alignment)
    next(arena_writer)
    if tmpdir is None:
        tmpdir = tempfile.gettempdir()
        
    total_size = 0
    fudge_factor = _getsizeof("") + 8
    save_files = []
    num_bytes = metadata.num_bytes
    assert num_bytes, num_bytes
    num_bits = metadata.num_bits
    assert num_bits, num_bits
    assert (num_bits+7)//8 == num_bytes, (num_bits, num_bytes)

    try:
        while 1:
            id_fp_pairs = (yield "next")
            if id_fp_pairs is None:
                # Polite request to end.
                break

            if max_spool_size is not None and total_size > max_spool_size:
                # Save to a temp file and restart the arena.
                tmpfile = tempfile.NamedTemporaryFile(prefix="fpserver-ordered-", suffix=".fpb", dir=tmpdir)
                arena = next(arena_writer)
                for _ in arena_writer:
                    raise AssertionError("should not happen")

                save_arena_as_fpb(tmpfile, arena, include_hash=False)
                save_files.append(tmpfile)
                arena_writer = arena_module._make_arena_writer(metadata, reorder=True, alignment=alignment)
                next(arena_writer)  # prime the pump
                total_size = 0
            # This is only approximate.
            # The utf8 string may be larger len(id).
            # However, that's probably not a big thing to worry about.
            sum_len_ids = arena_writer.send(id_fp_pairs)
            ##     for id, fp in id_fp_pairs:
            ##         total_size += len(id)
            total_size +=  (sum_len_ids + fudge_factor) * len(id_fp_pairs)

        # Finished processing the data files
        if not save_files:
            # Shortcut. There's only a single arena, in memory, so write it to a file.
            arena = next(arena_writer)
            for _ in arena_writer:
                raise AssertionError("should not happen")
            save_arena_as_fpb(output, arena, include_hash=include_hash)
            return

        #arenas = [open_fpb(tmpfile) for tmpfile in save_files]
        arenas = []
        for tmpfile in save_files:
            ##try:
            arena = open_fpb(tmpfile, "fpb", None)
            ## except AssertionError as err:
            ##     print tmpfile, err
            ##     tmpfile.seek(0)
            ##     with open("log.fpb", "wb") as f:
            ##         shutil.copyfileobj(tmpfile, f)
            ##     raw_input("Do something. ")
            ##     raise
            arenas.append(arena)
        arena = next(arena_writer)
        for _ in arena_writer:
            raise AssertionError("should not happen")
        if arena:
            arenas.append(arena)


        num_entries = sum(len(arena) for arena in arenas)
        write_magic(output)
        write_metadata_chunk(output, metadata)

        # Save the fingerprints into the ARENa chunk

        # All of the arenas have the same storage_size, and have ordered
        # fingerprints, so we can look into the internal data structure
        # and pull out all of the fingerprints for a given popcount.

        storage_size = arenas[0].storage_size  # Same for everyone
        chunk_start_position = output.tell()
        spacer = _get_arena_alignment_spacer(chunk_start_position+8+4+4+4, alignment)

        _write_arena_header(output, spacer, num_entries, num_bytes, storage_size)
        output.write(spacer)
        where = output.tell()
        if where % alignment != 0:
            raise AssertionError("Miscomputed the alignment. At %d with alignment %s" % (where, alignment))


        popcount_counts = [0] * (num_bytes*8+1)

        output.flush()   # Flush Python/C's FILE* so can pass a fileno() directly to write(2)
        for popcount in range(num_bits+1):
            for arena in arenas:
                popcount_bytes = arena.popcount_indices[popcount*4:popcount*4+8]
                start_popcount_index, end_popcount_index = struct.unpack("<II", popcount_bytes)
                num = end_popcount_index - start_popcount_index

                popcount_counts[popcount] += num

                # Copy all of the fingerprints with this popcount in this arena 
                start_offset = start_popcount_index*storage_size + arena.start_padding
                end_offset = end_popcount_index*storage_size + arena.start_padding
                _chemfp.write_substring(output.fileno(), arena.arena, start_offset, end_offset)
                # For large data sets, the above is about 10% faster than doing an
                # intermediate copy to a string, probably because of less swapping.
                # For small files the difference is not noticiable.
                #output.write(arena.arena[start_offset:end_offset])


        # Save the 'POPC'ount chunk
        if include_popc:
            _write_popcount_counts(output, popcount_counts)

        # Save the FPID chunk
        fpid_writer = _make_data_chunk_writer(output, b"FPID",
                                              data_spool=None, offset_spool=BytesIO())
        next(fpid_writer)
        hash_writer = _make_hash_writer(output)
        next(hash_writer)

        for popcount in range(num_bits+1):
            for arena in arenas:
                popcount_bytes = arena.popcount_indices[popcount*4:popcount*4+8]
                start_popcount_index, end_popcount_index = struct.unpack("<II", popcount_bytes)

                ids = arena.ids[start_popcount_index:end_popcount_index]
                ids = [id.encode("utf8") for id in ids]
                assert len(ids) == end_popcount_index-start_popcount_index
                fpid_writer.send(ids)

                if include_hash:
                    hash_writer.send(ids)

        for _ in fpid_writer:
            raise AssertionError("fpid_writer didn't close!")

        if include_hash:
            for _ in hash_writer:
                raise AssertionError("hash_writer didn't close!")

        # Finish writing the file.
        write_fend(output)
        output.flush()
    finally:
        if close is not None:
            try:
                close()
            finally:
                close = None

## Read the "AREN"a

def extract_arena(chunk):
    assert chunk.tag == b"AREN"
    num_bytes, storage_size = _unpack_arena_sizes(chunk[:8])
    spacer_num_bytes = _compat.indexbytes(chunk, 8)
    fp_start = 8+1+spacer_num_bytes
    assert fp_start <= chunk.end
    delta = len(chunk)-fp_start
    assert delta % storage_size == 0, (delta, storage_size, delta % storage_size)
    assert storage_size > 0

    # Figure out the best alignment, up to what chemfp can understand.
    alignment = 32  # Well more than chemfp requires.
    while alignment > storage_size:
        alignment //= 2
    while storage_size % alignment != 0:
        alignment //= 2
    while (chunk.start + fp_start) % alignment != 0:
        alignment //= 2

    return dict(num_bytes = num_bytes,
                storage_size = storage_size,
                fp_start = chunk.start + fp_start,
                fp_end = chunk.end,
                alignment = alignment,
                num_records = delta // storage_size,
                )



############ POPC chunk

_unpack_popc_offset = struct.Struct("<I").unpack

def _pack_popcount_offsets(offsets):
    #assert ((len(offsets)-2) % 8) == 0, ("there must be enough popcounts for a multiple of 8 bits",
    #                                   len(offsets))
    return struct.pack("<" + "I"*len(offsets), *offsets)

def _write_popc_chunk_from_ints(output, offsets):
    if len(offsets) < 2:
        raise ValueError("popcount offsets must contain at least two entries")
    write_chunk(output, b"POPC", _pack_popcount_offsets(offsets))


def _write_popcount_counts(output, popcount_counts):
    offsets = [0] * (len(popcount_counts)+1)
    total = 0
    for i, count in enumerate(popcount_counts):
        total += count
        offsets[i+1] = total
    _write_popc_chunk_from_ints(output, offsets) 

def _write_popc_chunk_from_string(output, popcount_indices):
    if len(popcount_indices) < 8:
        raise ValueError("popcount_indices must contain at least two entries")
    if len(popcount_indices) % 4 != 0:
        raise ValueError("popcount_indices string must be a multiple of 4 bytes")
    write_chunk(output, b"POPC", popcount_indices)


def extract_popc(chunk):
    if len(chunk) % 4 != 0:
        raise FPBError("POPC chunk size must be a multiple of 4 bytes: %d" % (len(chunk),))
    num_entries = len(chunk) // 4
    if num_entries < 9:
        raise FPBError("POPC must contain at least 9 offsets")

    max_num_bits = num_entries-2

    popc = chunk[:]
    offsets = struct.unpack("<" + "I"*num_entries, popc)
    if offsets[0] != 0:
        raise FPBError("First POPC entry must be 0")

    prev_offset = 0
    for popcount, offset in enumerate(offsets):
        if offset < prev_offset:
            raise FPBError("POPC order is incorrect at index %d (%d to %d)"
                           % (popcount, prev_offset, offset))
        prev_offset = offset

    return {
        "max_num_bits": max_num_bits,
        "last_index": offsets[-1],
        "popcount_indices": popc,
        }


############ FPID chunk

# The FPID chunk stores record identifiers, which are often smallish
# strings (~10-20 bytes) but may be longer InChI strings and IUPAC
# names (over 100 bytes). Actually, there is no size limit, but
# knowing the size helps understand the design rationale.
#
# This format is designed to handle more then 100 million records. If
# there are more than 50 bytes per record identifier then the section
# can be longer than 2**32 bytes long. The start and end positions for
# each identifier are stored in an offset table. An uncompressed table
# requires more than 4 bytes to access that range.
#
# However, most people will store under 5 million records with about
# 10 bytes per record, which is only 50 MB for the identifiers. A 4
# byte offset table takes an extra 20 MB while an 8 byte offset table
# takes 40 MB. I don't think the extra 20 MB of 0s is useful, while
# the complexity to handle two index tables is tricky but minor.

# Instead, I have two index tables. Once the identifier offset is too
# long for 4-bytes, I switch to using 8-bytes. It's still somewhat
# wasteful if you have very large data sets with small identifiers,
# but that's not something I expect to worry about in the next decade.
# (Namely, 300 million * 12 byte identifiers is still under 4GB, and
# only virtual libraries will likely be that large.)


# The actual layout inside of the chunk is:
#  <num_4byte_elements: 4 bytes> -- the number of 4 byte offsets.
#  <num_8byte_elements: 4 bytes> -- the number of 8 byte offsets
#  Note: the number of indicies is num_4byte_elements + num_8byte_elements + 1
#  because even with no elements there will be the initial '\0\0\0\0'.
#
#  <id 0> + NUL   -- the first string, with an added NUL terminator
#  <id 1> + NUL   -- the second string, with an added NUL terminator
#      ....
#  <id N> + NUL   -- the last string, with an added NUL terminator
#       
#  <offset 0: 4 bytes>    -- the offset relative to the start of <text 0>.
#                   (This always contains the 4 bytes "\0\0\0\0")
#     ...
#  <offset num_4byte_elements: 4 bytes>   -- the last offset stored in 4 bytes
#
#       (Note: This next section exists only when <num 8 byte offsets> > 0)
#  <offset num_4byte_elements+1: 8 bytes>    -- the first offset stored in 8 bytes
#     ...
#  <offset num_4byte_elements+num_8byte_elements: 8 bytes>   -- the last offset stored in 8 bytes

# To get the identifier for record at position P >= 0:
#   chunk_size = size of the chunk
#   num_4byte_elements = decode bytes[0:4] as uint32
#   num_8byte_elements = decode bytes[4:8] as uint32
#   if P >= num_4byte_elements + num_8byte_elements:
#       record does not exist
#   offset_start = chunk_size - num_4byte_elements*4 - num_8byte_elements*8
#   if P < num_4byte_elements:
#     start, end = decode bytes[offset_start:offset_start+8] as (uint32, uint32)
#   elif P == N4:
#     start, end = decode bytes[offset_start:offset_start+12] as (uint32, uint64)
#   else:
#     start, end = decode bytes[offset_start:offset_start+16] as (uint64, uint64)
#   id = bytes[start:end-1]

_4BYTE_CHUNK_SIZE = 2**32-1 - 16 # -8 for the two counts, -1 and -8 for safety

_pack_fpid_offset_counts = struct.Struct("<II").pack     # number of 4 and 8 byte counts
_unpack_fpid_offset_counts = struct.Struct("<II").unpack #

_pack_fpid_4byte_offset = struct.Struct("<I").pack       # an offset into the 4-byte section
_unpack_fpid_4byte_offset = struct.Struct("<I").unpack   # 

_pack_fpid_8byte_offset = struct.Struct("<Q").pack       # an offset into the 8-byte section
_unpack_fpid_8byte_offset = struct.Struct("<Q").unpack   # 

def _write_column_header(output, chunk_size, chunk_name, num_4byte_elements, num_8byte_elements):
    output.write(_pack_chunk_size(chunk_size) + chunk_name +
                 _pack_fpid_offset_counts(num_4byte_elements, num_8byte_elements))


def _make_data_chunk_writer(output, chunk_name, data_spool, offset_spool):
    """Send per-record data for column `chunk_name` to 'output', in data chunk format

    THIS IS A PRIVATE API

    If `data_spool` is None then row values are written immediately to
    output. Otherwise, the rows are written to data_spool as
    temporary storage and copied to `output` when this generator
    finishes.
    
    The row offsets are written to `offset_spool` as temporary storage
    and copied to `output` after all of the rows have been written.
    """
    assert len(chunk_name) == 4
    # XXX better validation here.
    
    # If data_spool is None then write directly to output.
    # Let 'input_data_spool' be the actual input parameter,
    # and 'data_spool' be the file where the ids are stored.
    input_data_spool = data_spool
    if data_spool is None:
        data_spool = output

    start_where = data_spool.tell()

    # Save a dummy header. When finished this will contain:
    #   chunk size (8 bytes)
    #   chunk tag <chunk_name>
    #   count of the number of 4-byte offsets (4 bytes)
    #   count of the number of 8-byte offsets (4 bytes)
    initial_signature = (b"\xFF\xFE\xFF\xFE\xFF\xEE\xAB\xCD" +
                         chunk_name +
                         b"\x01\x02\x03\x04\x05\x06\x07\x08")
    data_spool.write(initial_signature)

    offset = 8  # Current offset into the chunk
    num_4byte_elements = 0
    num_8byte_elements = 0

    
    # The start of the first row is directly after num_8byte_elements
    offset_start_where = offset_spool.tell()  # Will need to come back here later
    offset_spool.write(_pack_fpid_4byte_offset(offset))  # First offset MUST be 4 bytes long.

    while 1:
        # Get a list of terms, or 'None' to indicate the end of input.
        row_stream = yield "next"
        if row_stream is None:
            break

        # Assume the input is a list.
        #rows = list(row_stream)
        #assert isinstance(row_stream, (tuple, list)), row_stream
        rows = row_stream
        if not rows:
            # Sent an empty list
            continue
        data_spool.writelines(rows)
        if 0:
            # TODO: the following is a sketch of a way to make the code faster.
            # However, since I don't often use unordered arenas, I don't
            # see the point in spending the time to optimize it further.
            import bisect
            row_sizes = _chemfp.get_cumulative_sizes(rows, offset)
            split = bisect.bisect_left(row_sizes, _4BYTE_CHUNK_SIZE)
            if split == 0:
                offset4 = []
                offset8 = row_sizes
            elif split == len(row_sizes):
                offset4 = row_sizes
                offset8 = []
            else:
                offset4 = row_sizes[:split]
                offset8 = row_sizes[split:]
            offset = row_sizes[-1]
        else:
            row_sizes = map(len, rows)

            offset4 = []
            offset8 = []
            for n in row_sizes:
                # Tracks the *end* position
                offset += n

                if offset > _4BYTE_CHUNK_SIZE:
                    offset8.append(offset)
                else:
                    offset4.append(offset)
                
        if offset4:
            num_4byte_elements += len(offset4)
            offset_spool.write(struct.pack("<" + "I"*len(offset4), *offset4))
        if offset8:
            num_8byte_elements += len(offset8)
            offset_spool.write(struct.pack("<" + "Q"*len(offset8), *offset8))

    # size is the current offset plus the size of the offset_spool
    chunk_size = offset + (num_4byte_elements+1)*4 + num_8byte_elements*8

    # Finished reading rows.

    # Update the dummy chunk size and the 4-byte and 8-byte counts.

    if input_data_spool is None:
        # I've been writing the identifiers directly to output.
        # After I update the header I need to get back to this point
        # so I can append the offsets.
        end_where = output.tell()

    # Stick in the correct header sizes
    data_spool.seek(start_where)
    _write_column_header(data_spool, chunk_size, chunk_name, num_4byte_elements, num_8byte_elements)
    
    if input_data_spool is None:
        # I've been writing the identifiers directly to output.
        # Go back to the end of the identifiers.
        output.seek(end_where)
    else:
        # I've been writing the identifers to the input_data_spool.
        # I need to copy the spool to the actual output. This means
        # I need to get to the start of the spool before copying.
        data_spool.seek(start_where)
        shutil.copyfileobj(data_spool, output)
        
    # I've written the identifiers. Now copy over the offsets.
    offset_spool.seek(offset_start_where)
    shutil.copyfileobj(offset_spool, output)



# It's easy if you can go over the list of ids multiple times
_pack_fpid_10000_4byte_offsets = struct.Struct("<" + "I"*10000).pack
def write_fpid_chunk(output, utf8_ids):
    num_4byte_elements, num_8byte_elements, id_size = _chemfp.get_id_chunk_sizes(utf8_ids, _4BYTE_CHUNK_SIZE)
    offset_size = (num_4byte_elements+1) * 4 + num_8byte_elements * 8
    chunk_size = 4 + 4 + id_size + offset_size

    _write_column_header(output, chunk_size, b"FPID", num_4byte_elements, num_8byte_elements)
    if 1:
        # It's faster to concatenate than to do individual writes
        W = output.write
        for i in range(0, len(utf8_ids), 10000):
            W(b"".join(utf8_ids[i:i+10000]))
    else:
        # The above is about 4x faster than output.writelines().
        # Plus, the zstandard writer doesn't support writelines.
        output.writelines(utf8_ids)

    size = 4 + 4 + 0  # 4 for each count, and 0 for the lack of a header
    output.write(_pack_fpid_4byte_offset(size)) # First size is always 4 bytes

    if utf8_ids:
        if 1:
            # Using a C extension is about 20x faster than the old code
            buf, size = _chemfp.pack_4byte_offsets(size, utf8_ids, num_4byte_elements)
            output.write(buf)
            buf, size = _chemfp.pack_8byte_offsets(size, utf8_ids, num_4byte_elements)
            output.write(buf)
        else:
            # This is the old code
            sizes = []
            num_sizes = 0
            for id in itertools.islice(utf8_ids, 0, num_4byte_elements):
                size += len(id)
                sizes.append(size)
                num_sizes += 1
                if num_sizes == 10000:
                    output.write(_pack_fpid_10000_4byte_offsets(*sizes))
                    sizes = []
                    num_sizes = 0
            if num_sizes:
                output.write(struct.pack("<" + "I"*len(sizes), *sizes))

            for id in itertools.islice(utf8_ids, num_4byte_elements, len(utf8_ids)):
                size += len(id)
                output.write(_pack_fpid_8byte_offset(size))
                #print("Write 8", size, id)
                #print("Write 8", size, _pack_fpid_8byte_offset(size))

def extract_fpid_ids(chunk):
    num_4byte_records, num_8byte_records = _unpack_fpid_offset_counts(chunk[:8])
    min_chunk_size = (num_4byte_records+1)*4 +  (num_8byte_records+1)*8
    if len(chunk) < min_chunk_size:
        #print "sizes", len(chunk), min_chunk_size, num_4byte_records, num_8byte_records
        raise FPBError("%s chunk is too small for the given number of records: %d > %d"
                       % (repr(chunk.tag)[1:-1], min_chunk_size, len(chunk)))

    index_start = len(chunk) - 4*(num_4byte_records+1) - 8*num_8byte_records
    if num_8byte_records:
        last_data_end, = _unpack_fpid_8byte_offset(chunk[-8:])
    else:
        last_data_end, = _unpack_fpid_4byte_offset(chunk[-4:])
    
    assert last_data_end == index_start, (last_data_end, index_start, chunk[last_data_end:index_start])

    return IdLookup(content = chunk.content,
                    num_4byte_records = num_4byte_records,
                    num_8byte_records = num_8byte_records,
                    data_start = chunk.start,
                    index_start = index_start + chunk.start
                    )

# Look up fingerprint identifier strings in an FPID chunk.

# This implements an list-like API. It doesn't know about the djb_hash value.

class IdLookup(object):
    def __init__(self,
                 content, num_4byte_records, num_8byte_records, data_start, index_start):
        self.content = content
        self._num_records = num_4byte_records + num_8byte_records
        self.num_4byte_records = num_4byte_records
        self.num_8byte_records = num_8byte_records
        self.data_start = data_start
        self.index_start = index_start

    def __repr__(self):
        return "IdLookup(<%d bytes>, #4=%d, #8=%d, data_start=%d, index_start=%d)" % (
            len(self.content), self.num_4byte_records, self.num_8byte_records,
            self.data_start, self.index_start)

    def __len__(self):
        return self._num_records

    def __iter__(self):
        content = self.content
        data_start = self.data_start
        index_start = self.index_start
        num_4byte_records = self.num_4byte_records
        num_8byte_records = self.num_8byte_records
        get_record_data = _chemfp.get_record_data
        for i in xrange(self._num_records):
            yield get_record_data(i, num_4byte_records, num_8byte_records,
                                  content, data_start, index_start)
    
    def __getitem__(self, i):
        # This is a microoptimization.
        # isinstance() is slow and int is expected.
        if i.__class__ is not int and isinstance(i, slice):
            start, end, stride = i.indices(self._num_records)
            if stride != 1:
                raise ValueError("Only stride=1 is supported, not %r" % (stride,))

            # I do a slice by adjusting the index start and num_*records fields
            # then returning a new IdLookup using those values.

            # Remove from the start
            to_remove = start
            index_start = self.index_start
            num_4byte_records = self.num_4byte_records
            num_8byte_records = self.num_8byte_records

            if to_remove <= num_4byte_records:
                index_start += to_remove*4
                num_4byte_records -= to_remove
            else:
                index_start += num_4byte_records*4
                to_remove -= num_4byte_records
                
                num_4byte_records = 0
                index_start += to_remove*8
                num_8byte_records -= to_remove

            # Remove from the end
            to_remove = self._num_records - end
            if to_remove < num_8byte_records:
                num_8byte_records -= to_remove
            else:
                to_remove -= num_8byte_records
                num_8byte_records = 0

            if to_remove:
                num_4byte_records -= to_remove
            return IdLookup(self.content, num_4byte_records, num_8byte_records, self.data_start, index_start)

        ## print "QQQ", i, self.num_4byte_records, self.num_8byte_records
        ## s = _chemfp.get_text_record(i, self.num_4byte_records, self.num_8byte_records,
        ##                                self.content, self.data_start, self.index_start)
        ## print "return", repr(s)
        ## return s
        if 0:
            # Used for debugging
            if i < self.num_4byte_records:
                index_start = self.index_start + i*4
                index_bytes = self.content[index_start:index_start+8]
                start, end = struct.unpack("<II", index_bytes)
            elif i == self.num_4byte_records:
                index_start = self.index_start + i*4
                index_bytes = self.content[index_start:index_start+12]
                start, end = struct.unpack("<IQ", index_bytes)
            else:
                index_start = self.index_start + self.num_4byte_records*4 - 4
                index_start += (i - self.num_4byte_records) * 8
                index_bytes = self.content[index_start:index_start+16]
                start, end = struct.unpack("<QQ", index_bytes)
                
            record_bytes = self.content[self.data_start+start:self.data_start+end]
            
            if len(record_bytes) < 100:
                print("record bytes [%d](%d:%d):" % (i, start, end), repr(record_bytes))
            else:
                print("first and last record bytes [%d](%d:%d):" % (i, start, end),
                          repr(record_bytes[:30]) + " ... " + repr(record_bytes[-30:]))
            return record_bytes.decode("utf8")
        else:
            return _chemfp.get_record_data(i, self.num_4byte_records, self.num_8byte_records,
                                           self.content, self.data_start, self.index_start)





############ hash chunk

# The HASH chunk, along with the FPID chunk, is an associative array
# which maps a record identifier to a record index. More specifically,
# the HASH chunk stores the mapping from hash values to the identifier
# index. Since multiple identifiers may have the same hash, the FPID
# chunk is used to make the index really is for the query
# identifier. (This assumes that you have unique identifiers).

# The implementation uses a table of 256 hash tables, indexed by part
# of the query hash value. Each hash tables uses open adressing with
# linear probing and a load factor of 0.5.

# The format is heavily influenced by the "cdb" data format of Daniel
# J. Bernstein, described in:
#   http://cr.yp.to/cdb.html
#   http://en.wikipedia.org/wiki/Cdb_(software)
#   http://www.unixuser.org/~euske/doc/cdbinternals/index.html
#
# I use the djb hash function, and I've changed the format so that it
# maps a hash to the id offset.

# In more detail, the data layout contains a main table and 256 hash
# tables. The main table contains 256 entries, where entry 'i'
# contains information about hash table 'i'.

# Entry i contains 8 bytes. The first 4 bytes (P[i]) is the unsigned
# integer offset to the start of hash table i, relative to the end of
# the main table. The second 4 bytes (E[i]) is the unsigned integer
# count of the number of slots in hash table i. E[i] is always twice
# the number of entries in the hash table.

# Hash table i contains E[i] slots. Each slot has 8 bytes containing 4
# bytes for the uint32 hash value followed by 4 bytes for the uint32
# identifier index. Empty slots are represented with
# "\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF". A non-empty slots is also called a 'hash
# pair.'

# There is an upper limit to the hash table. The empty slot notation
# means that offset 2**32-1 is not allowed, but that's okay since the
# byte offset P[i] sets a limit of
#    (2**32-256*8) / (2 * 8) = 268,435,328
# records, which is well over twice the FPB format's design goal.
#
# (Note: this isn't quite right. If the last hash table starts at
# position 2**32-2 then it can continue beyond 2**32. With an evenly
# distributed hash, there's an extra million records before problems.

# It's of course possible to develop a hash format which handles
# a much higher number of entries. However, it's a lot of work.
# The current method suports 2.5x more than the design goal and 100x
# more than most people use, and there are major issues to make
# the current scheme handle more than 20x its current limit.
#
# Some of the issues to address in a new hash format are:
#  - The current 4 bytes for the id index handles only 2**32 entries.
#      Is there need for more than ~4 billion entries?
#  - If 'yes', then is the choice of 256 hash tables still appropriate?
#      What about the choice of a cdb-like hash table?
#  - The 4th byte of the hash is used for both the hash table index
#     and in the hash comparison, so only ~16.8 million unique hash
#     values are available for the hash chain. With under 4 million
#     records in a hash table the average probe length is about 1.75.
#     Between 4 and 10 million records, the performance is still
#     decent, with an average probe length of under 2.0. Then it
#     gets worse. Quickly.
#
#          number of    average
#           entries   probe length
#          =========  ============
#          10000000      2.1
#          11000000      2.5
#          12000000      3.0
#          13000000      3.7
#          14000000      5.0
#          15000000      7.6
#          16000000     17.
#          16100000     20.
#          16200000     23.
#          16300000     28.
#          16400000     35.
#          16500000     48.
#          16600000     75.
#          16700000    170
#
# This shows a performance impact starting at about 256*12000000 =
# 3,000,000,000 total records and a serious impact by 256*16000000 =
# 4,000,000,000 total records. At this point, building a hash table,
# in the current implementation, starts taking quadratic time, so
# that sets a pretty hard practicable limit.
#
# There are ways to fix all of these, but they require extensive
# changes which don't seem warranted now.

# Here's a more graphical depiction of the layout.

#  ---------------+-----------------------+------------------------
#  Main table     | entry[0]              |  P[0]    4 bytes  (= 0)
#                 |                       |  E[0]    4 bytes
#                 | entry[1]              |  P[1]    4 bytes  (= E[0])
#                 |                       |  E[1]    4 bytes
#                 | entry[2]              |  P[2]    4 bytes  (= E[0]+E[1])
#                 |                       |  E[2]    4 bytes
#        ...         ...                      ...
#                 | entry[255]            |  P[255]  4 bytes  (= E[0]+E[1]+...+E[254])
#                 |                       |  E[255]  4 bytes
#  ---------------+-----------------------+---------------------- <--- offset P[0]
#  hash table   0 | slot[0][0]            | hash[0][0]          4 bytes  
#                 |                       | id_index[0][0]      4 bytes
#                 | slot[0][1]            | hash[0][0]          4 bytes
#                 |                       | id_index[0][1]      4 bytes
#        ...          ...                     ...
#                 | slot[0][E[0]-1]       | hash[0][E[0]-1]     4 bytes
#                 |                       | id_index[0][E[0]-1] 4 bytes
#  ---------------+-----------------------+---------------------- <--- offset P[1]
#  hash table   1 | slot[1][0]            | hash[1][0]          4 bytes
#                 |                       | id_index[1][0]      4 bytes
#        ...          ...                     ...
#                 | slot[1][E[1]-1]       | hash[1][E[1]-1]     4 bytes
#                 |                       | id_index[1][E[1]-1] 4 bytes
#  ---------------+-----------------------+--------------
#        ...
#  ---------------+-----------------------+---------------------- <--- offset P[255]
#  hash table 255 | slot[255][0]          | hash[255][0]            4 bytes
#                 |                       | id_index[255][0]        4 bytes
#        ...          ...                     ...
#                 | slot[255][E[1]-1]     | hash[255][E[255]-1]     4 bytes
#                 |                       | id_index[255][E[255]-1] 4 bytes
#  ---------------+-----------------------+----------------------


# To find id index candidates for a given query identifier:
#
#   1) compute hash = djb_hash(query_identifier)
#   2) compute hash_table_index = hash % 256
#   3) get P = P[hash_table_index] and E = E[hash_table_index]
#   4) get the subtable starting at (256*8)+P, which is E*8 bytes long
#   5) compute the initial probe = hash % E
#   6) while slot[hash_table_index][probe] is not empty:
#     6a) if hash[hash_table_index][probe] == hash
#             then id_index[hash_table_index][probe] is a candidate
#     6b) probe = (probe + 1) % E
#     6c) go to 6

# To check if a given id index candidate actually matches the query
# identifier, use the FPID chunk to look up the record from its index,
# and check if the two strings match.


# These are the (P[i], E[i]) pairs in the main table. E[i] is a multiple of 2.
_hash__struct_II = struct.Struct("<II")
_pack_hash_main_table_entry = _hash__struct_II.pack
_unpack_hash_main_table_entry = _hash__struct_II.unpack

# These are the (hash[i], index[i]) hash pairs in a hash table slot
_pack_hash_hashpair = _hash__struct_II.pack
_unpack_hash_hashpair = _hash__struct_II.unpack

_is_python2 = (sys.version_info.major == 2)

def _make_hash_writer(output):
    """Send fingerprint identifier hashes to 'output' as a HASH chunk

    THIS IS A PRIVATE API

    The HASH chunk is written to output only all identifiers have been
    received.
    """
    # There are ways to spool the data, but it's more complicated than
    # I want to deal with. This is okay for now because in I expect
    # each record will take about 8*(17/16) bytes in the hash_chain
    # arrays and, assuming an equi-distributed hash function, about:
    #   (1 / 256 hash values) * (72 bytes for 8 character string + 8 bytes for two slots) * 9/8
    #   => 45/128  = ~ 0.35 bytes per entry in the hash chain output.
    #
    # (The 17/16 and 9/8 are the geometic overallocation strategies
    # for array and list types, to give amoratized linear growth.)
    # 
    # Thus, in the worst case of 60 million records, this uses
    # about 60 million * 8.85 bytes => 530 MB, using base-10 MB.
    # This is acceptable.

    # Store the (hash, term_index) as alternating integers.
    # Use an array in order to minimize memory use. (~8*(17/16) bytes per entry)
    hash_chains = [array.array("I", ()) for i in range(256)]

    term_index = 0
    djb_hash = _djb_hash
    while 1:
        # Get a list of terms, or 'None' to indicate the end of input.
        #print("Sizes:", list(map(len, hash_chains)))
        term_stream = (yield "next")
        if term_stream is None:
            break

        # Store the hash value and its index in the correct hash chain.
        for term in term_stream:
            hash = djb_hash(term)
            hash_chain_index = hash % 256
            hash_chains[hash_chain_index].extend((hash, term_index))
            term_index += 1

    # Variable renaming since 'num_terms' makes more sense at this point.
    num_terms = term_index

    # I know how large the output will be, so write that chunk header now.
    size = 256*8 + num_terms*2*8
    output.write(_pack_chunk_size(size) + b"HASH")

    # Write the main table of 256 entries. 
    # These contain offsets to the (hash, term_index) hash chains
    
    hash_chain_start_offset = 0
    for hash_chain in hash_chains:
        # The number of entries in the hash chain = len(hash_chain)//2.
        # The number of slots for a hash chain is E => len(hash_chain)
        E = len(hash_chain)
        output.write(_pack_hash_main_table_entry(hash_chain_start_offset, E))
        # Each slot contains a hash-part of (hash, term_index), both 4 bytes long.
        hash_chain_start_offset += E*8

    # Write each individual hash chain

    # I tried various pure-Python implementations, but they were
    # too slow. The fastest Python version on 5.5 million
    # identifiers took about 13 seconds, and 0.6 seconds in C, so
    # about 20x faster. Even with 28.5 million records the C code
    # takes only about 4 seconds.

    # XXX TODO: check if this is still a problem with the BytesIO
    # Grrr! cStringIO objects don't accept a ctypes buffer. I don't know why.
    # I get a message like:
    #    argument 1 must be string or read-only character buffer, not c_long_Array_5
    # Also, there's a bug in StringIO; it uses str() to convert to the output string.
    # So, use buffer() to get the correct bytes, then str() to convert to a real string.
    # Ugly, but it works.
    for hash_chain in hash_chains:
        if not hash_chain:
            continue
        E = len(hash_chain)
        # XXX Doesn't this assume that c_int is 4 bytes?
        # Not using create_string_buffer() because I want to be sure it's int aligned.
        # (Is that really a problem?)
        slots = (ctypes.c_int*(E*2))()
        #slots = ctypes.create_string_buffer(E*8)  # 4 byte integers * 50% occupancy
        _chemfp.assign_hash_slots(slots, hash_chain)
        if _is_python2:
            # This is the best solution I could figure out.
            # If I write the slots like I do with Python 3 then
            # I get "<chemfp.fpb_io.c_int_Array_4 object at ...>"
            output.write(str(buffer(slots)))
        else:
            output.write(slots)
    
        
# This implements a subset of the dict API.
# Not yet meant for public use.
class HashLookup(object):
    def __init__(self, hash_table, values):
        self.hash_table = hash_table
        self.values = values

    def __len__(self):
        return len(self.values)

    def get_index_by_id(self, id):
        hash = _djb_hash(id)
        id0 = id + "\0"
        for index in self.hash_table.get_matching_indices(hash):
            value = self.values[index]
            if id == value or value.startswith(id0):
                return index
        return None
    
    def get(self, id, default=None):
        hash = _djb_hash(id)
        id0 = id + "\0"
        for index in self.hash_table.get_matching_indices(hash):
            value = self.values[index]
            if id == value or value.startswith(id0):
                return value
        return default

    def __getitem__(self, id):
        hash = _djb_hash(id)
        id0 = id + "\0"
        for index in self.hash_table.get_matching_indices(hash):
            value = self.values[index]
            if id == value or value.startswith(id0):
                return value
        raise KeyError(id)
        
    def __contains__(self, id):
        hash = _djb_hash(id)
        id0 = id + "\0"
        for index in self.hash_table.get_matching_indices(hash):
            value = self.values[index]
            if id == value or value.startswith(id0):
                return True
        return False

    def iteritems(self):
        # By construction, this is the part of the data up to the
        # first "\0", if present.  This is a bit tricky because I must
        # exclude duplicate keys.
        seen = set()
        current_hash = None
        for value in self.values:
            id = value.partition("\0")[0]
            h = _djb_hash(id)
            if h == current_hash:
                if id in seen:
                    continue
                seen.add(id)
            else:
                yield id, value
                seen.clear()
                seen.add(id)


    def __iter__(self):
        for id, value in self.iteritems():
            yield id
            
    def iterkeys(self):
        for id, value in self.iteritems():
            yield id

    def itervalues(self):
        for id, value in self.iteritems():
            yield value

    def iter_all_keys(self):
        for value in self.values:
            yield value.partition("\0")[0]

    def iter_all_values(self):
        return iter(self.values)

    def iter_all_items(self):
        for value in self.values:
            yield value.partition("\0")[0], value
        
    
            
class HashTable(object):
    __slots__ = ["chain_start_offsets", "chain_sizes", "hash_table_data"]
    def __init__(self, chain_start_offsets, chain_sizes, hash_table_data):
        self.chain_start_offsets = chain_start_offsets
        self.chain_sizes = chain_sizes
        self.hash_table_data = hash_table_data

    def get_matching_indices(self, hash):
        #print("hash", hash)
        i = hash & 255
        chain_size = self.chain_sizes[i]
        #print("chain size", chain_size)
        if chain_size == 0:
            return
        
        start_offsets = self.chain_start_offsets[i]
        initial_slot = slot = (hash>>8) % chain_size
        #print("Start lookup", i, start_offsets, chain_size, slot)

        hash_table_data = self.hash_table_data
        while 1:
            s = hash_table_data[start_offsets+slot*8:start_offsets+slot*8+8]
            #print("check slot", i, slot, repr(s))
            assert len(s) == 8
            if s == b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF":
                return
            new_hash, target_index = struct.unpack("<II", s)
            #print("       ->", new_hash, target_index)
            if new_hash == hash:
                yield target_index
            slot = (slot + 1) % chain_size
            if slot == initial_slot:
                raise FPBError("Chain terminator not found for hash %r" % (hash,))

        

def extract_hash_table(chunk):
    if len(chunk) < 8*256:
        raise FPBError("HASH chunk doesn't have enough space for the offset table")
    
    start_positions = []
    chain_sizes = []
    total_size = 0
    N = len(chunk) - 8*256

    for i in range(256):
        start_position, chain_size = _unpack_hash_main_table_entry(chunk[i*8:(i+1)*8])
        ## if chain_size:
        ##     print "Check", i, start_position, chain_size, N
        assert start_position <= N, "bad offset"
        assert chain_size % 2 == 0, "must be a muliple of 2"
        assert start_position == total_size*8
        total_size += chain_size
        start_positions.append(start_position)
        chain_sizes.append(chain_size)

    return HashTable(start_positions, chain_sizes, chunk.get_subchunk(256*8, None))

############ chemfp license key

def extract_cfpl(cfpl):
    assert cfpl.tag == b"CFPL"
    license_key = cfpl[:16]
    if len(license_key) != 16:
        return b""
    return license_key


############ The FEND chunk

def write_fend(output):
    write_chunk(output, b"FEND", b"")


############  helper function

# no metadata. Write just the basics and quit
def _write_empty_fpb_file(output):
    write_magic(output)
    _write_arena_chunk(output, 0, 1, 1, 1, b"", 0, 0)
    write_fpid_chunk(output, [])
    write_fend(output)
    output.flush()
    

############  Public API

# Note: the constructors are not yet part of the public API.
# These should only be created through chemfp.open_fingerprint_writer().


class InputOrderFPBWriter(FingerprintWriter):
    """Fingerprint writer for FPB files which preserves the input fingerprint order

    This is a subclass of :class:`chemfp.FingerprintWriter`.

    Instances have the following public attributes:

    .. py:attribute:: metadata

       a :class:`chemfp.Metadata` instance
       
    .. py:attribute:: format

       the string 'fpb'
       
    .. py:attribute:: closed

       False when the file is open, else True

    Other attributes (like "alignment", "include_hash", "include_popc",
    "max_spool_size", and "tmpdir") are undocumented and subject
    to change in the future. Let me know if they are useful.

    An InputOrderFPBWriter is also is own context manager, and will
    close the writer on context exit.
    """
    format = "fpb"
    _writer = None
    def __init__(self, output, metadata=None, alignment=8, include_popc=True,
                 include_hash=True, max_spool_size=None, num_entries=None, tmpdir=None, close=None):
        """(the constructor is not part of the public API; use open_fingerprint_writer() instead)"""
        self.output = output
        self.metadata = metadata
        self.alignment = alignment
        self.include_popc = include_popc
        self.include_hash = include_hash
        self.max_spool_size = max_spool_size
        self.num_entries = num_entries
        if tmpdir is None:
            tmpdir = tempfile.gettempdir()
        self.tmpdir = tmpdir
        self._close = close
        
        if metadata is not None:
            self._num_bytes = metadata.num_bytes
        else:
            self._num_bytes = None
        self._id_spool = None
        self._offset_spool = None

        self.closed = False
        if self._num_bytes is not None:
            self._start_writer()

    def _start_writer(self):
        if self.max_spool_size is None:
            id_spool = BytesIO()
            offset_spool = BytesIO()
        else:
            id_spool = SpooledTemporaryFile(max_size = self.max_spool_size,
                                            prefix="fpserver-fpid-text", dir=self.tmpdir)
            offset_spool = SpooledTemporaryFile(max_size = self.max_spool_size,
                                                prefix="fpserver-fpid-offset", dir=self.tmpdir)
        #id_spool = _Watch(id_spool)
        #offset_spool = _Watch(offset_spool)
        self._id_spool = id_spool
        self._offset_spool = offset_spool
        metadata = self.metadata
        if metadata is None:
            metadata = Metadata(num_bytes=self._num_bytes)
        writer = _make_unordered_fpb_writer(self.output, metadata=metadata, alignment=self.alignment,
                                            include_popc=self.include_popc, include_hash=self.include_hash,
                                            id_spool=id_spool, offset_spool=offset_spool,
                                            num_entries=self.num_entries, close=self._close)
        # Prime the pump
        q = next(writer)
        assert q == "next"
        self._writer = writer

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def write_fingerprint(self, id, fp):
        """Write a single fingerprint record with the given id and fp to the destination

        :param string id: the record identifier
        :param bytes fp: the fingerprint
        """
        if self.closed:
            raise ValueError("Cannot write a fingerprint after calling close()")
        if self._num_bytes is None:
            self._num_bytes = len(fp)
            self._start_writer()
        self._writer.send( [(id, fp)] )

    def write_fingerprints(self, id_fp_iter):
        """Write a sequence of (id, fingerprint) pairs to the destination

        :param id_fp_pairs: An iterable of (id, fingerprint) pairs.
        """
        if self.closed:
            raise ValueError("Cannot write fingerprints after calling close()")
        writer = self._writer

        # Make sure it's an iterator, which I'll consume with islice.
        id_fp_iter = iter(id_fp_iter)
        
        while 1:
            # Feeding the iterator directly, one-by-one: 19.9 seconds
            # Chunks of size      1: 18.3
            # Chunks of size      3: 13.6
            # Chunks of size     10: 11.9
            # Chunks of size     30: 11.1
            # Chunks of size    100: 10.6
            # Chunks of size    300: 10.4
            # Chunks of size   1000: 10.4  <--- Decided on this one
            # Chunks of size   3000: 10.7
            # Chunks of size  10000: 10.8
            # Chunks of size  30000: 10.8
            # Chunks of size 100000: 11.5

            # Read the next 1000 records, as a list
            sub_id_fp_iter = list(itertools.islice(id_fp_iter, 1000))
            if not sub_id_fp_iter:
                break

            # I allow FPB files with no metadata, but that means I
            # don't know the expected number of bytes until I read
            # the first fingerprint.
            if self._num_bytes is None:
                self._num_bytes = len(sub_id_fp_iter[0][1])
                self._start_writer()
                writer = self._writer
            writer.send(sub_id_fp_iter)

    def close(self):
        """Close the output writer

        This will set self.closed to False
        """
        if self.closed:
            return
        self.closed = True
        writer = self._writer
        if writer is not None:
            try:
                self._writer.send(None)
                raise AssertionError
            except StopIteration:
                pass
            self._writer = None
        else:
            assert self._num_bytes is None
            _write_empty_fpb_file(self.output)
            if self._close is not None:
                self._close()
        self._close = None
    
class OrderedFPBWriter(FingerprintWriter):
    """Fingerprint writer for FPB files where the input fingerprint order is preserved

    This is a subclass of :class:`chemfp.FingerprintWriter`.

    Instances have the following public attributes:

    .. py:attribute:: metadata

       a :class:`chemfp.Metadata` instance
       
    .. py:attribute:: format

       the string 'fpb'
       
    .. py:attribute:: closed

       False when the file is open, else True

    Other attributes (like "alignment", "include_hash", "include_popc",
    "max_spool_size", and "tmpdir") are undocumented and subject
    to change in the future. Let me know if they are useful.

    An OrderedFPBWriter is also is own context manager, and will
    close the writer on context exit.
    """
    format = "fpb"
    _writer = None
    def __init__(self, output, metadata=None, alignment=8, include_popc=True, 
                 include_hash=True, max_spool_size=None, tmpdir=None, close=None):
        """(the constructor is not part of the public API; use open_fingerprint_writer() instead)"""
        self.output = output
        self.metadata = metadata
        self.alignment = alignment
        self.include_hash = include_hash
        self.include_popc = include_popc
        self.max_spool_size = max_spool_size
        if tmpdir is None:
            tmpdir = tempfile.gettempdir()
        self.tmpdir = tmpdir
        self._close = close

        if metadata is not None:
            self._num_bytes = metadata.num_bytes
        else:
            self._num_bytes = None

        self.closed = False
        if self._num_bytes is not None:
            self._start_writer()

    def _start_writer(self):
        metadata = self.metadata
        if metadata is None:
            metadata = Metadata(num_bytes=self._num_bytes)
        
        writer = _make_ordered_writer(self.output, metadata, alignment=self.alignment,
                                      include_popc=self.include_popc, include_hash=self.include_hash,
                                      max_spool_size=self.max_spool_size, tmpdir=self.tmpdir,
                                      close=self._close)
        q = next(writer)
        assert q == "next"
        self._writer = writer

    def write_fingerprint(self, id, fp):
        """Write a single fingerprint record with the given id and fp to the destination

        :param string id: the record identifier
        :param bytes fp: the fingerprint
        """
        if self.closed:
            raise ValueError("Cannot write a fingerprint after calling close()")
        if self._num_bytes is None:
            self._num_bytes = len(fp)
            self._start_writer()
        self._writer.send( [(id, fp)] )

    def write_fingerprints(self, id_fp_iter):
        """Write a sequence of (id, fingerprint) pairs to the destination

        :param id_fp_pairs: An iterable of (id, fingerprint) pairs.
        """
        if self.closed:
            raise ValueError("Cannot write fingerprints after calling close()")
        writer = self._writer

        id_fp_iter = iter(id_fp_iter)

        while 1:
            sub_id_fp_iter = list(itertools.islice(id_fp_iter, 1000))
            if not sub_id_fp_iter:
                break

            if self._num_bytes is None:
                self._num_bytes = len(sub_id_fp_iter[0][1])
                self._start_writer()
                writer = self._writer
            writer.send(sub_id_fp_iter)

    #def abort(self):
    #    self.writer.throw(StopIteration) # Perhaps this would be nice in the future? XXX

    def close(self):
        """Close the output writer"""
        if self.closed:
            return
        self.closed = True

        writer = self._writer
        if writer is not None:
            try:
                self._writer.send(None)
                raise AssertionError
            except StopIteration:
                pass
            self._writer = None
        else:
            assert self._num_bytes is None
            _write_empty_fpb_file(self.output)
            if self._close is not None:
                self._close()
        self._close = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

# This fixes an unfortunate interaction in a mmap'ed FPB arena between
# the close() and copy() methods. The new copy might contain the
# original memory map, or a slice of it. If one of {the original, the
# copy} is closed, how does it know to close the memory map and the
# underlying file?

# The CountedCloseMMap instance knows how to close the mmap and source
# file. It also tracks a reference count of the number of open
# references. When that count goes to 0, it closes everything.

# Originally I didn't track this at all, and let Python's garbage
# collector do it for me. For the 3.2 release I decided to get rid of
# ResourceWarning messages. Wow, chemfp really depended on the gc to
# close a lot of files.

class CountedCloseMMap(object):
    def __init__(self, source_close, content_close):
        self._counter = 1
        self._source_close = source_close
        self._content_close = content_close

    def _close(self):
        if self._source_close is not None:
            self._source_close()
            self._source_close = None
        if self._content_close is not None:
            self._content_close()
            self._content_close = None
        
    def inc_ref(self):
        i = self._counter
        if i > 0:
            self._counter = i + 1

    def dec_ref(self):
        i = self._counter
        if i > 0:
            self._counter = i = i-1
            if i == 0:
                self._close()

    def __call__(self):
        self.dec_ref()
        

class FPBFingerprintArena(arena_module.FingerprintArena):
    def __init__(self, metadata, alignment,
                 start_padding, end_padding, storage_size, arena,
                 popcount_indices, arena_ids, start=0, end=None,
                 id_lookup=None, close=None, license_key=None
                 ):
        super(FPBFingerprintArena, self).__init__(metadata, alignment,
                                                  start_padding, end_padding, storage_size, arena,
                                                  popcount_indices, arena_ids, start, end, id_lookup,
                                                  license_key=license_key)
        self._close = close
        self.closed = False

    def _new_copy(self, metadata, alignment, start_padding, end_padding, storage_size, arena,
             popcount_indices, arena_ids, start=0, end=None, id_lookup=None,
             num_bits=None, num_bytes=None):
        self._close.inc_ref()
        return FPBFingerprintArena(
            metadata, alignment, start_padding, end_padding, storage_size, arena,
            popcount_indices, arena_ids, start, end,
            id_lookup, close=self._close, license_key=self._license_key)

    # Both a slice and a copy need to share the reference count to close the mmap.
    _new_slice = _new_copy
        
    def close(self):
        """Close any resources associated with this arena

        More specifically, close the memory-map connection to the FPB file.
        """
        if self.closed:
            return
        self.closed = True
        if self._close is not None:
            self._close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


### End of public API

# Helper function to save an arena directly as an FPB file
def save_arena_as_fpb(output, arena, include_popc=True, include_hash=True, tmpdir=None):
    write_magic(output)
    write_metadata_chunk(output, arena.metadata)
    _write_arena_chunk(output, len(arena), arena.metadata.num_bytes, arena.storage_size,
                       arena.alignment, arena.arena, arena.start_padding, arena.end_padding)
    if include_popc and arena.popcount_indices:
        _write_popc_chunk_from_string(output, arena.popcount_indices)

    utf8_ids = [id.encode("utf8") for id in arena.ids]
    write_fpid_chunk(output, utf8_ids)

    if include_hash:
        hash_writer = _make_hash_writer(output)
        next(hash_writer)
        hash_writer.send(utf8_ids)
        for x in hash_writer:
            raise AssertionError("should not get here")

    write_fend(output)
    output.flush()


##

# Open for reading
def open_fpb(source, format_name, compression, allow_mmap=True):
    source_close = None
    content_close = None

    is_stdin = source is None
        
    source, source_close = io.open_compressed_input(source, compression, blame_name=None)

    try:
        # Open the file object.
        must_read = False
        if is_stdin or compression or (not allow_mmap):
            must_read = True
        else:
            try:
                source.fileno()
                source.seek(0, 0)
            except (AttributeError, stdlib_io.UnsupportedOperation):
                must_read = True
        if must_read:
            # I can't memory map this, so load into memory.
            # (Could be a BytesIO(), stdin, or compressed file.)
            s = source.read(0)
            if isinstance(s, type(u"")):
                _compat.raise_tb(ValueError("The file-like 'source' object must be opened in binary mode, not text"), None)

            # Do a quick check before potentially reading a large non-FPB file
            try:
                magic = source.read(8)
            except EOFError:
                magic = b""
            _check_magic(magic)
            content = source.read()
            size = len(content)
            check_magic = False
        else:
            # Get the file size (and also make sure the file is seekable and big enough)
            # the make it a memory-mapped file.
            source.seek(0, 2)
            size = source.tell()
            if size < 32:
                raise FPBError("The file is too small to be a valid FPB file")
            try:
                content = mmap.mmap(source.fileno(), size, access=mmap.ACCESS_READ)
            except OverflowError as err:
                raise OverflowError("This version of Python is unable to mmap a %d byte file: %s"
                                    % (size, err))
            content_close = content.close
            check_magic = True

        chunks = list(iter_chunks(content, check_magic=check_magic))
        named_chunks = _validate_chunk_order(chunks)

        arena_info = extract_arena(named_chunks[b"AREN"])
        #print arena_info
        metadata = extract_metadata(named_chunks[b"META"], arena_num_bytes = arena_info["num_bytes"])
        #print metadata
        ids = extract_fpid_ids(named_chunks[b"FPID"])
        #print list(ids)
        if named_chunks[b"HASH"] is None:
            id_lookup = None
        else:
            hash_data = extract_hash_table(named_chunks[b"HASH"])
            hash_lookup = HashLookup(hash_data, ids)
            id_lookup = hash_lookup.get_index_by_id

        popc = named_chunks[b"POPC"]
        if popc is None:
            popcount_indices = b""
        else:
            popc_info = extract_popc(popc)
            # consistency checks go here
            popcount_indices = popc_info["popcount_indices"]
            max_num_bits = popc_info["max_num_bits"]
            if metadata.num_bits > max_num_bits:
                raise FPBError("FPB metadata says there are %d bits but only %d popcount bits are defined" %
                               (metadata.num_bits, max_num_bits))

        cfpl = named_chunks[b"CFPL"]
        if cfpl is None:
            license_key = b""
        else:
            license_key = extract_cfpl(cfpl)

        return FPBFingerprintArena(
            metadata,
            alignment = arena_info["alignment"],
            start_padding = arena_info["fp_start"],
            end_padding = len(content) - arena_info["fp_end"],
            storage_size = arena_info["storage_size"],
            arena = content,
            popcount_indices = popcount_indices,
            arena_ids = ids,
            id_lookup = id_lookup,
            close = CountedCloseMMap(source_close, content_close),
            license_key = license_key,
            )
    except:
        if content_close is not None:
            content_close()
        if source_close is not None:
            source_close()
        raise

def open_fpb_writer(destination, metadata=None, alignment=8, reorder=True,
                    tmpdir=None, max_spool_size=None):
    if not _chemfp.is_licensed():
        # Do a check now, where it's obvious, rather than deeper in the C extension.
        raise NotImplementedError("A valid chemfp license key is required to create an FPB file.")
    
    # Do some extra checks now so that metadata errors appear early.
    # Otherwise they might not appear until the close()
    if metadata is not None:
        if metadata.num_bits is not None:
            if not isinstance(metadata.num_bits, int):
                raise ValueError("metadata num_bits must be an integer")
        if metadata.num_bytes is not None:
            if not isinstance(metadata.num_bytes, int):
                raise ValueError("metadata num_bytes must be an integer")
        # I don't like how this conversion is done twice.
        # I could store the string and use it later.
        # Seems needless as the overhead should be trivial.
        check_metadata_output = BytesIO()
        write_metadata_chunk(check_metadata_output, metadata)
        

    
    if destination is None:
        raise ValueError("FPB output to stdout is not supported")
    if isinstance(destination, _compat.filename_types):
        output = open(destination, "wb")
        close = output.close
    else:
        output = destination # Passed in a file-like object; someone else manages the close.
        close = None

    if reorder:
        return OrderedFPBWriter(output, metadata=metadata, alignment=alignment, include_popc=True, 
                                include_hash=True, max_spool_size=max_spool_size, tmpdir=tmpdir, close=close)
    else:
        return InputOrderFPBWriter(output, metadata=metadata, alignment=alignment, include_popc=True, 
                                   include_hash=True, max_spool_size=max_spool_size, tmpdir=tmpdir, close=close)

    
############################

class Describe(object):
    def __init__(self, fpb_file, log):
        self.fpb_file = fpb_file
        self.log = log
        try:
            self._offset = fpb_file.tell()
        except IOError:
            self._offset = 0
            self._is_seekable = False
        else:
            self._is_seekable = True

    def _die(self):
        raise SystemExit(1)

    def is_seekable(self):
        return self._is_seekable
    
    
    def read_raw(self, size):
        offset = self._offset
        s = self.fpb_file.read(size)
        self._offset += len(s)
        return offset, s
    
    def read_n(self, size, reason):
        offset = self._offset
        s = self.fpb_file.read(size)
        self._offset = offset + len(s)
        if len(s) == size:
            return (offset, s)

        why = ("Incomplete read of %d of %d %s bytes; reached the end of file"
               % (len(s), size, reason))
        self.log.read_error(offset, len(s), why)
        if len(s) <= 40:
            self.log.extra("Read: %r" % (s,))
        else:
            self.log.extra("  Read: %r ... %r" % (s[:20], s[-20:]))
        self._die()

    def skip_forward(self, size):
        start = self.fpb_file.tell()
        self.fpb_file.seek(size, 1)
        end = self.fpb_file.tell()
        if end - start != size:
            self.log.read_error(self._offset, end-start, "skip forward")
        self._offset += size
        return
    
        while size > 0:
            read_size = min(size, 10000000)
            s = self.fpb_file.read(read_size)
            n = len(s)
            if n != read_size:
                self.log.read_error(self._offset, len(s), "skip forward")
            self._offset += n
            size -= n
        
    def read_expected(self, expected, reason):
        offset, s = self.read_n(len(expected), reason)
        if s == expected:
            return (offset, s)
        self.log.read_error(offset, len(s),
                            "Expected to read %s %r but actually read %r" % (reason, expected, s))
        self._die()

    def tell(self):
        return self._offset

class OutputLog(object):
    def __init__(self, outfile):
        self.outfile = outfile
    def successful_read(self, offset, bytes_read, reason):
        self.outfile.write("%d-%d: %s (%r)\n" % (offset, offset+len(bytes_read), reason, bytes_read))
    def successful_large_read(self, offset, bytes_read, reason):
        # For potentially large reads (should I always do this for successful_read?)
        if len(bytes_read) > 50:
            s = repr(bytes_read[:25]) + " ... " + repr(bytes_read[-25:])
            self.outfile.write("%d-%d: %s (%r)\n" % (offset, offset+len(bytes_read), reason, s))
        else:
            self.outfile.write("%d-%d: %s (%r)\n" % (offset, offset+len(bytes_read), reason, bytes_read))
    def successful_read_size(self, offset, num_bytes_read, reason):
        self.outfile.write("%d-%d: %s\n" % (offset, offset+num_bytes_read, reason))
    def read_error(self, offset, num_bytes_read, reason):
        self.outfile.write("%d-%d: ERROR: %s\n" % (offset, offset+num_bytes_read, reason))
    def info(self, msg, depth=1):
        s = " " * depth
        self.outfile.write(s + "INFO: %s\n" % (msg,))
    def warn(self, msg, depth=1):
        s = " " * depth
        self.outfile.write(s + "WARNING: %s\n" % (msg,))
    def error(self, msg, depth=1):
        s = " " * depth
        self.outfile.write(s + "ERROR: %s\n" % (msg,))
    def exit(self, msg, depth=1):
        self.error(msg, depth)
        raise SystemExit(1)
    def extra(self, msg):
        self.outfile.write("> " + msg + "\n")
        

class Options(object):
    def __init__(self, skip_hash_table, skip_ids, show_ids, show_id_indices, skip_fps, show_fps):
        self.skip_hash_table = skip_hash_table
        self.skip_ids = skip_ids
        self.show_ids = show_ids
        self.show_id_indices = show_id_indices
        self.skip_fps = skip_fps
        self.show_fps = show_fps
        
    def __repr__(self):
        return "Options(skip_hash_table=%r, skip_ids=%r, show_ids=%r, show_id_indices=%r, skip_fps=%r, show_fps=%r)" % (
            self.skip_hash_table, self.skip_ids, self.show_ids, self.show_id_indices, self.skip_fps, self.show_fps)

_default_options = Options(
    skip_hash_table=False,
    skip_ids=False,
    show_ids=False,
    show_id_indices=False,
    skip_fps=False,
    show_fps=False,
    )
    
def describe(fpb_file, opts=_default_options, outfile=sys.stdout):
    log = OutputLog(outfile)
    fpb_file = Describe(fpb_file, log)
    try:
        _describe(fpb_file, opts, log)
    except SystemExit:
        log.error("Unable to continue. Stopping.")
        return 1
    return 0
    
def _describe(fpb_file, opts, log):
    offset, magic = fpb_file.read_expected(FPB_MAGIC, "FPB magic signature")
    log.successful_read(offset, magic, "FPB magic signature")
    
    while 1:
        offset, chunk_size_str = fpb_file.read_n(8, "chunk size")
        chunk_size, = _unpack_chunk_size(chunk_size_str)
        log.successful_read(offset, chunk_size_str, "chunk size %d" % (chunk_size,))
        offset, chunk_tag = fpb_file.read_n(4, "chunk tag")
        log.successful_read(offset, chunk_tag, "chunk tag")
        if not _is_valid_tag(chunk_tag):
            log.exit("invalid tag")

        f = globals().get("_describe_chunk_" + chunk_tag.decode("ascii"), None)
        start_where = fpb_file.tell()
        if f is None:
            log.info("===== Unknown chunk tag %s with %d bytes (range %d:%d) ====="
                     % (chunk_tag, chunk_size, start_where, start_where+chunk_size))
            f = _skip_chunk
        else:
            log.info("===== Describing %s chunk with %d bytes (range %d:%d) ===="
                     % (chunk_tag, chunk_size, start_where, start_where+chunk_size))
            

        f(fpb_file, opts, log, chunk_size)
        end_where = fpb_file.tell()
        if end_where - start_where != chunk_size:
            log.exit("%s chunk reader read %d bytes but should have read %d bytes"
                     % (chunk_tag, end_where - start_where, chunk_size))

        log.info("===== Finished with %s chunk ====" % (chunk_tag,))

        if chunk_tag == b"FEND":
            log.info("Reached end of FPB file.", depth=0)
            where = fpb_file.tell()
            offset, s = fpb_file.read_raw(1)
            if s:
                log.info("Extra bytes exist after FEND record, starting at %d" % (where,))
            break
        
def describe_chunks(fpb_file, opts=_default_options, outfile=sys.stdout):
    log = OutputLog(outfile)
    fpb_file = Describe(fpb_file, log)
    
    offset, magic = fpb_file.read_expected(FPB_MAGIC, "FPB magic signature")
    log.successful_read(offset, magic, "FPB magic signature")

    while 1:
        offset, chunk_size_str = fpb_file.read_n(8, "chunk size")
        chunk_size, = _unpack_chunk_size(chunk_size_str)
        log.successful_read(offset, chunk_size_str, "chunk size %d" % (chunk_size,))
        offset, chunk_tag = fpb_file.read_n(4, "chunk tag")
        log.successful_read(offset, chunk_tag, "chunk tag")
        if not _is_valid_tag(chunk_tag):
            log.error("invalid tag")
            
        if chunk_tag == b"FEND":
            log.info("Reached end of FPB file.", depth=0)
            where = fpb_file.tell()
            offset, s = fpb_file.read_raw(1)
            if s:
                log.info("Extra bytes exist after FEND record, starting at %d" % (where,))
            break
    
        fpb_file.skip_forward(chunk_size)
        

def _skip_chunk(fpb_file, opts, log, chunk_size):
    num_read = 0
    while num_read < chunk_size:
        to_read = min(1000000, chunk_size-num_read)
        offset, s = fpb_file.read_raw(to_read)
        log.successful_read_size(offset, len(s), "Skipping ...")
        num_read += len(s)
        if len(s) != to_read:
            log.exit("Only able to read %d bytes of %d.\n" % (num_read, chunk_size))

def _describe_chunk_META(fpb_file, opts, log, chunk_size):
    offset, s = fpb_file.read_n(chunk_size, "META data")
    for line in s.splitlines():
        log.extra(" " + repr(line))

    f = BytesIO(s)
    location = io.Location("META")
    metadata, chunk, chunk_reader = fps_io.read_header(f, location)
    if chunk is not None:
        log.exit("META field contains lines which don't start with '#'")

    if metadata.num_bits is not None:
        log.info("num_bits = %d" % (metadata.num_bits,))
    if metadata.num_bytes is not None:
        log.info("num_bytes = %d" % (metadata.num_bytes,))
    if metadata.type is not None:
        log.info("type = %r" % (metadata.type,))
    if metadata.aromaticity is not None:
        log.info("aromaticity = %r" % (metadata.aromaticity,))
    if metadata.software is not None:
        log.info("software = %r" % (metadata.software,))
    if metadata.sources:
        for i, source in enumerate(metadata.sources):
            log.info("source[%d] = %r" % (i, source))
    if metadata.date:
        log.info("date = %s" % (metadata.date,))

def _describe_chunk_TEXT(fpb_file, opts, log, chunk_size):
    offset, s = fpb_file.read_n(chunk_size, "TEXT data")
    label, nl, content = s.partition(b"\n")
    if not nl:
        log.error("Missing newline between the label and content")
        return
    
    if not label:
        log.error("Empty label")
        return
    
    try:
        unicode_label = label.decode("ascii")
    except UnicodeDecodeError as err:
        log.error("label contains non-ASCII bytes: %r" % (label,))
    else:
        log.info("label = %r" % (unicode_label,))
        
    
    if not content:
        log.info("empty content")
    else:
        try:
            unicode_content = content.decode("utf8")
        except UnicodeDecodeError as err:
            log.error("content does not contain UTF-8 text")
        else:
            log.info("content contains %d bytes and %d characters" %
                         (len(content), len(unicode_content)))
            if len(unicode_content) < 70:
                log.info("content: %r" % (unicode_content,))
            else:
                log.info("content starts: %r" % (unicode_content[:60],))
        
    

def _describe_chunk_AREN(fpb_file, opts, log, chunk_size):
    offset, fingerprint_size_bytes = fpb_file.read_n(4, "fingerprint size")
    fingerprint_size, = _unpack_arena_fingerprint_size(fingerprint_size_bytes)
    log.successful_read(offset, fingerprint_size_bytes, "fingerprint size = %d" % (fingerprint_size, ))

    offset, storage_size_bytes = fpb_file.read_n(4, "storage size")
    storage_size, = _unpack_arena_storage_size(storage_size_bytes)
    log.successful_read(offset, storage_size_bytes, "storage size = %d" % (storage_size,))
    if storage_size > 16384:
        log.warn("Fingerprint storage size of %d is almost certainly too large" % (storage_size,))
    if storage_size == 0:
        log.exit("Fingerprints must use at least one byte of storage, not 0")
    
    offset, spacer_size_byte = fpb_file.read_n(1, "spacer size")
    spacer_size = ord(spacer_size_byte)
    log.successful_read(offset, spacer_size_byte, "spacer size = %d" % (spacer_size,))

    offset, spacer = fpb_file.read_n(spacer_size, "spacer")
    log.successful_read(offset, spacer, "spacer")
    
    if spacer != b"\0" * spacer_size:
        log.warn("Spacer should only contain %d NUL bytes, not %r" % (spacer_size, spacer))

    where = fpb_file.tell()
    for alignment_size in (256, 128, 64, 32, 16, 8, 4, 2, 1):
        if (storage_size % alignment_size == 0 and
            where % alignment_size == 0):
            log.info("fingerprint data is %d byte aligned" % (alignment_size,))
            break
    else:
        raise AssertionError("How can it not be 1 byte aligned?")

    if opts.skip_fps:
        log.info("Skipping fingerprint validation")
        fpb_file.skip_forward(chunk_size - 4 - 4 - 1 - spacer_size)
        return
        
    chunk_size -= (4+4+1+spacer_size)
    fp_index = 0
    extra_size = storage_size - fingerprint_size
    extra_expected_bytes = b"\0" * extra_size
    prev_popcount = 0
    byte_popcount = bitops.byte_popcount
    
    while chunk_size >= storage_size:
        offset, fp = fpb_file.read_n(fingerprint_size, "fp[%d]" % (fp_index,))
        popcount = byte_popcount(fp)
        if opts.show_fps:
            log.successful_read_size(offset, len(fp), "fp[%d]: %s (%d)" % (fp_index, bitops.hex_encode(fp), popcount))
        if prev_popcount is not None and popcount < prev_popcount:
            log.info("Fingerprints are not in popcount order (%d < %d)" % (popcount, prev_popcount))
            prev_popcount = None
        else:
            prev_popcount = popcount
        if extra_size:
            offset, extra_bytes = fpb_file.read_n(extra_size, "+alignment[%d]" % (fp_index,))
            log.successful_read_size(offset, len(extra_bytes), "+alignment[%d]:" % (fp_index,))
            if extra_bytes != extra_expected_bytes:
                log.warn("Extra alignment bytes should only contain %d NUL bytes, not %r"
                         % (extra_size, extra_bytes))
        fp_index += 1
            
        chunk_size -= storage_size

    if chunk_size == 0:
        return

    offset, extra_data = fpb_file.read_raw(min(20, chunk_size))
    chunk_size -= len(extra_data)
    if extra_data:
        log.error("Extra data found after fingerprint fingerprint[%d] and the end of the chunk" % (fp_index,))
        if chunk_size <= 20:
            log.extra("Found %r" % (extra_data,))
        else:
            log.extra("Extra data starts with %r" % (extra_data,))
        raise SystemExit(1)

    if chunk_size:
        if chunk_size > 10000:
            log.exit("Extra unknown fingerprint data is too large to skip.")
        log.info("Skipping %d unknown bytes to get to the end of the chunk" % (chunk_size,))
        fpb_file.read_n(chunk_size, "unknown fingerprint data")
            

def _describe_chunk_POPC(fpb_file, opts, log, chunk_size):
    popcount = 0
    prev_start = None
    while chunk_size >= 4:
        msg = "start of popcount = %d" % popcount
        offset, start_offset_str = fpb_file.read_n(4, msg)
        start_offset, = _unpack_popc_offset(start_offset_str)
        log.successful_read(offset, start_offset_str, msg + " at %d" % (start_offset,))

        if prev_start is not None:
            if prev_start > start_offset:
                log.error("start index of popcount %d (%d) must not be before the start of popcount %d (%d)"
                          % (popcount, start_offset, popcount-1, prev_start))
            else:
                num_fingerprints = (start_offset - prev_start)
                log.info("%d fingerprints in popc[%d:%d]" % (num_fingerprints, popcount-1, popcount))

        prev_start = start_offset
        
        chunk_size -= 4
        popcount += 1
            

    if chunk_size != 0:
        offset, junk = log.read_raw(chunk_size)
        log.successful_read(offset, junk, "unknown data after last popcount index")
        log.exit("Unknown data at end of chunk: %r" % (junk,))
    
class SeekableChunk(object):
    def __init__(self, fpb_file, start_offset, chunk_size):
        self.fpb_file = fpb_file
        self.start_offset = start_offset
        self.chunk_size = chunk_size
    def __getitem__(self, x):
        if x.start < 0:
            start, stop, step = x.indices(self.chunk_size)
        else:
            start, stop, step = x.start, x.stop, x.step
        assert step is None or step == 1, (x, step)
        assert start != stop, (x, start, stop)
        self.fpb_file.seek(self.start_offset + start)
        data = self.fpb_file.read(stop-start)
        assert len(data) == stop-start, (len(data), start, stop)
        #print("Spam", x, (start, stop), (self.start_offset, self.chunk_size), repr(data))
        #print("Spam", x, (start, stop), (self.start_offset, self.chunk_size), repr(data[:10]))
        return data
        
def _describe_chunk_FPID(fpb_file, opts, log, chunk_size):
    unpack = struct.Struct("<I").unpack

    start_offset = fpb_file.tell()
    offset, num_4byte_str = fpb_file.read_n(4, "number of 4-byte identifier offset")
    num_4byte_elements, = unpack(num_4byte_str)
    log.successful_read(offset, num_4byte_str, "%d identifiers with 4-byte offsets" % (num_4byte_elements,))
    
    offset, num_8byte_str = fpb_file.read_n(4, "number of 8-byte identifier offset")
    num_8byte_elements, = unpack(num_8byte_str)
    log.successful_read(offset, num_8byte_str, "%d identifiers with 8-byte offsets" % (num_8byte_elements,))

    if opts.skip_ids:
        log.info("Skipping identifier verification")
        fpb_file.skip_forward(chunk_size-8)
        return
    
    if fpb_file.is_seekable():
        chunk = SeekableChunk(fpb_file.fpb_file, start_offset, chunk_size)
        fpb_file.skip_forward(chunk_size-8)
    else:
        # Cannot seek.
        # I have to read the entire chunk into memory before I can figure out which
        # terms are which.
        offset, chunk = fpb_file.read_n(chunk_size-8, "remaining FPID chunk") # could fail when there isn't enough memory
        chunk = num_4byte_str + num_8byte_str + chunk

    if num_8byte_elements:
        offset = start_offset + chunk_size - 8
        s = chunk[-8:]
        last_data_end, = _unpack_fpid_8byte_offset(s)
        log.successful_read(offset, s, "last_data_end (8-byte offset)")
    else:
        offset = start_offset + chunk_size - 4
        s = chunk[-4:]
        last_data_end, = _unpack_fpid_4byte_offset(s)
        log.successful_read(offset, s, "last_data_end (4-byte offset)")
    log.info("last_data_end = %d" % last_data_end)

    if last_data_end > chunk_size:
        log.error("last_data_end (%d) must be smaller than the chunk size (%d)" %
                      (last_data_end, chunk_size))

    log.info("==== Reading fingerprint id using 4-byte index table with %d entries" % (num_4byte_elements,))
    recno = 0
    for i in range(num_4byte_elements):
        index_start = last_data_end + 4*i
        offset_bytes = chunk[index_start:index_start+8]
        id_start, id_end = struct.unpack("<II", offset_bytes)
        if opts.show_ids:
            log.successful_large_read(id_start, chunk[id_start:id_end], "id[%d]" % (recno,))
                
        recno += 1

    log.info("==== Reading fingerprint id using 8-byte index table with %d entries" % (num_8byte_elements,))
    if num_8byte_elements:
        i = num_4byte_elements
        index_start = last_data_end + 4*i
        offset_bytes = chunk[index_start:index_start+12]
        id_start, id_end = struct.unpack("<IQ", offset_bytes)
        if opts.show_ids:
            log.successful_large_read(id_start, chunk[id_start:id_end], "id[%d]" % (recno,))
        recno += 1

        for i in range(num_8byte_elements):
            index_start = last_data_end + 4*num_4byte_elements + i*8
            offset_bytes = chunk[index_start:index_start+16]
            #print("offset_bytes:", len(offset_bytes), repr(offset_bytes))
            id_start, id_end = struct.unpack("<QQ", offset_bytes)
            if opts.show_ids:
                log.successful_large_read(id_start, chunk[id_start:id_end], "id[%d]" % (recno,))
            recno += 1

    log.info("==== Reading 4-byte index table with %d entries" % (num_4byte_elements,))
    recno = 0
    prev_id_offset = 7
    index_start = last_data_end
    for i in range(num_4byte_elements):
        index_start = last_data_end + 4*i
        data = chunk[index_start:index_start+4]
        id_offset, = struct.unpack("<I", data)
        if opts.show_id_indices:
            log.successful_read_size(index_start, 4, "[%d] = 4[%d] = %d" % (recno, i, id_offset))
        if id_offset <= prev_id_offset:
            log.error("offsets must be monotonically increasing (%d, %d)" % (prev_id_offset, id_offset))
        recno += 1

    if not num_8byte_elements:
        index_start = last_data_end + 4*num_4byte_elements
        data = chunk[index_start:index_start+4]
        id_offset, = struct.unpack("<I", data)
        if opts.show_id_indices:
            log.successful_read_size(index_start, 4, "[%d] = 4[%d] = %d" % (recno, i, id_offset))
        if id_offset <= prev_id_offset:
            log.error("offsets must be monotonically increasing (%d, %d)" % (prev_id_offset, id_offset))
        recno += 1

    log.info("==== Reading 8-byte index table with %d entries" % (num_8byte_elements,))
    for i in range(num_8byte_elements):
        index_start = last_data_end + 4*num_4byte_elements + 8*i
        data = chunk[index_start:index_start+8]
        id_offset, = struct.unpack("<Q", data)
        if opts.show_id_indices:
            log.successful_read_size(index_start, 8, "[%d] = 8[%d] = %d" % (recno, i, id_offset))
        if id_offset <= prev_id_offset:
            log.error("offsets must be monotonically increasing (%d, %d)" % (prev_id_offset, id_offset))
        recno += 1

    if num_8byte_elements:
        index_start = last_data_end + 4*num_4byte_elements + 8*num_8byte_elements
        data = chunk[index_start:index_start+8]
        id_offset, = struct.unpack("<Q", data)
        if opts.show_id_indices:
            log.successful_read_size(index_start, 8, "[%d] = 8[%d] = %d" % (recno, i, id_offset))
        if id_offset <= prev_id_offset:
            log.error("offsets must be monotonically increasing (%d, %d)" % (prev_id_offset, id_offset))
        recno += 1
        index_end = index_start + 8
    else:
        index_end = index_start + 4


    if id_offset != last_data_end:
        log.error("Last offset is %d, which should be the same as last_data_end of %d"
                  % (id_offset, last_data_end))

    if index_end != chunk_size:
        log.error("Last index ends at %d, which should be the same as chunk_size of %d"
                  % (index_end, chunk_size))

    # If this is a SeekableChunk, go to the end
    chunk[-1:]
    
def _describe_chunk_HASH(fpb_file, opts, log, chunk_size):
    if opts.skip_hash_table:
        log.info("Skipping hash table verification")
        fpb_file.skip_forward(chunk_size)
        return
    
    prev_hash_table_end = 0
    hash_table_num_slots = []
    for i in range(256):
        offset, subtable_bytes = fpb_file.read_n(8, "subtable[%d]" % (i,))
        hash_table_start, num_slots = _unpack_hash_main_table_entry(subtable_bytes)
        log.successful_read(offset, subtable_bytes, "subtable[%d] offset=%d num_slots=%d"
                            % (i, hash_table_start, num_slots))
        if hash_table_start != prev_hash_table_end:
            log.error("Hash table %d is supposed to start at offset %d, not %d"
                      % (i, prev_hash_table_end, hash_table_start))
        if num_slots % 2 != 0:
            log.exit("Hash table %d has %d slots, which is not a multiple of 2"
                     % (i, num_slots))
        hash_table_num_slots.append(num_slots)

        prev_hash_table_end = hash_table_start + 8*num_slots

    prev_hash_table_end += 256*8
    if prev_hash_table_end != chunk_size:
        log.exit("The end of hash table 255 (%d) is not aligned with the chunk size (%d)"
                 % (prev_hash_table_end, chunk_size))

    for hash_table_index in range(256):
        num_slots = hash_table_num_slots[hash_table_index]
        log.info("hash table[%d] has %d slots and %d hash pairs"
                 % (hash_table_index, num_slots, num_slots//2))
        num_empty = 0
        for slot in xrange(num_slots):
            offset, slot_bytes = fpb_file.read_n(8, "hash table %d slot %d" % (hash_table_index, slot))
            if slot_bytes == b"\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF":
#                log.successful_read(offset, slot_bytes, "  [%d] slot[%d] (empty)" % (hash_table_index, slot))
                num_empty += 1
            else:
                hash_value, term_index = _unpack_hash_hashpair(slot_bytes)
#                log.successful_read(offset, slot_bytes, "  [%d] slot[%d] hash=%d index=%d"
#                                    % (hash_table_index, slot, hash_value, term_index))
                if hash_value % 256 != hash_table_index:
                    log.exit("Hash value %d is supposed to be in table %d, not %d"
                             % (hash_value, (hash_value % 256), hash_table_index))
        if num_empty*2 != num_slots:
            log.error("hash table %d has %d empty slots but is supposed to have %d"
                      % (hash_table_index, num_empty, num_slots//2))
            
def _describe_chunk_CFPL(fpb_file, opts, log, chunk_size):
    if chunk_size != 16:
        log.error("CFPL block size must be 16 bytes, not %d" % (chunk_size,))

    offset, license_key = fpb_file.read_n(16, "arena license key")
    log.successful_read(offset, license_key, "arena license key (binary)")
    log.info("chemfp arena license key (hex): %s" % (bitops.hex_encode(license_key),))
    if chunk_size > 16:
        fpb_file.skip_forward(chunk_size-16)
    
        

def _describe_chunk_FEND(fpb_file, opts, log, chunk_size):
    if chunk_size != 0:
        log.error("FEND must not contain any data")
        _skip_chunk(fpb_file, log, chunk_size)

def main(argv=None):
    try:
        import argparse
    except ImportError:
        from . import argparse
    parser = argparse.ArgumentParser(
        description="show the low-level contents of an FPB file")
    parser.add_argument("--chunk-info", action="store_true",
                            help="only print chunk information")
    parser.add_argument("--skip-hash-table", action="store_true",
                            help="do not show or validate the HASH chunk")
    parser.add_argument("--skip-ids", action="store_true",
                            help="do not show or validate the FPID chunk")
    parser.add_argument("--show-ids", action="store_true",
                            help="include ids in the output")
    parser.add_argument("--show-id-indices", action="store_true",
                            help="include id indices in the output")
    parser.add_argument("--skip-fps", action="store_true",
                            help="do not show or validate fingerprints")
    parser.add_argument("--show-fps", action="store_true",
                            help="include fingerprints in the output")
    parser.add_argument("filename", help="FPB filename")
    args = parser.parse_args(args=argv)

    try:
        f = open(args.filename, "rb")
    except Exception as err:
        sys.stderr.write("Cannot open FPB file: %s\n" % (err,))
        raise SystemExit(1)

    opts = Options(
        skip_hash_table = args.skip_hash_table,
        skip_ids = args.skip_ids,
        show_ids = args.show_ids,
        show_id_indices = args.show_id_indices,
        skip_fps = args.skip_fps,
        show_fps = args.show_fps)
    try:
        if args.chunk_info:
            describe_chunks(f, opts)
        else:
            describe(f, opts)
    finally:
        f.close()
    
if __name__ == "__main__":
    import signal
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    main()
