# Copyright (c) 2010-2020 Andrew Dalke Scientific, AB (Sweden)
# All Rights Reserved.

## LICENSE ##

# This file contains proprietary and unpublished source code.
#
# You may use and modify this software ONLY under the terms of the
# License agreement you or your organization made with Andrew Dalke
# Scientific AB.
#
# If no other agreeement exists then you may use chemfp under the
# terms of the Chemfp Base License Agreement v1.0 or later.
# See http://chemfp.com/licenses/ for details.
#
# Unless required by applicable law or agreed to in writing, this
# software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.  See the License
# agreement for the specific language governing permissions and
# limitations under the License.

## /LICENSE ##

from __future__ import absolute_import

import openbabel
if getattr(openbabel, "__version__", "2.0")[:2] == "2.":
    from openbabel import OBSmartsPattern, OBBitVec, OBAtom
else:
    from openbabel.openbabel import OBSmartsPattern, OBBitVec, OBAtom


from . import openbabel_toolkit
from . import openbabel_types
from . import pattern_fingerprinter
from . import types
from . import __version__ as chemfp_version
from ._compat import ints2bytes

from . import SOFTWARE as CHEMFP_SOFTWARE
SOFTWARE = openbabel_toolkit.software + " " + CHEMFP_SOFTWARE

# HasMatch was added to OpenBabel 2.3 .
_HAS_HASMATCH = hasattr(OBSmartsPattern(), "HasMatch")

# ImplicitHydrogenCount() was removed on May 2017 and replaced with
# GetImplicitHCount().

if hasattr(OBAtom, "ImplicitHydrogenCount"):
    def _get_hcount(obatom):
        return obatom.ImplicitHydrogenCount()
else:
    def _get_hcount(obatom):
        return obatom.GetImplicitHCount()


class Matcher(object):
    def __init__(self, ob_matcher):
        self.ob_matcher = ob_matcher

    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self.ob_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self.ob_matcher.Match(mol, True)
        
    def NumUniqueMatches(self, mol):
        self.ob_matcher.Match(mol)
        return sum(1 for x in self.ob_matcher.GetUMapList())

class HydrogenMatcher(object):
    def __init__(self, max_count):
        self.max_count = max_count
    def HasMatch(self, mol):
        for i in range(1, mol.NumAtoms()+1):
            atom = mol.GetAtom(i)
            if atom.GetAtomicNum() == 1:
                return True
            if _get_hcount(atom):
                return True
        return False
        
    def NumUniqueMatches(self, mol):
        count = 0
        for i in range(1, mol.NumAtoms()+1):
            atom = mol.GetAtom(i)
            if atom.GetAtomicNum() == 1:
                count += 1
            count += _get_hcount(atom)
            if count >= self.max_count:
                return count
        return count

class AromaticRings(object):
    def __init__(self, max_count):
        self.max_count = max_count
        # The single ring case is easy; if there's an aromatic atom in a ring
        # then there's an aromatic ring.
        self._single_matcher = OBSmartsPattern()
        assert self._single_matcher.Init("[aR]")
        
    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self._single_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self._single_matcher.Match(mol, True)
        

    def NumUniqueMatches(self, mol):
        num_rings = 0
        for ring in mol.GetSSSR():
            if ring.IsAromatic():
                num_rings += 1
                if num_rings == self.max_count:
                    break
        return num_rings


###

class HeteroAromaticRings(object):
    def __init__(self, max_count):
        self.max_count = max_count
        # The single ring case is easy; if there's an non-carbon aromatic atom
        # in a ring then there's a hetero-aromatic ring
        self._single_matcher = OBSmartsPattern()
        assert self._single_matcher.Init("[aR;!#6]")

    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self._single_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self._single_matcher.Match(mol, True)
        

    def NumUniqueMatches(self, mol):
        num_rings = 0
        for ring in mol.GetSSSR():
            if (ring.IsAromatic() and
                  any(mol.GetAtom(atom_id).GetAtomicNum() != 6 for atom_id in ring._path)):
                num_rings += 1
                if num_rings == self.max_count:
                    break
        return num_rings


class NumFragments(object):
    def __init__(self, max_count):
        if max_count >= 3:
            raise NotImplementedError("<fragment> >= 3 not implemented")
    def HasMatch(self, mol):
        # Has at least one fragment, which is the same as having atoms
        return (mol.NumAtoms() > 0)
    def NumUniqueMatches(self, mol):
        # If the largest fragment is not the same size as the entire molecule
        # then there are at least two fragments
        n = mol.NumAtoms()
        if n == 0:
            # No atoms means no fragments
            return 0
        
        v = OBBitVec()
        mol.FindLargestFragment(v)
        if v.CountBits() == n:
            return 1
        
        return 2

_pattern_classes = {
    "<H>": HydrogenMatcher,
    "<aromatic-rings>": AromaticRings,
    "<hetero-aromatic-rings>": HeteroAromaticRings,
    "<fragments>": NumFragments,
    }

def ob_compile_pattern(pattern, max_count):
    if pattern in _pattern_classes:
        return _pattern_classes[pattern](max_count)
    
    if pattern.startswith("<"):
        raise NotImplementedError(pattern)

    matcher = OBSmartsPattern()
    if not matcher.Init(pattern):
        raise pattern_fingerprinter.UnsupportedPatternError(
            pattern, "Uninterpretable SMARTS pattern")
    return Matcher(matcher)

class OBPatternFingerprinter(pattern_fingerprinter.PatternFingerprinter):
    def __init__(self, patterns):
        assert patterns is not None
        super(OBPatternFingerprinter, self).__init__(patterns, ob_compile_pattern)
        
    def fingerprint(self, mol):
        bytes = [0] * self.num_bytes
        for matcher, largest_count, count_info_tuple in self.matcher_definitions:
            #print matcher, largest_count, count_info_tuple
            if largest_count == 1:
                if matcher.HasMatch(mol):
                    count_info = count_info_tuple[0]
                    bytes[count_info.byteno] |= count_info.bitmask
            else:
                # There's no good way to get the number of unique
                # matches without iterating over all of them.
                actual_count = matcher.NumUniqueMatches(mol)
                for count_info in count_info_tuple:
                    if actual_count >= count_info.count:
                        bytes[count_info.byteno] |= count_info.bitmask
                    else:
                        break
        return ints2bytes(bytes)


class _CachedFingerprinters(dict):
    def __missing__(self, name):
        patterns = pattern_fingerprinter._load_named_patterns(name)
        fingerprinter = OBPatternFingerprinter(patterns)
        self[name] = fingerprinter
        return fingerprinter
_cached_fingerprinters = _CachedFingerprinters()

# XXX Why are there two "Fingerprinter" classes?
# XX Shouldn't they be merged?

class SubstructOpenBabelFingerprinter_v1(types.NoFingerprintParametersMixin, openbabel_types.OpenBabelBaseFingerprintType):
    """chemfp's Substruct fingerprint implementation for Open Babel, version 1

    WARNING: these fingerprints have not been validated.

    The Substruct fingerprints are CACTVS/PubChem-like fingerprints designed
    for use across multiple toolkits.

    The ChemFP-Substruct-OpenBabel/1 :class:`.FingerprintType` has no parameters.
    """
    name = "ChemFP-Substruct-OpenBabel/1"
    num_bits = 881
    software = SOFTWARE
    def make_fingerprinter(self):
        return _cached_fingerprinters["substruct"].fingerprint

class RDMACCSOpenBabelFingerprinter_v1(types.NoFingerprintParametersMixin, openbabel_types.OpenBabelBaseFingerprintType):
    """chemfp's RDMACCS fingerprint implementation for Open Babel, version 1

    The RDMACSS keys are MACCS-166-like fingerprints based on RDKit's
    MACCS116 definition, but designed to be (slightly) more portable
    across multiple chemistry toolkits.

    This version does not define key 44.

    The RDMACSS-OpenBabel/1 :class:`.FingerprintType` has no parameters.
    """
    name = "RDMACCS-OpenBabel/1"
    num_bits = 166
    def make_fingerprinter(self):
        return _cached_fingerprinters["rdmaccs"].fingerprint

class RDMACCSOpenBabelFingerprinter_v2(types.NoFingerprintParametersMixin, openbabel_types.OpenBabelBaseFingerprintType):
    """chemfp's RDMACCS fingerprint implementation for Open Babel, version 2

    The RDMACSS keys are MACCS-166-like fingerprints based on RDKit's
    MACCS116 definition, but designed to be (slightly) more portable
    across multiple chemistry toolkits.

    This version defines key 44.

    The RDMACSS-OpenBabel/2 :class:`.FingerprintType` has no parameters.
    """
    name = "RDMACCS-OpenBabel/2"
    num_bits = 166
    def make_fingerprinter(self):
        return _cached_fingerprinters["rdmaccs2"].fingerprint

def resolve_chemfp_fingerprint_types(resolver):
    register = resolver.register_fingerprint_type
    register(SubstructOpenBabelFingerprinter_v1)
    register(RDMACCSOpenBabelFingerprinter_v2)
    register(RDMACCSOpenBabelFingerprinter_v1)
